@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,12 +21,63 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * $Log: asner.h,v $
+ * Revision 1.41  2003/08/18 23:32:22  rjongbloed
+ * Micro optimisation suggested by Chih-Wei Huang
+ *
+ * Revision 1.40  2003/08/01 16:00:51  csoutheren
+ * Changed #if to #ifdef to (maybe) avoid compiler problems with gcc 2.95.2
+ *
+ * Revision 1.39  2003/08/01 02:12:34  csoutheren
+ * Changed to allow easy isolation of PER, BER and XER encoding/decoding routines
+ *
+ * Revision 1.38  2003/04/22 23:39:09  craigs
+ * Changed some functions from protected to public for MacOSX. Thanks to Hugo Santos
+ *
+ * Revision 1.37  2003/04/17 14:44:44  craigs
+ * Removed MacOS specific defines to make some attributes public
+ * Thanks to Hugo Santos and apologies to Roger Hardiman
+ *
+ * Revision 1.36  2003/02/26 01:57:44  robertj
+ * Added XML encoding rules to ASN system, thanks Federico Pinna
+ *
+ * Revision 1.35  2003/02/01 13:25:52  robertj
+ * Added function to add new elements directly to ASN array.
+ *
+ * Revision 1.34  2003/01/24 23:43:43  robertj
+ * Fixed subtle problems with the use of MAX keyword for unsigned numbers,
+ *   should beUINT_MAX not INT_MAX, thanks Stevie Gray for pointing it out.
+ *
+ * Revision 1.33  2002/11/26 23:29:18  robertj
+ * Added missing const to DecodeSubType() function.
+ *
+ * Revision 1.32  2002/11/06 22:47:23  robertj
+ * Fixed header comment (copyright etc)
+ *
+ * Revision 1.31  2002/10/31 05:50:49  robertj
+ * Changed to use new UTF-8/UCS-2 conversion functions on PString.
+ *
+ * Revision 1.30  2002/10/10 14:37:40  rogerh
+ * In two of the PASN classes make the protected members public. This
+ * makes OpenH323 compile in Mac OS X 10.2.1
+ *
+ * Revision 1.29  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.28  2002/05/29 01:22:35  robertj
+ * Added ability to set object id from unsigned integer arrays.
+ *
+ * Revision 1.27  2002/05/21 04:23:40  robertj
+ * Fixed problem with ASN encoding/decoding unsconstrained negative numbers,
+ *
+ * Revision 1.26  2002/05/14 06:59:31  robertj
+ * Added more bullet proofing so a malformed PDU cannot cause teh decoder
+ *   to try and allocate huge arrays and consume all CPU and memory on a
+ *   system. A configurable limit of 100 is set for things like SEQUENCE OF.
+ *
  * Revision 1.25  2001/12/13 09:13:28  robertj
  * Added function get get oid as a string.
  * Added functions to compare oid to PString.
@@ -101,7 +152,7 @@
  * Fixed problems with using copy constructor/assignment oeprator on PASN_Objects.
  *
  * Revision 1.3  1997/12/18 05:08:13  robertj
- * Added function to get choice discriminator name.
+ * Added function to get choice discriminat`or name.
  *
  * Revision 1.2  1997/12/11 10:35:42  robertj
  * Support for new ASN file parser.
@@ -111,15 +162,26 @@
 #ifndef _ASNER_H
 #define _ASNER_H
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
+// provide options to omit vertain encodings, if needed
+#define     P_INCLUDE_PER
+#define     P_INCLUDE_BER
+#define     P_INCLUDE_XER
 
 class PASN_Stream;
 class PBER_Stream;
 class PPER_Stream;
 
+#if P_EXPAT
+class PXER_Stream;
+class PXMLElement;
+#else
+#undef      P_INCLUDE_XER
+#endif
+
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -195,8 +257,8 @@ class PASN_Object : public PObject
     enum MaximumValueTag { MaximumValue = INT_MAX };
     void SetConstraints(ConstraintType type, int value)
       { SetConstraintBounds(type, value, value); }
-    void SetConstraints(ConstraintType, int lower, MaximumValueTag upper)
-      { SetConstraintBounds(PartiallyConstrained, (int)lower, (unsigned)upper); }
+    void SetConstraints(ConstraintType, int lower, MaximumValueTag /*upper*/)
+      { SetConstraintBounds(PartiallyConstrained, (int)lower, lower < 0 ? INT_MAX : UINT_MAX); }
     void SetConstraints(ConstraintType, MinimumValueTag lower, unsigned upper)
       { SetConstraintBounds(PartiallyConstrained, (int)lower, (unsigned)upper); }
     void SetConstraints(ConstraintType, MinimumValueTag lower, MaximumValueTag upper)
@@ -208,6 +270,11 @@ class PASN_Object : public PObject
     virtual void SetCharacterSet(ConstraintType ctype, const char * charSet);
     virtual void SetCharacterSet(ConstraintType ctype, unsigned firstChar, unsigned lastChar);
 
+    static PINDEX GetMaximumArraySize();
+    static void SetMaximumArraySize(PINDEX sz);
+    static PINDEX GetMaximumStringSize();
+    static void SetMaximumStringSize(PINDEX sz);
+
   protected:
     PASN_Object(unsigned tag, TagClass tagClass, BOOL extend = FALSE);
 
@@ -317,8 +384,12 @@ class PASN_Integer : public PASN_ConstrainedObject
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
+
+    BOOL IsUnsigned() const;
 
   protected:
     unsigned value;
@@ -360,8 +431,15 @@ class PASN_Enumeration : public PASN_Object
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
+
+#ifdef P_INCLUDE_XER
+    virtual BOOL DecodeXER(PXER_Stream & strm);
+    virtual void EncodeXER(PXER_Stream & strm) const;
+#endif
 
   protected:
     unsigned maxEnumValue;
@@ -412,7 +490,10 @@ class PASN_ObjectId : public PASN_Object
 
     PASN_ObjectId & operator=(const char * dotstr);
     PASN_ObjectId & operator=(const PString & dotstr);
+    PASN_ObjectId & operator=(const PUnsignedArray & numbers);
     void SetValue(const PString & dotstr);
+    void SetValue(const PUnsignedArray & numbers) { value = numbers; }
+    void SetValue(const unsigned * numbers, PINDEX size);
 
     BOOL operator==(const char * dotstr) const;
     BOOL operator!=(const char * dotstr) const      { return !operator==(dotstr); }
@@ -477,10 +558,15 @@ class PASN_BitString : public PASN_ConstrainedObject
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_BER
     BOOL DecodeBER(PBER_Stream & strm, unsigned len);
     void EncodeBER(PBER_Stream & strm) const;
+#endif
+
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
 
     BOOL DecodeSequenceExtensionBitmap(PPER_Stream & strm);
     void EncodeSequenceExtensionBitmap(PPER_Stream & strm) const;
@@ -518,7 +604,7 @@ class PASN_OctetString : public PASN_ConstrainedObject
     BYTE & operator[](PINDEX i) { return value[i]; }
     BYTE * GetPointer(PINDEX sz = 0) { return value.GetPointer(sz); }
     PINDEX GetSize() const { return value.GetSize(); }
-    BOOL SetSize(PINDEX newSize) { return value.SetSize(newSize); }
+    BOOL SetSize(PINDEX newSize);
 
     virtual Comparison Compare(const PObject & obj) const;
     virtual PObject * Clone() const;
@@ -530,10 +616,12 @@ class PASN_OctetString : public PASN_ConstrainedObject
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
 
-    BOOL DecodeSubType(PASN_Object &);
+    BOOL DecodeSubType(PASN_Object &) const;
     void EncodeSubType(const PASN_Object &);
 
   protected:
@@ -567,10 +655,15 @@ class PASN_ConstrainedString : public PASN_ConstrainedObject
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_BER
     BOOL DecodeBER(PBER_Stream & strm, unsigned len);
     void EncodeBER(PBER_Stream & strm) const;
+#endif
+
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
 
   protected:
     PASN_ConstrainedString(const char * canonicalSet, PINDEX setSize,
@@ -618,15 +711,15 @@ class PASN_BMPString : public PASN_ConstrainedObject
     PASN_BMPString(const PASN_BMPString & other);
     PASN_BMPString & operator=(const PASN_BMPString & other);
 
-    PASN_BMPString & operator=(const char * v);
-    PASN_BMPString & operator=(const PString & v) { return operator=((const char *)v); }
+    PASN_BMPString & operator=(const char * v) { return operator=(PString(v).AsUCS2()); }
+    PASN_BMPString & operator=(const PString & v) { return operator=(v.AsUCS2()); }
     PASN_BMPString & operator=(const PWORDArray & v);
     operator PString() const { return GetValue(); }
     operator PWORDArray() const { return value; }
-    PString GetValue() const;
+    PString GetValue() const { return value; }
     void GetValue(PWORDArray & v) const { v = value; }
-    void SetValue(const char * v) { operator=(v); }
-    void SetValue(const PString & v) { operator=((const char *)v); }
+    void SetValue(const char * v) { operator=(PString(v).AsUCS2()); }
+    void SetValue(const PString & v) { operator=(v.AsUCS2()); }
     void SetValue(const PWORDArray & v) { operator=(v); }
     void SetValue(const PASN_BMPString & v) { operator=(v.value); }
 
@@ -643,10 +736,15 @@ class PASN_BMPString : public PASN_ConstrainedObject
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_BER
     BOOL DecodeBER(PBER_Stream & strm, unsigned len);
     void EncodeBER(PBER_Stream & strm) const;
+#endif
+
+#ifdef P_INCLUDE_PER
     BOOL DecodePER(PPER_Stream & strm);
     void EncodePER(PPER_Stream & strm) const;
+#endif
 
   protected:
     void Construct();
@@ -774,8 +872,17 @@ class PASN_Choice : public PASN_Object
     virtual BOOL Decode(PASN_Stream &);
     virtual void Encode(PASN_Stream &) const;
 
+#ifdef P_INCLUDE_PER
     virtual BOOL DecodePER(PPER_Stream &);
     virtual void EncodePER(PPER_Stream &) const;
+#endif
+
+#ifdef P_INCLUDE_XER
+    BOOL DecodeXER(PXER_Stream &);
+    void EncodeXER(PXER_Stream &) const;
+#endif
+
+    PASN_Choice & operator=(const PASN_Choice & other);
 
   protected:
     PASN_Choice(unsigned nChoices = 0, BOOL extend = FALSE);
@@ -783,7 +890,6 @@ class PASN_Choice : public PASN_Object
     PASN_Choice(unsigned tag, TagClass tagClass, unsigned nChoices, BOOL extend, const PString & nameSpec);
 
     PASN_Choice(const PASN_Choice & other);
-    PASN_Choice & operator=(const PASN_Choice & other);
 
     BOOL CheckCreate() const;
 
@@ -810,7 +916,7 @@ class PASN_Sequence : public PASN_Object
     PASN_Sequence & operator=(const PASN_Sequence & other);
 
     PINDEX GetSize() const { return fields.GetSize(); }
-    void SetSize(PINDEX newSize);
+    BOOL SetSize(PINDEX newSize);
     PASN_Object & operator[](PINDEX i) const { return fields[i]; }
 
     BOOL HasOptionalField(PINDEX opt) const;
@@ -834,19 +940,32 @@ class PASN_Sequence : public PASN_Object
     BOOL UnknownExtensionsDecode(PASN_Stream & strm);
     void UnknownExtensionsEncode(PASN_Stream & strm) const;
 
+#ifdef P_INCLUDE_BER
     BOOL PreambleDecodeBER(PBER_Stream & strm);
     void PreambleEncodeBER(PBER_Stream & strm) const;
     BOOL KnownExtensionDecodeBER(PBER_Stream & strm, PINDEX fld, PASN_Object & field);
     void KnownExtensionEncodeBER(PBER_Stream & strm, PINDEX fld, const PASN_Object & field) const;
     BOOL UnknownExtensionsDecodeBER(PBER_Stream & strm);
     void UnknownExtensionsEncodeBER(PBER_Stream & strm) const;
+#endif
 
+#ifdef P_INCLUDE_PER
     BOOL PreambleDecodePER(PPER_Stream & strm);
     void PreambleEncodePER(PPER_Stream & strm) const;
     BOOL KnownExtensionDecodePER(PPER_Stream & strm, PINDEX fld, PASN_Object & field);
     void KnownExtensionEncodePER(PPER_Stream & strm, PINDEX fld, const PASN_Object & field) const;
     BOOL UnknownExtensionsDecodePER(PPER_Stream & strm);
     void UnknownExtensionsEncodePER(PPER_Stream & strm) const;
+#endif
+
+#ifdef P_INCLUDE_XER
+    virtual BOOL PreambleDecodeXER(PXER_Stream & strm);
+    virtual void PreambleEncodeXER(PXER_Stream & strm) const;
+    virtual BOOL KnownExtensionDecodeXER(PXER_Stream & strm, PINDEX fld, PASN_Object & field);
+    virtual void KnownExtensionEncodeXER(PXER_Stream & strm, PINDEX fld, const PASN_Object & field) const;
+    virtual BOOL UnknownExtensionsDecodeXER(PXER_Stream & strm);
+    virtual void UnknownExtensionsEncodeXER(PXER_Stream & strm) const;
+#endif
 
   protected:
     BOOL NoExtensionsToDecode(PPER_Stream & strm);
@@ -883,8 +1002,9 @@ class PASN_Array : public PASN_ConstrainedObject
     PCLASSINFO(PASN_Array, PASN_ConstrainedObject);
   public:
     PINDEX GetSize() const { return array.GetSize(); }
-    void SetSize(PINDEX newSize);
+    BOOL SetSize(PINDEX newSize);
     PASN_Object & operator[](PINDEX i) const { return array[i]; }
+    void Append(PASN_Object * obj) { array.SetAt(array.GetSize(), obj); }
     void RemoveAt(PINDEX i) { array.RemoveAt(i); }
     void RemoveAll() { array.RemoveAll(); }
 
@@ -900,12 +1020,13 @@ class PASN_Array : public PASN_ConstrainedObject
 
     virtual PASN_Object * CreateObject() const = 0;
 
+    PASN_Array & operator=(const PASN_Array & other);
+
   protected:
     PASN_Array(unsigned tag = UniversalSequence,
                TagClass tagClass = UniversalTagClass);
 
     PASN_Array(const PASN_Array & other);
-    PASN_Array & operator=(const PASN_Array & other);
 
     PASN_ObjectArray array;
 };
@@ -982,136 +1103,16 @@ class PASN_Stream : public PBYTEArray
     void Construct();
 };
 
+#ifdef  P_INCLUDE_PER
+#include "asnper.h"
+#endif
 
-/** Class for ASN basic Encoding Rules stream.
-*/
-class PBER_Stream : public PASN_Stream
-{
-    PCLASSINFO(PBER_Stream, PASN_Stream);
-  public:
-    PBER_Stream();
-    PBER_Stream(const PBYTEArray & bytes);
-    PBER_Stream(const BYTE * buf, PINDEX size);
-
-    PBER_Stream & operator=(const PBYTEArray & bytes);
-
-    virtual BOOL Read(PChannel & chan);
-    virtual BOOL Write(PChannel & chan);
-
-    virtual BOOL NullDecode(PASN_Null &);
-    virtual void NullEncode(const PASN_Null &);
-    virtual BOOL BooleanDecode(PASN_Boolean &);
-    virtual void BooleanEncode(const PASN_Boolean &);
-    virtual BOOL IntegerDecode(PASN_Integer &);
-    virtual void IntegerEncode(const PASN_Integer &);
-    virtual BOOL EnumerationDecode(PASN_Enumeration &);
-    virtual void EnumerationEncode(const PASN_Enumeration &);
-    virtual BOOL RealDecode(PASN_Real &);
-    virtual void RealEncode(const PASN_Real &);
-    virtual BOOL ObjectIdDecode(PASN_ObjectId &);
-    virtual void ObjectIdEncode(const PASN_ObjectId &);
-    virtual BOOL BitStringDecode(PASN_BitString &);
-    virtual void BitStringEncode(const PASN_BitString &);
-    virtual BOOL OctetStringDecode(PASN_OctetString &);
-    virtual void OctetStringEncode(const PASN_OctetString &);
-    virtual BOOL ConstrainedStringDecode(PASN_ConstrainedString &);
-    virtual void ConstrainedStringEncode(const PASN_ConstrainedString &);
-    virtual BOOL BMPStringDecode(PASN_BMPString &);
-    virtual void BMPStringEncode(const PASN_BMPString &);
-    virtual BOOL ChoiceDecode(PASN_Choice &);
-    virtual void ChoiceEncode(const PASN_Choice &);
-    virtual BOOL ArrayDecode(PASN_Array &);
-    virtual void ArrayEncode(const PASN_Array &);
-    virtual BOOL SequencePreambleDecode(PASN_Sequence &);
-    virtual void SequencePreambleEncode(const PASN_Sequence &);
-    virtual BOOL SequenceKnownDecode(PASN_Sequence &, PINDEX, PASN_Object &);
-    virtual void SequenceKnownEncode(const PASN_Sequence &, PINDEX, const PASN_Object &);
-    virtual BOOL SequenceUnknownDecode(PASN_Sequence &);
-    virtual void SequenceUnknownEncode(const PASN_Sequence &);
-
-    virtual PASN_Object * CreateObject(unsigned tag,
-                                       PASN_Object::TagClass tagClass,
-                                       BOOL primitive) const;
-
-    BOOL HeaderDecode(unsigned & tagVal,
-                      PASN_Object::TagClass & tagClass,
-                      BOOL & primitive,
-                      unsigned & len);
-    BOOL HeaderDecode(PASN_Object & obj, unsigned & len);
-    void HeaderEncode(const PASN_Object & obj);
-};
-
-
-/** Class for ASN Packed Encoding Rules stream.
-*/
-class PPER_Stream : public PASN_Stream
-{
-    PCLASSINFO(PPER_Stream, PASN_Stream);
-  public:
-    PPER_Stream(BOOL aligned = TRUE);
-    PPER_Stream(const PBYTEArray & bytes, BOOL aligned = TRUE);
-    PPER_Stream(const BYTE * buf, PINDEX size, BOOL aligned = TRUE);
-
-    PPER_Stream & operator=(const PBYTEArray & bytes);
-
-    unsigned GetBitsLeft() const;
-
-    virtual BOOL Read(PChannel & chan);
-    virtual BOOL Write(PChannel & chan);
-
-    virtual BOOL NullDecode(PASN_Null &);
-    virtual void NullEncode(const PASN_Null &);
-    virtual BOOL BooleanDecode(PASN_Boolean &);
-    virtual void BooleanEncode(const PASN_Boolean &);
-    virtual BOOL IntegerDecode(PASN_Integer &);
-    virtual void IntegerEncode(const PASN_Integer &);
-    virtual BOOL EnumerationDecode(PASN_Enumeration &);
-    virtual void EnumerationEncode(const PASN_Enumeration &);
-    virtual BOOL RealDecode(PASN_Real &);
-    virtual void RealEncode(const PASN_Real &);
-    virtual BOOL ObjectIdDecode(PASN_ObjectId &);
-    virtual void ObjectIdEncode(const PASN_ObjectId &);
-    virtual BOOL BitStringDecode(PASN_BitString &);
-    virtual void BitStringEncode(const PASN_BitString &);
-    virtual BOOL OctetStringDecode(PASN_OctetString &);
-    virtual void OctetStringEncode(const PASN_OctetString &);
-    virtual BOOL ConstrainedStringDecode(PASN_ConstrainedString &);
-    virtual void ConstrainedStringEncode(const PASN_ConstrainedString &);
-    virtual BOOL BMPStringDecode(PASN_BMPString &);
-    virtual void BMPStringEncode(const PASN_BMPString &);
-    virtual BOOL ChoiceDecode(PASN_Choice &);
-    virtual void ChoiceEncode(const PASN_Choice &);
-    virtual BOOL ArrayDecode(PASN_Array &);
-    virtual void ArrayEncode(const PASN_Array &);
-    virtual BOOL SequencePreambleDecode(PASN_Sequence &);
-    virtual void SequencePreambleEncode(const PASN_Sequence &);
-    virtual BOOL SequenceKnownDecode(PASN_Sequence &, PINDEX, PASN_Object &);
-    virtual void SequenceKnownEncode(const PASN_Sequence &, PINDEX, const PASN_Object &);
-    virtual BOOL SequenceUnknownDecode(PASN_Sequence &);
-    virtual void SequenceUnknownEncode(const PASN_Sequence &);
-
-    BOOL IsAligned() const { return aligned; }
-
-    BOOL SingleBitDecode();
-    void SingleBitEncode(BOOL value);
-
-    BOOL MultiBitDecode(unsigned nBits, unsigned & value);
-    void MultiBitEncode(unsigned value, unsigned nBits);
-
-    BOOL SmallUnsignedDecode(unsigned & value);
-    void SmallUnsignedEncode(unsigned value);
-
-    int LengthDecode(unsigned lower, unsigned upper, unsigned & len);
-    void LengthEncode(unsigned len, unsigned lower, unsigned upper);
-
-    int UnsignedDecode(unsigned lower, unsigned upper, unsigned & value);
-    void UnsignedEncode(int value, unsigned lower, unsigned upper);
-
-    void AnyTypeEncode(const PASN_Object * value);
-
-  protected:
-    BOOL aligned;
-};
+#ifdef  P_INCLUDE_BER
+#include "asnber.h"
+#endif
 
+#ifdef  P_INCLUDE_XER
+#include "asnxer.h"
+#endif
 
 #endif // _ASNER_H
