@@ -27,6 +27,13 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: args.h,v $
+ * Revision 1.23  2003/03/27 07:27:07  robertj
+ * Added function to get a bunch of arguments as a string array.
+ *
+ * Revision 1.22  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
  * Revision 1.21  2001/12/15 04:49:06  robertj
  * Added stream I/O functions for argument list.
  *
@@ -98,7 +105,7 @@
 #ifndef _PARGLIST
 #define _PARGLIST
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -312,6 +319,15 @@ class PArgList : public PObject
      */
     PINDEX GetCount() const;
 
+    /** Get the parameters that were parsed in the argument list.
+
+       @return array of parameter strings at the specified index range.
+     */
+    PStringArray GetParameters(
+      PINDEX first = 0,
+      PINDEX last = P_MAX_INDEX
+    ) const;
+
     /** Get the parameter that was parsed in the argument list.
 
        @return parameter string at the specified index.
