@@ -27,6 +27,16 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: file.h,v $
+ * Revision 1.39  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.38  2002/07/02 08:00:55  craigs
+ * Also made GetPosition, SetPosition and SetLength virtual as well
+ *
+ * Revision 1.37  2002/07/02 07:59:42  craigs
+ * Added virtual to GetLength call
+ *
  * Revision 1.36  2002/01/13 20:54:55  rogerh
  * Make the other Open() virtual so it can be overridden (eg in PWAVFile)
  *
@@ -144,7 +154,7 @@
 
 #define _PFILE
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -549,7 +559,7 @@ class PFile : public PChannel
        @return
        length of file in bytes.
      */
-    off_t GetLength() const;
+    virtual off_t GetLength() const;
       
     /**Set the size of the file, padding with 0 bytes if it would require
        expanding the file, or truncating it if being made shorter.
@@ -557,7 +567,7 @@ class PFile : public PChannel
        @return
        TRUE if the file size was changed to the length specified.
      */
-    BOOL SetLength(
+    virtual BOOL SetLength(
       off_t len   // New length of file.
     );
 
@@ -581,7 +591,7 @@ class PFile : public PChannel
        @return
        TRUE if the new file position was set.
      */
-    BOOL SetPosition(
+    virtual BOOL SetPosition(
       off_t pos,                         /// New position to set.
       FilePositionOrigin origin = Start  /// Origin for position change.
     );
@@ -592,7 +602,7 @@ class PFile : public PChannel
        @return
        current file position relative to start of file.
      */
-    off_t GetPosition() const;
+    virtual off_t GetPosition() const;
 
     /**Determine if the current file position is at the end of the file. If
        this is TRUE then any read operation will fail.
