@@ -27,6 +27,44 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: ipsock.h,v $
+ * Revision 1.60  2003/05/21 09:34:43  rjongbloed
+ * Name lookup support for IPv6, thanks again Sbastien Josset
+ *
+ * Revision 1.59  2003/04/28 02:55:08  robertj
+ * Added function to see at run time if IPv6 available, thanks Sebastien Josset
+ *
+ * Revision 1.58  2003/04/03 08:43:23  robertj
+ * Added IPv4 mapping into IPv6, thanks Sebastien Josset
+ *
+ * Revision 1.57  2003/03/26 05:36:37  robertj
+ * More IPv6 support (INADDR_ANY handling), thanks Sbastien Josset
+ *
+ * Revision 1.56  2003/02/03 11:23:32  robertj
+ * Added function to get pointer to IP address data.
+ *
+ * Revision 1.55  2003/02/03 08:51:44  robertj
+ * Fixed compatibility with old code so taking address of PIPSocket::Address
+ *   gets address of 4 or 16 byte IP address.
+ *
+ * Revision 1.54  2002/12/02 03:57:18  robertj
+ * More RTEMS support patches, thank you Vladimir Nesic.
+ *
+ * Revision 1.53  2002/11/02 00:32:21  robertj
+ * Further fixes to VxWorks (Tornado) port, thanks Andreas Sikkema.
+ *
+ * Revision 1.52  2002/10/29 07:59:45  robertj
+ * Changed in_addr6 to more universally used in6_addr.
+ *
+ * Revision 1.51  2002/10/08 14:31:43  robertj
+ * Changed for IPv6 support, thanks Sbastien Josset.
+ *
+ * Revision 1.50  2002/10/08 12:41:51  robertj
+ * Changed for IPv6 support, thanks Sbastien Josset.
+ *
+ * Revision 1.49  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
  * Revision 1.48  2001/12/13 09:17:01  robertj
  * Added function to convert PString to IP address with error checking that can
  *   distinguish between 0.0.0.0 or 255.255.255.255 and illegal address.
@@ -180,7 +218,7 @@
 
 #define _PIPSOCKET
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -189,8 +227,14 @@
 #include <ptlib/socket.h>
 #endif
 
+
+
 /** This class describes a type of socket that will communicate using the
    Internet Protocol.
+   If P_USE_IPV6 is not set, IPv4 only is supported.
+   If P_USE_IPV6 is set, both IPv4 and IPv6 adresses are supported, with 
+   IPv4 as default. This allows to transparently use IPv4, IPv6 or Dual 
+   stack operating systems.
  */
 class PIPSocket : public PSocket
 {
@@ -201,43 +245,56 @@ class PIPSocket : public PSocket
      */
     PIPSocket();
 
-
   public:
     /**
       A class describing an IP address
      */
-    class Address : public in_addr {
+    class Address : public PObject {
       public:
+
         /**@name Address constructors */
         //@{
-        /// Create an IP address with the default address
+        /// Create an IPv4 address with the default address: 127.0.0.1 (loopback)
         Address();
 
-        /// Create an IP address with the default address
+        /** Create an IP address from string notation.
+            eg dot notation x.x.x.x. for IPv4, or colon notation x:x:x::xxx for IPv6
+          */
         Address(const PString & dotNotation);
 
+        /// Create an IPv4 or IPv6 address from 4 or 16 byte values
+        Address(PINDEX len, const BYTE * bytes);
+
         /// Create an IP address from four byte values
         Address(BYTE b1, BYTE b2, BYTE b3, BYTE b4);
 
-        /// Create an IP address from a four byte value in network byte order
+        /// Create an IPv4 address from a four byte value in network byte order
         Address(DWORD dw);
 
-        /// Create an IP address from an in_addr structure
+        /// Create an IPv4 address from an in_addr structure
         Address(const in_addr & addr);
 
-        /// Copy an address from another Address
-        Address(const Address & addr);
+#if P_HAS_IPV6
+        /// Create an IPv6 address from an in_addr structure
+        Address(const in6_addr & addr);
+
+        /// Create an IP (v4 or v6) address from a sockaddr (sockaddr_in,
+        /// sockaddr_in6 or sockaddr_in6_old) structure
+	Address(const int ai_family, const int ai_addrlen,struct sockaddr *ai_addr);
+#endif
 
 #ifdef __NUCLEUS_NET__
         Address(const struct id_struct & addr);
         Address & operator=(const struct id_struct & addr);
 #endif
 
-        /// Copy an address from another Address
+        /// Copy an address from another IP v4 address
         Address & operator=(const in_addr & addr);
 
-        /// Copy an address from another Address
-        Address & operator=(const Address & addr);
+#if P_HAS_IPV6
+        /// Copy an address from another IPv6 address
+        Address & operator=(const in6_addr & addr);
+#endif
 
         /// Copy an address from a string
         Address & operator=(const PString & dotNotation);
@@ -246,6 +303,33 @@ class PIPSocket : public PSocket
         Address & operator=(DWORD dw);
         //@}
 
+        /// Compare two adresses
+        Comparison Compare(const PObject & obj) const;
+        BOOL operator==(const Address & addr) const { return Compare(addr) == EqualTo; }
+        BOOL operator!=(const Address & addr) const { return Compare(addr) != EqualTo; }
+#if P_HAS_IPV6
+        BOOL operator==(in6_addr & addr) const;
+        BOOL operator!=(in6_addr & addr) const { return !operator==(addr); }
+#endif
+        BOOL operator==(in_addr & addr) const;
+        BOOL operator!=(in_addr & addr) const { return !operator==(addr); }
+        BOOL operator==(DWORD dw) const;
+        BOOL operator!=(DWORD dw) const   { return !operator==(dw); }
+#ifdef P_TORNADO 
+        BOOL operator==(long unsigned int u) const { return  operator==((DWORD)u); }
+        BOOL operator!=(long unsigned int u) const { return !operator==((DWORD)u); }
+#endif
+#ifdef _WIN32
+        BOOL operator==(unsigned u) const { return  operator==((DWORD)u); }
+        BOOL operator!=(unsigned u) const { return !operator==((DWORD)u); }
+#endif
+#ifdef P_RTEMS
+        BOOL operator==(u_long u) const { return  operator==((DWORD)u); }
+        BOOL operator!=(u_long u) const { return !operator==((DWORD)u); }
+#endif
+        BOOL operator==(int i) const      { return  operator==((DWORD)i); }
+        BOOL operator!=(int i) const      { return !operator==((DWORD)i); }
+
         /// Format an address as a string
         PString AsString() const;
 
@@ -257,32 +341,80 @@ class PIPSocket : public PSocket
         /// Format an address as a string
         operator PString() const;
 
-        /// Return address in network order
+        /// Return IPv4 address in network order
+        operator in_addr() const;
+
+#if P_HAS_IPV6
+        /// Return IPv4 address in network order
+        operator in6_addr() const;
+#endif
+
+        /// Return IPv4 address in network order
         operator DWORD() const;
 
-        /// Return first byte of IP address
+        /// Return first byte of IPv4 address
         BYTE Byte1() const;
 
-        /// Return second byte of IP address
+        /// Return second byte of IPv4 address
         BYTE Byte2() const;
 
-        /// Return third byte of IP address
+        /// Return third byte of IPv4 address
         BYTE Byte3() const;
 
-        /// Return fourth byte of IP address
+        /// Return fourth byte of IPv4 address
         BYTE Byte4() const;
 
-        /// return specified byte of IP address
+        /// return specified byte of IPv4 or IPv6 address
         BYTE operator[](PINDEX idx) const;
 
-        /// output address as a string to the specified string
-        friend ostream & operator<<(ostream & s, const Address & a);
+        /// Get the address length (will be either 4 or 16)
+        PINDEX GetSize() const;
 
-        /// output address as a string to the specified string
-        friend istream & operator>>(istream & s, Address & a);
+        /// Get the pointer to IP address data
+        const char * GetPointer() const { return (const char *)&v; }
+
+        /// Get the version of the IP address being used
+        unsigned GetVersion() const { return version; }
+
+        /// Check address 0.0.0.0 or :: 
+        BOOL IsValid() const;
+        BOOL IsAny();
+
+        /// Check address 127.0.0.1 or ::1
+        BOOL IsLoopback() const;
+
+        /// Check for Broadcast address 255.255.255.255
+        BOOL IsBroadcast() const;
+
+#if P_HAS_IPV6
+        /// Check for v4 mapped i nv6 address ::ffff:a.b.c.d
+        BOOL IsV4Mapped() const;
+#endif
+        
+        static const Address & GetLoopback();
+#if P_HAS_IPV6
+        static const Address & GetLoopback6();
+        static const Address & GetAny6();
+#endif
+        static const Address & GetBroadcast();
+
+      protected:
+        /// Runtime test of IP addresse type
+        union {
+          in_addr four;
+#if P_HAS_IPV6
+          in6_addr six;
+#endif
+        } v;
+        unsigned version;
+
+      /// output IPv6 & IPv4 address as a string to the specified string
+      friend ostream & operator<<(ostream & s, const Address & a);
+
+      /// input IPv4 (not IPv6 yet!) address as a string from the specified string
+      friend istream & operator>>(istream & s, Address & a);
     };
 
-
   // Overrides from class PChannel
     /** Get the platform and I/O channel type name of the channel. For an IP
        socket this returns the host name of the peer the socket is connected
@@ -293,6 +425,25 @@ class PIPSocket : public PSocket
      */
     virtual PString GetName() const;
 
+    // Set the default IP address familly.
+    // Needed as lot of IPv6 stack are not able to receive IPv4 packets in IPv6 sockets
+    // They are not RFC 2553, chapter 7.3, compliant.
+    // As a concequence, when opening a socket to listen to port 1720 (for exemple) from any remot host
+    // one must decide whether this an IPv4 or an IPv6 socket...
+    static int GetDefaultIpAddressFamily();
+    static void SetDefaultIpAddressFamily(int ipAdressFamily); // PF_INET, PF_INET6
+    static void SetDefaultIpAddressFamilyV4(); // PF_INET
+#if P_HAS_IPV6
+    static void SetDefaultIpAddressFamilyV6(); // PF_INET6
+    static BOOL IsIpAddressFamilyV6Supported();
+#endif
+    static PIPSocket::Address GetDefaultIpAny();
+
+    // Open an IPv4 or IPv6 socket
+    virtual BOOL OpenSocket(
+      int ipAdressFamily=PF_INET
+    ) = 0;
+
 
   // Overrides from class PSocket.
     /** Connect a socket to a remote host on the specified port number. This is
