@@ -27,6 +27,35 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: object.h,v $
+ * Revision 1.86  2002/10/14 21:42:37  rogerh
+ * Only use malloc.h on Windows
+ *
+ * Revision 1.85  2002/10/10 04:43:43  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.84  2002/10/08 12:41:51  robertj
+ * Changed for IPv6 support, thanks Sbastien Josset.
+ *
+ * Revision 1.83  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.82  2002/08/06 02:27:58  robertj
+ * GNU C++ v3 compatibility.
+ *
+ * Revision 1.81  2002/06/25 02:22:47  robertj
+ * Improved assertion system to allow C++ class name to be displayed if
+ *   desired, especially relevant to container classes.
+ *
+ * Revision 1.80  2002/06/14 10:29:43  rogerh
+ * STL + gcc 3.1 compile fix. Submitted by Klaus Kaempf <kkaempf@suse.de>
+ *
+ * Revision 1.79  2002/06/13 08:34:05  rogerh
+ * gcc 3.1 needs iostream instead of iostream.h
+ *
+ * Revision 1.78  2002/05/22 00:23:31  craigs
+ * Added GMTTime flag to tracing options
+ *
  * Revision 1.77  2002/04/19 00:20:51  craigs
  * Added option to append to log file rather than create anew each time
  *
@@ -271,25 +300,43 @@
  *
  */
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
+#if defined(P_VXWORKS)
+#include <private/stdiop.h>
+#endif
+
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#ifdef _WIN32
+#include <malloc.h>
+#endif
 #include <string.h>
 #ifdef __USE_STL__
 #include <string>
 #include <iomanip>
 #include <iostream>
+#if (__GNUC__ >= 3)
+#include <sstream>
+#else
 #include <strstream>
+#endif
+#else
+#if (__GNUC__ >= 3)
+#include <iostream>
+#ifndef __MWERKS__
+#include <iomanip>
+#endif
 #else
 #include <iostream.h>
 #ifndef __MWERKS__
 #include <iomanip.h>
 #endif
 #endif
+#endif
 
 #ifdef _WIN32_WCE
 #include <stdlibx.h>
@@ -297,6 +344,8 @@
 
 #if (__GNUC__ >= 3)
 using namespace std;
+#else
+typedef long _Ios_Fmtflags;
 #endif
 
 ///////////////////////////////////////////////////////////////////////////////
@@ -339,13 +388,24 @@ enum PStandardAssertMessage {
   PMaxStandardAssertMessage
 };
 
+#define __CLASS__ NULL
+
 /** This macro is used to assert that a condition must be TRUE.
 If the condition is FALSE then an assert function is called with the source
 file and line number the macro was instantiated on, plus the message described
 by the #msg# parameter. This parameter may be either a standard value
 from the #PStandardAssertMessage# enum or a literal string.
 */
-#define PAssert(b, m) if(b);else PAssertFunc(__FILE__, __LINE__, (m))
+#define PAssert(b, m) if(b);else PAssertFunc(__FILE__, __LINE__, __CLASS__, (m))
+
+/** This macro is used to assert that a condition must be TRUE.
+If the condition is FALSE then an assert function is called with the source
+file and line number the macro was instantiated on, plus the message described
+by the #msg# parameter. This parameter may be either a standard value
+from the #PStandardAssertMessage# enum or a literal string.
+The #c# parameter specifies the class name that the error occurred in
+*/
+#define PAssert2(b, c, m) if(b);else PAssertFunc(__FILE__, __LINE__, (c), (m))
 
 /** This macro is used to assert that an operating system call succeeds.
 If the condition is FALSE then an assert function is called with the source
@@ -354,7 +414,7 @@ described by the #POperatingSystemError# value in the #PStandardAssertMessage#
 enum.
  */
 #define PAssertOS(b) \
-              if(b);else PAssertFunc(__FILE__, __LINE__, POperatingSystemError)
+              if(b);else PAssertFunc(__FILE__, __LINE__, __CLASS__, POperatingSystemError)
 
 /** This macro is used to assert that a pointer must be non-null.
 If the pointer is NULL then an assert function is called with the source file
@@ -366,7 +426,7 @@ prevent incorrect behaviour with this, the macro will assume that the
 #ptr# parameter is an L-Value.
  */
 #define PAssertNULL(p) ((&(p)&&(p)!=NULL)?(p):(PAssertFunc(__FILE__, \
-                                        __LINE__, PNullPointerReference), (p)))
+                                        __LINE__, __CLASS__, PNullPointerReference), (p)))
 
 /** This macro is used to assert immediately.
 The assert function is called with the source file and line number the macro
@@ -374,11 +434,20 @@ was instantiated on, plus the message described by the #msg# parameter. This
 parameter may be either a standard value from the #PStandardAssertMessage#
 enum or a literal string.
 */
-#define PAssertAlways(m) PAssertFunc(__FILE__, __LINE__, (m))
+#define PAssertAlways(m) PAssertFunc(__FILE__, __LINE__, __CLASS__, (m))
+
+/** This macro is used to assert immediately.
+The assert function is called with the source file and line number the macro
+was instantiated on, plus the message described by the #msg# parameter. This
+parameter may be either a standard value from the #PStandardAssertMessage#
+enum or a literal string.
+*/
+#define PAssertAlways2(c, m) PAssertFunc(__FILE__, __LINE__, (c), (m))
 
 
-void PAssertFunc(const char * file, int line, PStandardAssertMessage msg);
-void PAssertFunc(const char * file, int line, const char * msg);
+void PAssertFunc(const char * file, int line, const char * className, PStandardAssertMessage msg);
+void PAssertFunc(const char * file, int line, const char * className, const char * msg);
+void PAssertFunc(const char * full_msg);
 
 
 /** Get the stream being used for error output.
@@ -453,6 +522,9 @@ public:
     /** SystemLog flag for tracing within a PServiceProcess application. Must
         be set in conjection with SetStream(new PSystemLog).
       */
+    GMTTime = 256,
+    /** Output timestamps in GMT time rather than local time
+      */
     SystemLogStream = 32768
   };
 
