@@ -27,6 +27,30 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: pprocess.h,v $
+ * Revision 1.58  2002/12/11 22:23:59  robertj
+ * Added ability to set user identity temporarily and permanently.
+ * Added get and set users group functions.
+ *
+ * Revision 1.57  2002/12/02 03:57:18  robertj
+ * More RTEMS support patches, thank you Vladimir Nesic.
+ *
+ * Revision 1.56  2002/10/17 13:44:27  robertj
+ * Port to RTEMS, thanks Vladimir Nesic.
+ *
+ * Revision 1.55  2002/10/17 07:17:42  robertj
+ * Added ability to increase maximum file handles on a process.
+ *
+ * Revision 1.54  2002/10/10 04:43:43  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.53  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.52  2002/07/30 02:55:48  craigs
+ * Added program start time to PProcess
+ * Added virtual to GetVersion etc
+ *
  * Revision 1.51  2002/02/14 05:13:33  robertj
  * Fixed possible deadlock if a timer is deleted (however indirectly) in the
  *   OnTimeout of another timer.
@@ -183,7 +207,7 @@
 
 #define _PPROCESS
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -197,12 +221,28 @@
    instance of the application.
  */
 #ifndef P_MAC_MPTHREADS
+#ifdef P_VXWORKS
+#define PCREATE_PROCESS(cls) \
+  PProcess::PreInitialise(0, NULL, NULL); \
+  cls instance; \
+  instance._main();
+#elif defined(P_RTEMS)
+#define PCREATE_PROCESS(cls) \
+extern "C" {\
+   void* POSIX_Init( void* argument) \
+     { PProcess::PreInitialise(0, 0, 0); \
+       static cls instance; \
+       exit( instance._main() ); \
+     } \
+}
+#else
 #define PCREATE_PROCESS(cls) \
   int main(int argc, char ** argv, char ** envp) \
     { PProcess::PreInitialise(argc, argv, envp); \
       static cls instance; \
       return instance._main(); \
     }
+#endif // P_VXWORKS
 #else // Mac MP Tasks implementation
 // Alas, a Macintosh main thread has a number of important responsibilities.
 // The goal of this code is to warm up the Toolbox and create an independant
@@ -435,7 +475,7 @@ class PProcess : public PThread
        @return
        string for the manufacturer name eg "Equivalence".
      */
-    const PString & GetManufacturer() const;
+    virtual const PString & GetManufacturer() const;
 
     /**Get the name of the process. This is used in the
        default "About" dialog box and for determining the location of the
@@ -446,7 +486,7 @@ class PProcess : public PThread
        @return
        string for the process name eg "MyApp".
      */
-    const PString & GetName() const;
+    virtual const PString & GetName() const;
 
     /**Get the version of the software. This is used in the default "About"
        dialog box and for determining the location of the configuration
@@ -462,7 +502,7 @@ class PProcess : public PThread
        @return
        string for the version eg "1.0b3".
      */
-    PString GetVersion(
+    virtual PString GetVersion(
       BOOL full = TRUE /// TRUE for full version, FALSE for short version.
     ) const;
 
@@ -496,17 +536,87 @@ class PProcess : public PThread
        This is a platform dependent string only provided by platforms that are
        multi-user.
 
-       For unix systems the string #uid can be used to specify numeric uids.
+       For unix systems if the username may consist exclusively of digits and
+       there is no actual username consisting of that string then the numeric
+       uid value is used. For example "0" is the superuser. For the rare
+       occassions where the users name is the same as their uid, if the
+       username field starts with a '#' then the numeric form is forced.
 
        If an empty string is provided then original user that executed the
        process in the first place (the real user) is set as the effective user.
 
+       The permanent flag indicates that the user will not be able to simple
+       change back to the original user as indicated above, ie for unix
+       systems setuid() is used instead of seteuid(). This is not necessarily
+       meaningful for all platforms.
+
        @return
        TRUE if processes owner changed. The most common reason for failure is
        that the process does not have the privilege to change the effective user.
       */
     BOOL SetUserName(
-      const PString & username  /// New effective user
+      const PString & username, /// New user name or uid
+      BOOL permanent = FALSE    /// Flag for if effective or real user
+    );
+
+    /**Get the effective group name of the owner of the process, eg "root" etc.
+       This is a platform dependent string only provided by platforms that are
+       multi-user. Note that some value may be returned as a "simulated" user.
+       For example, in MS-DOS an environment variable
+
+       @return
+       group name of processes owner.
+     */
+    PString GetGroupName() const;
+
+    /**Set the effective group of the process.
+       This is a platform dependent string only provided by platforms that are
+       multi-user.
+
+       For unix systems if the groupname may consist exclusively of digits and
+       there is no actual groupname consisting of that string then the numeric
+       uid value is used. For example "0" is the superuser. For the rare
+       occassions where the groups name is the same as their uid, if the
+       groupname field starts with a '#' then the numeric form is forced.
+
+       If an empty string is provided then original group that executed the
+       process in the first place (the real group) is set as the effective
+       group.
+
+       The permanent flag indicates that the group will not be able to simply
+       change back to the original group as indicated above, ie for unix
+       systems setgid() is used instead of setegid(). This is not necessarily
+       meaningful for all platforms.
+
+       @return
+       TRUE if processes group changed. The most common reason for failure is
+       that the process does not have the privilege to change the effective
+       group.
+      */
+    BOOL SetGroupName(
+      const PString & groupname, /// New group name or gid
+      BOOL permanent = FALSE     /// Flag for if effective or real group
+    );
+
+    /**Get the maximum file handle value for the process.
+       For some platforms this is meaningless.
+
+       @return
+       user name of processes owner.
+     */
+    int GetMaxHandles() const;
+
+    /**Set the maximum number of file handles for the process.
+       For unix systems the user must be run with the approriate privileges
+       before this function can set the value above the system limit.
+
+       For some platforms this is meaningless.
+
+       @return
+       TRUE if successfully set the maximum file hadles.
+      */
+    BOOL SetMaxHandles(
+      int newLimit  /// New limit on file handles
     );
 
     /**Get the default file to use in PConfig instances.
@@ -594,6 +704,10 @@ class PProcess : public PThread
     virtual int _main(void * arg = NULL);
     // Main function for process, called from real main after initialisation
 
+    PTime GetStartTime() const;
+    /* return the time at which the program was started 
+    */
+
   protected:
 #if !defined(P_PLATFORM_HAS_THREADS)
 
@@ -647,6 +761,12 @@ class PProcess : public PThread
     PTimerList timers;
     // List of active timers in system
 
+    PTime programStartTime;
+    // time at which process was intantiated, i.e. started
+
+    int maxHandles;
+    // Maximum number of file handles process can open.
+
 #if !defined(P_PLATFORM_HAS_THREADS)
 
     PThread * currentThread;
