@@ -27,6 +27,16 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: thread.h,v $
+ * Revision 1.32  2002/10/04 04:33:27  robertj
+ * Added functions for getting operating system thread identifer values.
+ *
+ * Revision 1.31  2002/09/16 01:08:59  robertj
+ * Added #define so can select if #pragma interface/implementation is used on
+ *   platform basis (eg MacOS) rather than compiler, thanks Robert Monaghan.
+ *
+ * Revision 1.30  2002/06/27 06:44:28  robertj
+ * Changed "" to PString::Empty() where assigning to PString.
+ *
  * Revision 1.29  2002/04/24 01:49:22  robertj
  * Fixed error in PTRACE_BLOCk nesting level to now work when no tracing enabled.
  *
@@ -117,7 +127,7 @@
  */
 
 
-#ifdef __GNUC__
+#ifdef P_USE_PRAGMA
 #pragma interface
 #endif
 
@@ -227,7 +237,7 @@ class PThread : public PObject
       AutoDeleteFlag deletion = AutoDeleteThread,
         /// Automatically delete PThread instance on termination of thread.
       Priority priorityLevel = NormalPriority,  /// Initial priority of thread.
-      const PString & threadName = "" /// The name of the thread (for Debug/Trace)
+      const PString & threadName = PString::Empty() /// The name of the thread (for Debug/Trace)
     );
 
     /** Destroy the thread, this simply calls the #Terminate()# function
@@ -370,6 +380,14 @@ class PThread : public PObject
 
   /**@name Miscellaneous */
   //@{
+    /**Get operating system specific thread identifier for this thread.
+      */
+    virtual PThreadIdentifer GetThreadId() const;
+
+    /**Get operating system specific thread identifier for current thread.
+      */
+    static PThreadIdentifer GetCurrentThreadId();
+
     /** User override function for the main execution routine of the thread. A
        descendent class must provide the code that will be executed in the
        thread within this function.
@@ -412,7 +430,7 @@ class PThread : public PObject
       AutoDeleteFlag deletion = AutoDeleteThread,
         /// Automatically delete PThread instance on termination of thread.
       Priority priorityLevel = NormalPriority,  /// Initial priority of thread.
-      const PString & threadName = "", /// The name of the thread (for Debug/Trace)
+      const PString & threadName = PString::Empty(), /// The name of the thread (for Debug/Trace)
       PINDEX stackSize = 10000         /// Stack size on some platforms
     );
   //@}
