@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,12 +21,92 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * $Log: http.cxx,v $
+ * Revision 1.93  2003/07/22 03:26:10  csoutheren
+ * Fixed problem with parsing default H323 addresses
+ *
+ * Revision 1.92  2003/06/23 15:31:40  ykiryanov
+ * Slightly changed call to ShellExecuteEx to make compatible with Win32
+ *
+ * Revision 1.91  2003/06/23 14:31:33  ykiryanov
+ * Modified for WinCE - used ShellExecuteEx instead of ShellExecute
+ *
+ * Revision 1.90  2003/06/05 00:15:54  rjongbloed
+ * Fixed callto bug created by previous patch.
+ *
+ * Revision 1.89  2003/06/04 01:42:05  rjongbloed
+ * Fixed h323 scheme, does not have a "password" field.
+ *
+ * Revision 1.88  2003/06/02 02:46:45  rjongbloed
+ * Fixed issue with callto URL parsing incorrect username field.
+ * Added automatic removal of illegal (though common) "//" in callto URL.
+ *
+ * Revision 1.87  2003/05/05 07:30:17  craigs
+ * Fixed problem with URLs that do not specify schemes
+ *
+ * Revision 1.86  2003/05/02 13:50:23  craigs
+ * Fixed a problem with callto:localhost
+ *
+ * Revision 1.85  2003/05/02 13:20:33  craigs
+ * Fixed callto problems
+ *
+ * Revision 1.84  2003/04/28 04:41:22  robertj
+ * Changed URL parsing so if a default scheme is present then explicit scheme
+ *   must be "known" to avoid ambiguity with host:port parsing.
+ *
+ * Revision 1.83  2003/04/10 00:13:56  robertj
+ * Fixed correct decoding of user/password/host/port field, for non h323 schemes.
+ *
+ * Revision 1.82  2003/04/08 06:28:14  craigs
+ * Fixed introduced problem with HTTP server mistaking relative URLs for proxy requests
+ *
+ * Revision 1.81  2003/04/04 08:03:55  robertj
+ * Fixed special case of h323 URL default port changing depending on
+ *   if it the host is an endpoint or gatekeeper.
+ *
+ * Revision 1.80  2003/04/04 05:18:08  robertj
+ * Added "callto", "tel" and fixed "h323" URL types.
+ *
+ * Revision 1.79  2002/12/02 00:17:03  robertj
+ * Fixed URL parsing/display problems with non-path URL type eg mailto
+ *
+ * Revision 1.78  2002/11/22 06:16:49  robertj
+ * Fixed usage of URI (relative http/https URL).
+ *
+ * Revision 1.77  2002/11/20 02:10:56  robertj
+ * Fixed some more realtive/absolute path issues.
+ *
+ * Revision 1.76  2002/11/20 01:01:49  robertj
+ * Fixed GNU compatibility
+ *
+ * Revision 1.75  2002/11/20 00:49:37  robertj
+ * Fixed correct interpretation of url re double slashes as per latest RFC,
+ *   including file: mapping and relative paths. Probably still more to do.
+ *
+ * Revision 1.74  2002/11/19 22:45:03  robertj
+ * Fixed support for file: scheme under unix
+ *
+ * Revision 1.73  2002/11/19 10:36:50  robertj
+ * Added functions to set anf get "file:" URL. as PFilePath and do the right
+ *   things with platform dependent directory components.
+ *
+ * Revision 1.72  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
+ * Revision 1.71  2002/09/23 07:17:24  robertj
+ * Changes to allow winsock2 to be included.
+ *
+ * Revision 1.70  2002/08/28 08:06:11  craigs
+ * Fixed problem (again) with file:// URLs
+ *
+ * Revision 1.69  2002/08/28 05:11:23  craigs
+ * Fixed problem with file:// URLs
+ *
+ * Revision 1.68  2002/05/02 05:11:29  craigs
+ * Fixed problem with not translating + chars in URL query parameters
+ *
  * Revision 1.67  2002/03/19 23:39:57  robertj
  * Fixed string output to include PathOnly variant, lost in previous mod.
  *
@@ -262,6 +342,10 @@
 
 #include <ctype.h>
 
+#ifdef WIN32
+#include <shellapi.h>
+#endif
+
 
 // RFC 1738
 // http://host:port/path...
@@ -288,73 +372,98 @@
 #define DEFAULT_RTSPU_PORT	554
 #define DEFAULT_PROSPERO_PORT	1525
 #define	DEFAULT_H323_PORT       1720
+#define	DEFAULT_H323RAS_PORT    1719
 #define	DEFAULT_SIP_PORT        5060
 
-enum SchemeFormat {
-  HostPort = 1,
-  UserPasswordHostPort = 2,
-  HostOnly = 3,
-  Other = 4
+
+struct schemeStruct {
+  const char * name;
+  BOOL hasUsername;
+  BOOL hasPassword;
+  BOOL hasHostPort;
+  BOOL defaultToUserIfNoAt;
+  BOOL defaultHostToLocal;
+  BOOL hasQuery;
+  BOOL hasParameters;
+  BOOL hasFragments;
+  BOOL hasPath;
+  BOOL relativeImpliesScheme;
+  WORD defaultPort;
 };
 
-class schemeStruct {
-  public:
-    const char * name;
-    int  type;
-    BOOL hasDoubleSlash;
-    WORD defaultPort;
+#define DEFAULT_SCHEME 0
+#define FILE_SCHEME    1
+
+static schemeStruct const SchemeTable[] = {
+//  scheme       user   pass   host   @def   defhost query  params frags  path   rel    port
+  { "http",      TRUE,  TRUE,  TRUE,  FALSE, TRUE,   TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  DEFAULT_HTTP_PORT     }, // Must be first
+  { "file",      FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, 0                     }, // Must be second
+  { "https",     FALSE, FALSE, TRUE,  FALSE, TRUE,   TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  DEFAULT_HTTPS_PORT    },
+  { "gopher",    FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_GOPHER_PORT   },
+  { "wais",      FALSE, FALSE, TRUE,  FALSE, FALSE,  FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_WAIS_PORT     },
+  { "nntp",      FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_NNTP_PORT     },
+  { "prospero",  FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_PROSPERO_PORT },
+  { "rtsp",      FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_RTSP_PORT     },
+  { "rtspu",     FALSE, FALSE, TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_RTSPU_PORT    },
+  { "ftp",       TRUE,  TRUE,  TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, TRUE,  FALSE, DEFAULT_FTP_PORT      },
+  { "telnet",    TRUE,  TRUE,  TRUE,  FALSE, TRUE,   FALSE, FALSE, FALSE, FALSE, FALSE, DEFAULT_TELNET_PORT   },
+  { "mailto",    FALSE, FALSE, FALSE, FALSE, TRUE,   TRUE,  FALSE, FALSE, FALSE, FALSE, 0                     },
+  { "news",      FALSE, FALSE, FALSE, FALSE, TRUE,   FALSE, FALSE, FALSE, FALSE, FALSE, 0                     },
+  { "h323",      TRUE,  FALSE, TRUE,  TRUE,  FALSE,  FALSE, TRUE,  FALSE, FALSE, FALSE, DEFAULT_H323_PORT     },
+  { "sip",       TRUE,  TRUE,  TRUE,  FALSE, FALSE,  FALSE, TRUE,  FALSE, FALSE, FALSE, DEFAULT_SIP_PORT      },
+  { "tel",       FALSE, FALSE, FALSE, TRUE,  FALSE,  FALSE, TRUE,  FALSE, FALSE, FALSE, 0                     },
+  { "fax",       FALSE, FALSE, FALSE, TRUE,  FALSE,  FALSE, TRUE,  FALSE, FALSE, FALSE, 0                     },
+  { "callto",    FALSE, FALSE, FALSE, TRUE,  FALSE,  FALSE, TRUE,  FALSE, FALSE, FALSE, 0                     },
+  { NULL,        FALSE, FALSE, FALSE, FALSE, FALSE,  FALSE, FALSE, FALSE, FALSE, FALSE, 0                     }
 };
 
-static schemeStruct const schemeInfo[] = {
-  { "http",      UserPasswordHostPort, TRUE, DEFAULT_HTTP_PORT },
-  { "https",     HostPort, TRUE, DEFAULT_HTTPS_PORT },
-  { "gopher",    HostPort, TRUE, DEFAULT_GOPHER_PORT },
-  { "wais",      HostPort, TRUE, DEFAULT_WAIS_PORT },
-  { "nntp",      HostPort, TRUE, DEFAULT_NNTP_PORT },
-  { "prospero",  HostPort, TRUE, DEFAULT_PROSPERO_PORT },
-  { "rtsp",      HostPort, TRUE, DEFAULT_RTSP_PORT },
-  { "rtspu",     HostPort, TRUE, DEFAULT_RTSPU_PORT },
-
-  { "ftp",       UserPasswordHostPort, TRUE, DEFAULT_FTP_PORT },
-  { "telnet",    UserPasswordHostPort, TRUE, DEFAULT_TELNET_PORT },
-  { "file",      HostOnly,             TRUE },
-  { "mailto",    Other, FALSE},
-  { "news",      Other, FALSE},
-  { "h323",      UserPasswordHostPort, FALSE, DEFAULT_H323_PORT },
-  { "sip",       UserPasswordHostPort, FALSE, DEFAULT_SIP_PORT },
-  { NULL }
-};
-
-static schemeStruct const defaultSchemeInfo = { "other", Other, FALSE};
-
-static const schemeStruct & GetSchemeInfo(const PString & scheme)
+static const schemeStruct * GetSchemeInfo(const PCaselessString & scheme)
 {
   PINDEX i;
-  for (i = 0; schemeInfo[i].name != NULL; i++)
-    if (scheme *= schemeInfo[i].name)
-      return schemeInfo[i];
-  return defaultSchemeInfo;
+  for (i = 0; SchemeTable[i].name != NULL; i++) {
+    if (scheme == SchemeTable[i].name)
+      return &SchemeTable[i];
+  }
+  return NULL;
 }
 
 //////////////////////////////////////////////////////////////////////////////
 // PURL
 
 PURL::PURL()
+  : scheme(SchemeTable[DEFAULT_SCHEME].name),
+    port(0),
+    relativePath(FALSE)
 {
-  scheme = schemeInfo[0].name;
-  port   = 0;
 }
 
 
-PURL::PURL(const char * str)
+PURL::PURL(const char * str, const char * defaultScheme)
 {
-  Parse(str);
+  Parse(str, defaultScheme);
 }
 
 
-PURL::PURL(const PString & str)
+PURL::PURL(const PString & str, const char * defaultScheme)
 {
-  Parse(str);
+  Parse(str, defaultScheme);
+}
+
+
+PURL::PURL(const PFilePath & filePath)
+  : scheme(SchemeTable[FILE_SCHEME].name),
+    port(0),
+    relativePath(FALSE)
+{
+  PStringArray pathArray = filePath.GetDirectory().GetPath();
+  hostname = pathArray[0];
+
+  PINDEX i;
+  for (i = 1; i < pathArray.GetSize(); i++)
+    pathArray[i-1] = pathArray[i];
+  pathArray[i-1] = filePath.GetFileName();
+
+  SetPath(pathArray);
 }
 
 
@@ -389,22 +498,16 @@ PString PURL::TranslateString(const PString & str, TranslationType type)
 {
   PString xlat = str;
 
-  if (type == QueryTranslation) {
-    PINDEX space = (PINDEX)-1;
-    while ((space = xlat.Find(' ', space+1)) != P_MAX_INDEX)
-      xlat[space] = '+';
-  }
-
   PString safeChars = "abcdefghijklmnopqrstuvwxyz"
                       "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-                      "0123456789$-_.+!*'(),";
+                      "0123456789$-_.!*'(),";
   switch (type) {
     case LoginTranslation :
-      safeChars += ";?&=";
+      safeChars += "+;?&=";
       break;
 
     case PathTranslation :
-      safeChars += ":@&=";
+      safeChars += "+:@&=";
       break;
 
     case QueryTranslation :
@@ -414,6 +517,12 @@ PString PURL::TranslateString(const PString & str, TranslationType type)
   while ((pos += 1+strspn(&xlat[pos+1], safeChars)) < xlat.GetLength())
     xlat.Splice(psprintf("%%%02X", (BYTE)xlat[pos]), pos, 1);
 
+  if (type == QueryTranslation) {
+    PINDEX space = (PINDEX)-1;
+    while ((space = xlat.Find(' ', space+1)) != P_MAX_INDEX)
+      xlat[space] = '+';
+  }
+
   return xlat;
 }
 
@@ -479,13 +588,14 @@ void PURL::SplitQueryVars(const PString & queryStr, PStringToString & queryVars)
 }
 
 
-void PURL::Parse(const char * cstr)
+void PURL::Parse(const char * cstr, const char * defaultScheme)
 {
   hostname = PCaselessString();
   pathStr = username = password = fragment = PString::Empty();
   path.SetSize(0);
   queryVars.RemoveAll();
   port = 0;
+  relativePath = FALSE;
 
   // copy the string so we can take bits off it
   while (isspace(*cstr))
@@ -494,118 +604,252 @@ void PURL::Parse(const char * cstr)
 
   PINDEX pos;
 
-  static const PString reservedChars = "=;/#?";
-
+  // get information which tells us how to parse URL for this
+  // particular scheme
+  const schemeStruct * schemeInfo = NULL;
+    
   // determine if the URL has a scheme
-  scheme = "";
-  if (isalpha(url[0])) {
-    for (pos = 0; url[pos] != '\0' &&
-                          reservedChars.Find(url[pos]) == P_MAX_INDEX; pos++) {
-      if (url[pos] == ':') {
-        scheme = url.Left(pos);
-        url.Delete(0, pos+1);
-        break;
-      }
-    }
+  scheme = PString::Empty();
+
+  // Character set as per RFC2396
+  pos = 0;
+  while (isalnum(url[pos]) || url[pos] == '+' || url[pos] == '-' || url[pos] == '.')
+    pos++;
+
+  // Have explicit scheme
+  if (url[pos] == ':') {
+    schemeInfo = GetSchemeInfo(url.Left(pos));
+    if (schemeInfo == NULL && defaultScheme == NULL)
+      schemeInfo = &SchemeTable[PARRAYSIZE(SchemeTable)-1];
+    if (schemeInfo != NULL)
+      url.Delete(0, pos+1);
   }
 
-  // if there is no scheme, then default to http for the local
-  // on the default port
-  if (scheme.IsEmpty()) {
-    scheme = schemeInfo[0].name;
-    if (url.GetLength() > 2 && url[0] == '/' && url[1] == '/') 
+  // if we could not match a scheme, then use the default scheme
+  if (schemeInfo == NULL && defaultScheme != NULL)
+    schemeInfo = GetSchemeInfo(defaultScheme);
+
+  if (schemeInfo == NULL)
+    schemeInfo = &SchemeTable[DEFAULT_SCHEME];
+
+  scheme = schemeInfo->name;
+
+  // Super special case!
+  if (scheme *= "callto") {
+    // Actually not part of MS spec, but a lot of people put in the // into
+    // the URL, so we take it out of it is there.
+    if (url.GetLength() > 2 && url[0] == '/' && url[1] == '/')
       url.Delete(0, 2);
-  } else {
 
-    // get information which tells us how to parse URL for this
-    // particular scheme
-    const schemeStruct & schemeInfo = GetSchemeInfo(scheme);
-    
-    // if the URL should have leading slash, then remove it if it has one
-    if (schemeInfo.hasDoubleSlash &&
-      url.GetLength() > 2 && url[0] == '/' && url[1] == '/') 
-    url.Delete(0, 2);
+    // For some bizarre reason callto uses + instead of ; for paramters
+    // We do a loop so that phone numbers of the form +61243654666 still work
+    do {
+      pos = url.Find('+');
+    } while (pos != P_MAX_INDEX && isdigit(url[pos+1]));
 
-    // if the rest of the URL isn't a path, then we are finished!
-    if (schemeInfo.type == Other) {
-      pathStr = url;
-      return;
+    if (pos != P_MAX_INDEX) {
+      SplitVars(url(pos+1, P_MAX_INDEX), paramVars, '+', '=');
+      url.Delete(pos, P_MAX_INDEX);
     }
 
-    // parse user/password/host/port
-    if (schemeInfo.type == HostPort ||
-        schemeInfo.type == UserPasswordHostPort ||
-        schemeInfo.type == HostOnly) {
-      pos = url.FindOneOf("/;?#");
-      PString uphp = url.Left(pos);
-      if (pos != P_MAX_INDEX)
-        url.Delete(0, pos);
-      else
-        url = PString();
+    hostname = paramVars("gateway");
+    if (!hostname)
+      username = UntranslateString(url, LoginTranslation);
+    else {
+      PCaselessString type = paramVars("type");
+      if (type == "directory") {
+        pos = url.Find('/');
+        if (pos == P_MAX_INDEX)
+          username = UntranslateString(url, LoginTranslation);
+        else {
+          hostname = UntranslateString(url.Left(pos), LoginTranslation);
+          username = UntranslateString(url.Mid(pos+1), LoginTranslation);
+        }
+      }
+      else {
+        // Now look for an @ and split user and host
+        pos = url.Find('@');
+        if (pos != P_MAX_INDEX) {
+          username = UntranslateString(url.Left(pos), LoginTranslation);
+          hostname = UntranslateString(url.Mid(pos+1), LoginTranslation);
+        }
+        else {
+          if (type == "ip" || type == "host")
+            hostname = UntranslateString(url, LoginTranslation);
+          else
+            username = UntranslateString(url, LoginTranslation);
+        }
+      }
+    }
 
-      // if the URL is of type HostOnly, then this is the hostname
-      if (schemeInfo.type == HostOnly) {
-        hostname = UntranslateString(uphp, LoginTranslation);
-      } 
+    // Allow for [ipv6] form
+    pos = hostname.Find(']');
+    if (pos == P_MAX_INDEX)
+      pos = 0;
+    pos = hostname.Find(':', pos);
+    if (pos != P_MAX_INDEX) {
+      port = (WORD)hostname.Mid(pos+1).AsUnsigned();
+      hostname.Delete(pos, P_MAX_INDEX);
+    }
 
-      // if the URL is of type UserPasswordHostPort, then parse it
-      if (schemeInfo.type == UserPasswordHostPort) {
+    password = paramVars("password");
+    return;
+  }
 
-        // extract username and password
-        PINDEX pos2 = uphp.Find('@');
-        if (pos2 != P_MAX_INDEX && pos2 > 0) {
-          PINDEX pos3 = uphp.Find(":");
-          // if no password...
+  // if the URL should have leading slash, then remove it if it has one
+  if (schemeInfo != NULL && schemeInfo->hasHostPort && schemeInfo->hasPath) {
+    if (url.GetLength() > 2 && url[0] == '/' && url[1] == '/')
+      url.Delete(0, 2);
+    else
+      relativePath = TRUE;
+  }
+
+  // parse user/password/host/port
+  if (!relativePath && schemeInfo->hasHostPort) {
+    PString endHostChars;
+    if (schemeInfo->hasPath)
+      endHostChars += '/';
+    if (schemeInfo->hasQuery)
+      endHostChars += '?';
+    if (schemeInfo->hasParameters)
+      endHostChars += ';';
+    if (schemeInfo->hasFragments)
+      endHostChars += '#';
+    if (endHostChars.IsEmpty())
+      pos = P_MAX_INDEX;
+    else
+      pos = url.FindOneOf(endHostChars);
+
+    PString uphp = url.Left(pos);
+    if (pos != P_MAX_INDEX)
+      url.Delete(0, pos);
+    else
+      url = PString::Empty();
+
+    // if the URL is of type UserPasswordHostPort, then parse it
+    if (schemeInfo->hasUsername) {
+      // extract username and password
+      PINDEX pos2 = uphp.Find('@');
+      PINDEX pos3 = P_MAX_INDEX;
+      if (schemeInfo->hasPassword)
+        pos3 = uphp.Find(':');
+      switch (pos2) {
+        case 0 :
+          uphp.Delete(0, 1);
+          break;
+
+        case P_MAX_INDEX :
+          if (schemeInfo->defaultToUserIfNoAt) {
+            if (pos3 == P_MAX_INDEX)
+              username = UntranslateString(uphp, LoginTranslation);
+            else {
+              username = UntranslateString(uphp.Left(pos3), LoginTranslation);
+              password = UntranslateString(uphp.Mid(pos3+1), LoginTranslation);
+            }
+            uphp = PString::Empty();
+          }
+          break;
+
+        default :
           if (pos3 > pos2)
-            username = UntranslateString(uphp(0, pos2-1), LoginTranslation);
+            username = UntranslateString(uphp.Left(pos2), LoginTranslation);
           else {
-            username = UntranslateString(uphp(0, pos3-1), LoginTranslation);
+            username = UntranslateString(uphp.Left(pos3), LoginTranslation);
             password = UntranslateString(uphp(pos3+1, pos2-1), LoginTranslation);
           }
           uphp.Delete(0, pos2+1);
-        }
+      }
+    }
+
+    // if the URL does not have a port, then this is the hostname
+    if (schemeInfo->defaultPort == 0)
+      hostname = UntranslateString(uphp, LoginTranslation);
+    else {
+      // determine if the URL has a port number
+      // Allow for [ipv6] form
+      pos = uphp.Find(']');
+      if (pos == P_MAX_INDEX)
+        pos = 0;
+      pos = uphp.Find(':', pos);
+      if (pos == P_MAX_INDEX)
+        hostname = UntranslateString(uphp, LoginTranslation);
+      else {
+        hostname = UntranslateString(uphp.Left(pos), LoginTranslation);
+        port = (WORD)uphp.Mid(pos+1).AsUnsigned();
       }
 
-      // determine if the URL has a port number
-      if (schemeInfo.type == HostPort ||
-          schemeInfo.type == UserPasswordHostPort) {
-        pos = uphp.Find(":");
-        if (pos == P_MAX_INDEX) {
-          hostname = UntranslateString(uphp, LoginTranslation);
-          port = schemeInfo.defaultPort;
-        } else {
-          hostname = UntranslateString(uphp.Left(pos), LoginTranslation);
-          port = (WORD)uphp(pos+1, P_MAX_INDEX).AsInteger();
-        }
-        if (hostname.IsEmpty())
-          hostname = PIPSocket::GetHostName();
-      }
+      if (hostname.IsEmpty() && schemeInfo->defaultHostToLocal)
+        hostname = PIPSocket::GetHostName();
     }
   }
 
-  // chop off any trailing query
-  pos = url.Find('?');
-  if (pos != P_MAX_INDEX /* && pos > 0 */) {
-    SplitQueryVars(url(pos+1, P_MAX_INDEX), queryVars);
-    url.Delete(pos, P_MAX_INDEX);
+  if (schemeInfo->hasQuery) {
+    // chop off any trailing query
+    pos = url.Find('?');
+    if (pos != P_MAX_INDEX) {
+      SplitQueryVars(url(pos+1, P_MAX_INDEX), queryVars);
+      url.Delete(pos, P_MAX_INDEX);
+    }
   }
 
-  // chop off any trailing parameters
-  pos = url.Find(';');
-  if (pos != P_MAX_INDEX /* && pos > 0 */) {
-    SplitVars(url(pos+1, P_MAX_INDEX), paramVars, ';', '=');
-    url.Delete(pos, P_MAX_INDEX);
+  if (schemeInfo->hasParameters) {
+    // chop off any trailing parameters
+    pos = url.Find(';');
+    if (pos != P_MAX_INDEX) {
+      SplitVars(url(pos+1, P_MAX_INDEX), paramVars, ';', '=');
+      url.Delete(pos, P_MAX_INDEX);
+    }
   }
 
-  // chop off any trailing fragment
-  pos = url.Find('#');
-  if (pos != P_MAX_INDEX /* && pos > 0 */) {
-    fragment = UntranslateString(url(pos+1, P_MAX_INDEX), PathTranslation);
-    url.Delete(pos, P_MAX_INDEX);
+  if (schemeInfo->hasFragments) {
+    // chop off any trailing fragment
+    pos = url.Find('#');
+    if (pos != P_MAX_INDEX) {
+      fragment = UntranslateString(url(pos+1, P_MAX_INDEX), PathTranslation);
+      url.Delete(pos, P_MAX_INDEX);
+    }
   }
 
-  // the hierarchy is what is left
-  SetPathStr(url);
+  if (schemeInfo->hasPath)
+    SetPathStr(url);   // the hierarchy is what is left
+  else {
+  // if the rest of the URL isn't a path, then we are finished!
+    pathStr = UntranslateString(url, PathTranslation);
+    Recalculate();
+  }
+
+  if (port == 0 && schemeInfo->defaultPort != 0 && !relativePath) {
+    // Yes another horrible, horrible special case!
+    if (scheme == "h323" && paramVars("type") == "gk")
+      port = DEFAULT_H323RAS_PORT;
+    else
+      port = schemeInfo->defaultPort;
+  }
+}
+
+
+PFilePath PURL::AsFilePath() const
+{
+  if (scheme != SchemeTable[FILE_SCHEME].name)
+    return PString::Empty();
+
+  PStringStream str;
+
+  if (relativePath) {
+    for (PINDEX i = 0; i < path.GetSize(); i++) {
+      if (i > 0)
+        str << PDIR_SEPARATOR;
+      str << path[i];
+    }
+  }
+  else {
+    if (hostname != "localhost")
+      str << hostname;
+    for (PINDEX i = 0; i < path.GetSize(); i++)
+      str << PDIR_SEPARATOR << path[i];
+  }
+
+  return str;
 }
 
 
@@ -617,54 +861,55 @@ PString PURL::AsString(UrlFormat fmt) const
   if (fmt == FullURL)
     return urlString;
 
+  if (scheme.IsEmpty())
+    return PString::Empty();
+
+  const schemeStruct * schemeInfo = GetSchemeInfo(scheme);
+  if (schemeInfo == NULL)
+    schemeInfo = &SchemeTable[PARRAYSIZE(SchemeTable)-1];
+
   if (fmt == HostPortOnly) {
-    if (scheme.IsEmpty())
-      return PString::Empty();
-
     str << scheme << ':';
-    const schemeStruct & schemeInfo = GetSchemeInfo(scheme);
 
-    if (schemeInfo.hasDoubleSlash)
+    if (relativePath) {
+      if (schemeInfo->relativeImpliesScheme)
+        return PString::Empty();
+      return str;
+    }
+
+    if (schemeInfo->hasPath && schemeInfo->hasHostPort)
       str << "//";
 
-    if (schemeInfo.type == Other) 
-      str << pathStr;
-    else {
-      if (schemeInfo.type == HostOnly)
-        str << hostname;
-
-      if (schemeInfo.type == UserPasswordHostPort) {
-        if (!username) {
-          str << TranslateString(username, LoginTranslation);
-          if (!password)
-            str << ':' << TranslateString(password, LoginTranslation);
-          str << '@';
-        }
-      }
-
-      if (schemeInfo.type == HostPort || schemeInfo.type == UserPasswordHostPort) {
-        if (hostname.IsEmpty())
-          str = PString();
-        else {
-          str << hostname;
-          if (port != schemeInfo.defaultPort)
-            str << ':' << port;
-        }
+    if (schemeInfo->hasUsername) {
+      if (!username) {
+        str << TranslateString(username, LoginTranslation);
+        if (schemeInfo->hasPassword && !password)
+          str << ':' << TranslateString(password, LoginTranslation);
+        str << '@';
       }
     }
+
+    if (schemeInfo->hasHostPort)
+      str << hostname;
+
+    if (schemeInfo->defaultPort != 0) {
+      if (port != schemeInfo->defaultPort)
+        str << ':' << port;
+    }
+
     return str;
   }
 
   // URIOnly and PathOnly
-  if (!path.IsEmpty()) {
-    PINDEX count = path.GetSize();
-    str << '/';
-    for (i = 0; i < count; i++) {
-      str << TranslateString(path[i], PathTranslation);
-      if (i < count-1)
+  if (schemeInfo->hasPath) {
+    for (i = 0; i < path.GetSize(); i++) {
+      if (i > 0 || !relativePath)
         str << '/';
+      str << TranslateString(path[i], PathTranslation);
     }
   }
+  else
+    str << TranslateString(pathStr, PathTranslation);
 
   if (fmt == URIOnly) {
     if (!fragment)
@@ -837,11 +1082,22 @@ void PURL::SetQueryVar(const PString & key, const PString & data)
 BOOL PURL::OpenBrowser(const PString & url)
 {
 #ifdef WIN32
-  if ((int)ShellExecute(NULL, "open", url, NULL, NULL, 0) > 32)
+  SHELLEXECUTEINFO sei;
+  ZeroMemory(&sei, sizeof(SHELLEXECUTEINFO));
+  sei.cbSize = sizeof(SHELLEXECUTEINFO);
+  sei.lpVerb = TEXT("open");
+
+  if ((int) ShellExecuteEx(&sei) > 32)
     return TRUE;
 
+#ifndef _WIN32_WCE
   MessageBox(NULL, "Unable to open page"&url, PProcess::Current().GetName(), MB_TASKMODAL);
-#endif
+#else
+  USES_CONVERSION;
+  MessageBox(NULL, _T("Unable to open page"), A2T(PProcess::Current().GetName()), MB_APPLMODAL);
+#endif // _WIN32_WCE
+
+#endif // WIN32
   return FALSE;
 }
 
@@ -849,7 +1105,7 @@ BOOL PURL::OpenBrowser(const PString & url)
 void PURL::Recalculate()
 {
   if (scheme.IsEmpty())
-    scheme = schemeInfo[0].name;
+    scheme = SchemeTable[DEFAULT_SCHEME].name;
 
   urlString = AsString(HostPortOnly) + AsString(URIOnly);
 }
