@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,12 +21,31 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * $Log: httpsrvr.cxx,v $
+ * Revision 1.46  2003/03/19 01:55:26  robertj
+ * Fixed bugs in deleteing HTTP resources from server, thanks Diego Trtara
+ *
+ * Revision 1.45  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
+ * Revision 1.44  2002/10/10 04:43:44  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.43  2002/10/02 08:54:01  craigs
+ * Added support for XMLRPC server
+ *
+ * Revision 1.42  2002/08/27 23:49:08  robertj
+ * Fixed security hole where possible to get any file on disk when using
+ *   PHTTPDirectory HTTP resource.
+ *
+ * Revision 1.41  2002/07/17 08:43:52  robertj
+ * Fixed closing of html msg on generated post output.
+ *
+ * Revision 1.40  2002/05/08 05:38:54  robertj
+ * Added PHTTPTailFile resource to do a unix 'tail -f' of a file.
+ *
  * Revision 1.39  2002/04/12 08:15:23  robertj
  * Fixed warning on older GNU compilers, also guarantees numeric output.
  *
@@ -272,18 +291,22 @@ BOOL PHTTPSpace::DelResource(const PURL & url)
 
     node = &node->children[pos];
 
-    if (node->resource != NULL)
+    // If have resource and not last node, then trying to remove something
+    // further down the tree than a leaf node.
+    if (node->resource != NULL && i < (path.GetSize()-1))
       return FALSE;
   }
 
   if (!node->children.IsEmpty())
     return FALSE;   // Still a resource in tree further down path.
 
-  do {
-    Node * par = node->parent;
-    par->children.Remove(node);
-    node = par;
-  } while (node != NULL && node->children.IsEmpty());
+  if (node->parent != NULL) {
+    do {
+      Node * par = node->parent;
+      par->children.Remove(node);
+      node = par;
+    } while (node != NULL && node->children.IsEmpty());
+  }
 
   return TRUE;
 }
@@ -595,7 +618,7 @@ BOOL PHTTPServer::ProcessCommand()
       (!url.GetHostName() && !PIPSocket::IsLocalHost(url.GetHostName())))
     persist = OnProxy(connectInfo);
   else {
-    PString entityBody = ReadEntityBody();
+    connectInfo.entityBody = ReadEntityBody();
 
     // Handle the local request
     PStringToString postData;
@@ -613,9 +636,9 @@ BOOL PHTTPServer::ProcessCommand()
           // check for multi-part form POSTs
           PString postType = (connectInfo.GetMIME())(ContentTypeTag);
           if (postType.Find("multipart/form-data") == 0)
-            connectInfo.DecodeMultipartFormInfo(postType, entityBody);
+            connectInfo.DecodeMultipartFormInfo(postType, connectInfo.entityBody);
           else  // if (postType *= "x-www-form-urlencoded)
-            PURL::SplitQueryVars(entityBody, postData);
+            PURL::SplitQueryVars(connectInfo.entityBody, postData);
         }
         persist = OnPOST(url, connectInfo.GetMIME(), postData, connectInfo);
         break;
@@ -763,7 +786,7 @@ static const httpStatusCodeStruct * GetStatusCodeStruct(int code)
   static const httpStatusCodeStruct httpStatusDefn[] = {
     // First entry MUST be InternalServerError
     { "Internal Server Error",         PHTTP::InternalServerError, 1 },
-    { "OK",                            PHTTP::OK, 1 },
+    { "OK",                            PHTTP::RequestOK, 1 },
     { "Unauthorised",                  PHTTP::UnAuthorised, 1 },
     { "Forbidden",                     PHTTP::Forbidden, 1 },
     { "Not Found",                     PHTTP::NotFound, 1 },
@@ -910,7 +933,7 @@ BOOL PHTTPServer::OnError(StatusCode code,
 
   if (!statusInfo->allowedBody) {
     StartResponse(code, headers, 0);
-    return statusInfo->code == OK;
+    return statusInfo->code == RequestOK;
   }
 
   PString reply;
@@ -936,7 +959,7 @@ BOOL PHTTPServer::OnError(StatusCode code,
   headers.SetAt(ContentTypeTag, "text/html");
   StartResponse(code, headers, reply.GetLength());
   WriteString(reply);
-  return statusInfo->code == OK;
+  return statusInfo->code == RequestOK;
 }
 
 
@@ -1076,7 +1099,7 @@ PHTTPRequest::PHTTPRequest(const PURL & _url,
     localAddr(0),
     localPort(0)
 {
-  code        = PHTTP::OK;
+  code        = PHTTP::RequestOK;
   contentSize = P_MAX_INDEX;
 
   PIPSocket * socket = server.GetSocket();
@@ -1273,9 +1296,11 @@ BOOL PHTTPResource::OnGETOrHEAD(PHTTPServer & server,
             const PHTTPConnectionInfo & connectInfo,
                                    BOOL isGET)
 {
-  if (isGET && info.Contains(PHTTP::IfModifiedSinceTag) &&
-                           !IsModifiedSince(PTime(info[PHTTP::IfModifiedSinceTag]))) 
-    return server.OnError(PHTTP::NotModified, url.AsString(), connectInfo);
+  // Nede to split songle if into 2 so the Tornado compiler won't end with
+  // 'internal compiler error'
+  if (isGET && info.Contains(PHTTP::IfModifiedSinceTag))
+    if (!IsModifiedSince(PTime(info[PHTTP::IfModifiedSinceTag])))
+      return server.OnError(PHTTP::NotModified, url.AsString(), connectInfo);
 
   PHTTPRequest * request = CreateRequest(url,
                                          info,
@@ -1329,11 +1354,13 @@ BOOL PHTTPResource::OnPOST(PHTTPServer & server,
                                          connectInfo.GetMultipartFormInfo(),
                                          server);
 
+  request->entityBody = connectInfo.GetEntityBody();
+
   BOOL persist = TRUE;
   if (CheckAuthority(server, *request, connectInfo)) {
     server.SetDefaultMIMEInfo(request->outMIME, connectInfo);
     persist = OnPOSTData(*request, data);
-    if (request->code != PHTTP::OK)
+    if (request->code != PHTTP::RequestOK)
       persist = server.OnError(request->code, "", connectInfo) && persist;
   }
 
@@ -1347,17 +1374,17 @@ BOOL PHTTPResource::OnPOSTData(PHTTPRequest & request,
 {
   PHTML msg;
   BOOL persist = Post(request, data, msg);
-  if (request.code != PHTTP::OK)
+
+  if (msg.Is(PHTML::InBody))
+    msg << PHTML::Body();
+
+  if (request.code != PHTTP::RequestOK)
     return persist;
 
   if (msg.IsEmpty())
-    msg << PHTML::Title()    << (unsigned)PHTTP::OK << " OK" << PHTML::Body()
-        << PHTML::Heading(1) << (unsigned)PHTTP::OK << " OK" << PHTML::Heading(1)
+    msg << PHTML::Title()    << (unsigned)PHTTP::RequestOK << " OK" << PHTML::Body()
+        << PHTML::Heading(1) << (unsigned)PHTTP::RequestOK << " OK" << PHTML::Heading(1)
         << PHTML::Body();
-  else {
-    if (msg.Is(PHTML::InBody))
-      msg << PHTML::Body();
-  }
 
   request.outMIME.SetAt(PHTTP::ContentTypeTag, "text/html");
 
@@ -1721,6 +1748,82 @@ PString PHTTPFile::LoadText(PHTTPRequest & request)
 }
 
 
+//////////////////////////////////////////////////////////////////////////////
+// PHTTPTailFile
+
+PHTTPTailFile::PHTTPTailFile(const PString & filename)
+  : PHTTPFile(filename)
+{
+}
+
+
+PHTTPTailFile::PHTTPTailFile(const PString & filename,
+                             const PHTTPAuthority & auth)
+  : PHTTPFile(filename, auth)
+{
+}
+
+
+PHTTPTailFile::PHTTPTailFile(const PURL & url,
+                             const PFilePath & file)
+  : PHTTPFile(url, file)
+{
+}
+
+
+PHTTPTailFile::PHTTPTailFile(const PURL & url,
+                             const PFilePath & file,
+                             const PString & contentType)
+  : PHTTPFile(url, file, contentType)
+{
+}
+
+
+PHTTPTailFile::PHTTPTailFile(const PURL & url,
+                             const PFilePath & file,
+                             const PHTTPAuthority & auth)
+  : PHTTPFile(url, file, auth)
+{
+}
+
+
+PHTTPTailFile::PHTTPTailFile(const PURL & url,
+                             const PFilePath & file,
+                             const PString & contentType,
+                             const PHTTPAuthority & auth)
+  : PHTTPFile(url, file, contentType, auth)
+{
+}
+
+
+BOOL PHTTPTailFile::LoadHeaders(PHTTPRequest & request)
+{
+  if (!PHTTPFile::LoadHeaders(request))
+    return FALSE;
+
+  request.contentSize = P_MAX_INDEX;
+  return TRUE;
+}
+
+
+BOOL PHTTPTailFile::LoadData(PHTTPRequest & request, PCharArray & data)
+{
+  PFile & file = ((PHTTPFileRequest&)request).file;
+
+  if (file.GetPosition() == 0)
+    file.SetPosition(file.GetLength()-request.url.GetQueryVars()("offset", "10000").AsUnsigned());
+
+  while (file.GetPosition() >= file.GetLength()) {
+    if (!request.server.Write(NULL, 0))
+      return FALSE;
+    PThread::Sleep(200);
+  }
+
+  PINDEX count = file.GetLength() - file.GetPosition();
+  return file.Read(data.GetPointer(count), count);
+}
+
+
 //////////////////////////////////////////////////////////////////////////////
 // PHTTPDirectory
 
@@ -1752,7 +1855,22 @@ PHTTPRequest * PHTTPDirectory::CreateRequest(const PURL & url,
                           const PMultipartFormInfoArray & multipartFormInfo,
 			                    PHTTPServer & socket)
 {
-  return new PHTTPDirRequest(url, inMIME, multipartFormInfo, socket);
+  PHTTPDirRequest * request = new PHTTPDirRequest(url, inMIME, multipartFormInfo, socket);
+
+  const PStringArray & path = url.GetPath();
+  request->realPath = basePath;
+  PINDEX i;
+  for (i = GetURL().GetPath().GetSize(); i < path.GetSize()-1; i++)
+    request->realPath += path[i] + PDIR_SEPARATOR;
+
+  // append the last path element
+  if (i < path.GetSize())
+    request->realPath += path[i];
+
+  if (request->realPath.Find(basePath) != 0)
+    request->realPath = basePath;
+
+  return request;
 }
 
 
@@ -1792,24 +1910,11 @@ BOOL PHTTPDirectory::CheckAuthority(PHTTPServer & server,
                              const PHTTPRequest & request,
                       const PHTTPConnectionInfo & conInfo)
 {
-  PFilePath & realPath = ((PHTTPDirRequest&)request).realPath;
-
-  // construct the real path name
-  const PStringArray & path = request.url.GetPath();
-  realPath = basePath;
-  PINDEX i;
-  for (i = GetURL().GetPath().GetSize(); i < path.GetSize()-1; i++)
-    realPath += path[i] + PDIR_SEPARATOR;
-
-  // append the last path element
-  if (i < path.GetSize())
-    realPath += path[i];
-
   // if access control is enabled, then search parent directories for password files
   PStringToString authorisations;
   PString newRealm;
   if (authorisationRealm.IsEmpty() ||
-      !FindAuthorisations(realPath.GetDirectory(), newRealm, authorisations) ||
+      !FindAuthorisations(((PHTTPDirRequest&)request).realPath.GetDirectory(), newRealm, authorisations) ||
       authorisations.GetSize() == 0)
     return TRUE;
 
