@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,12 +21,51 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * $Log: httpsvc.cxx,v $
+ * Revision 1.89  2003/02/19 07:23:45  robertj
+ * Changes to allow for single threaded HTTP service processes.
+ *
+ * Revision 1.88  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
+ * Revision 1.87  2002/10/10 04:43:44  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.86  2002/08/14 00:43:40  robertj
+ * Added ability to have fixed maximum length PStringStream's so does not do
+ *   unwanted malloc()'s while outputing data.
+ *
+ * Revision 1.85  2002/08/13 05:39:17  robertj
+ * Fixed GNU compatibility
+ *
+ * Revision 1.84  2002/08/13 01:57:15  robertj
+ * Fixed "last dump object" position in Memory Dump macro.
+ *
+ * Revision 1.83  2002/08/13 01:30:27  robertj
+ * Added UpTime macro for time service has been running.
+ * Added IfQuery macro blcok to add chunks of HTML depending on the value
+ *   of query parameters in the URL.
+ * Added memory statistics dump and memory object dump macros to help in
+ *   leak finding.
+ *
+ * Revision 1.82  2002/07/30 08:37:34  robertj
+ * Removed peer host as bad DNS makes it useless due to huge timeout.
+ *
+ * Revision 1.81  2002/07/30 04:51:26  craigs
+ * Added MonitorInfo macro
+ *
+ * Revision 1.80  2002/07/30 03:16:57  craigs
+ * Added StartTime macro
+ *
+ * Revision 1.79  2002/07/17 09:18:00  robertj
+ * made detection of gif file more intelligent for debug version.
+ *
+ * Revision 1.78  2002/07/17 08:03:45  robertj
+ * Allowed for adjustable copyright holder.
+ * Allowed for not having gif file for product name in default header.
+ *
  * Revision 1.77  2001/10/10 08:06:49  robertj
  * Fixed problem with not shutting down threads when closing listener.
  *
@@ -314,22 +353,30 @@ PHTTPServiceProcess::PHTTPServiceProcess(const Info & inf)
     compilationDate(inf.compilationDate),
     manufacturersHomePage(inf.manufHomePage != NULL ? inf.manufHomePage : HOME_PAGE),
     manufacturersEmail(inf.email != NULL ? inf.email : EMAIL),
-    productNameHTML(inf.productHTML != NULL ? inf.productHTML : inf.productName)
+    productNameHTML(inf.productHTML != NULL ? inf.productHTML : inf.productName),
+    gifHTML(inf.gifHTML),
+    copyrightHolder(inf.copyrightHolder != NULL ? inf.copyrightHolder : EQUIVALENCE),
+    copyrightHomePage(inf.copyrightHomePage != NULL ? inf.copyrightHomePage : HOME_PAGE),
+    copyrightEmail(inf.copyrightEmail != NULL ? inf.copyrightEmail : EMAIL)
 {
   ignoreSignatures = FALSE;
 
-  if (inf.gifHTML != NULL)
-    gifHTML = inf.gifHTML;
-  else {
-    gifHTML = psprintf("<img src=\"/%s\" alt=\"%s!\"", inf.gifFilename, inf.productName);
-    if (inf.gifWidth != 0 && inf.gifHeight != 0)
-      gifHTML += psprintf(" width=%i height=%i", inf.gifWidth, inf.gifHeight);
-    gifHTML += " align=absmiddle>";
+  if (inf.gifFilename != NULL) {
+    PDirectory exeDir = GetFile().GetDirectory();
+#if defined(_WIN32) && defined(_DEBUG)
+    // Special check to aid in using DevStudio for debugging.
+    if (exeDir.Find("\\Debug\\") != P_MAX_INDEX)
+      exeDir = exeDir.GetParent();
+#endif
+    httpNameSpace.AddResource(new PServiceHTTPFile(inf.gifFilename, exeDir+inf.gifFilename));
+    if (gifHTML.IsEmpty()) {
+      gifHTML = psprintf("<img src=\"/%s\" alt=\"%s!\"", inf.gifFilename, inf.productName);
+      if (inf.gifWidth != 0 && inf.gifHeight != 0)
+        gifHTML += psprintf(" width=%i height=%i", inf.gifWidth, inf.gifHeight);
+      gifHTML += " align=absmiddle>";
+    }
   }
 
-  if (inf.gifFilename != NULL)
-    httpNameSpace.AddResource(new PServiceHTTPFile(inf.gifFilename, GetFile().GetDirectory()+inf.gifFilename));
-
   restartThread = NULL;
   httpListeningSocket = NULL;
   httpThreads.DisallowDeleteObjects();
@@ -421,7 +468,7 @@ BOOL PHTTPServiceProcess::ListenForHTTP(PSocket * listener,
   }
 
   if (stackSize > 1000)
-    new PHTTPServiceThread(stackSize, *this, *httpListeningSocket);
+    new PHTTPServiceThread(stackSize, *this);
 
   return TRUE;
 }
@@ -462,12 +509,12 @@ PString PHTTPServiceProcess::GetCopyrightText()
   PHTML html(PHTML::InBody);
   html << "Copyright &copy;"
        << compilationDate.AsString("yyyy") << " by "
-       << PHTML::HotLink(HOME_PAGE)
-       << EQUIVALENCE
+       << PHTML::HotLink(copyrightHomePage)
+       << copyrightHolder
        << PHTML::HotLink()
        << ", "
-       << PHTML::HotLink("mailto:" EMAIL)
-       << EMAIL
+       << PHTML::HotLink("mailto:" + copyrightEmail)
+       << copyrightEmail
        << PHTML::HotLink();
   return html;
 }
@@ -482,17 +529,24 @@ PString PHTTPServiceProcess::GetPageGraphic()
   PHTML html(PHTML::InBody);
   html << PHTML::TableStart()
        << PHTML::TableRow()
-       << PHTML::TableData()
-       << gifHTML
-       << PHTML::TableData()
+       << PHTML::TableData();
+
+  if (gifHTML.IsEmpty())
+    html << PHTML::Heading(1) << productNameHTML << "&nbsp;" << PHTML::Heading(1);
+  else
+    html << gifHTML;
+
+  html << PHTML::TableData()
        << GetOSClass() << ' ' << GetOSName()
-       << " Version " << GetVersion(TRUE)
+       << " Version " << GetVersion(TRUE) << PHTML::BreakLine()
+       << ' ' << GetCompilationDate().AsString("d MMMM yyyy")
        << PHTML::BreakLine()
        << "By "
        << PHTML::HotLink(manufacturersHomePage) << GetManufacturer() << PHTML::HotLink()
        << ", "
        << PHTML::HotLink("mailto:" + manufacturersEmail) << manufacturersEmail << PHTML::HotLink()
-       << PHTML::TableEnd();
+       << PHTML::TableEnd()
+       << PHTML::HRule();
 
   return html;
 }
@@ -512,6 +566,56 @@ void PHTTPServiceProcess::GetPageHeader(PHTML & html, const PString & title)
 }
 
 
+PTCPSocket * PHTTPServiceProcess::AcceptHTTP()
+{
+  if (httpListeningSocket == NULL)
+    return NULL;
+
+  if (!httpListeningSocket->IsOpen())
+    return NULL;
+
+  // get a socket when a client connects
+  PTCPSocket * socket = new PTCPSocket;
+  if (socket->Accept(*httpListeningSocket))
+    return socket;
+
+  if (socket->GetErrorCode() != PChannel::Interrupted)
+    PSYSTEMLOG(Error, "Accept failed for HTTP: " << socket->GetErrorText());
+
+  if (httpListeningSocket != NULL && httpListeningSocket->IsOpen())
+    return socket;
+
+  delete socket;
+  return NULL;
+}
+
+
+BOOL PHTTPServiceProcess::ProcessHTTP(PTCPSocket & socket)
+{
+  if (!socket.IsOpen())
+    return TRUE;
+
+  PHTTPServer * server = CreateHTTPServer(socket);
+  if (server == NULL) {
+    PSYSTEMLOG(Error, "HTTP server creation/open failed.");
+    return TRUE;
+  }
+
+  // process requests
+  while (server->ProcessCommand())
+    ;
+
+  // always close after the response has been sent
+  delete server;
+
+  // if a restart was requested, then do it, but only if we are not shutting down
+  if (httpListeningSocket->IsOpen())
+    CompleteRestartSystem();
+
+  return TRUE;
+}
+
+
 void PHTTPServiceProcess::BeginRestartSystem()
 {
   if (restartThread == NULL) {
@@ -573,25 +677,26 @@ PHTTPServer * PHTTPServiceProcess::CreateHTTPServer(PTCPSocket & socket)
   return NULL;
 }
 
+
 PHTTPServer * PHTTPServiceProcess::OnCreateHTTPServer(const PHTTPSpace & httpNameSpace)
 {
   return new PHTTPServer(httpNameSpace);
 }
 
+
 //////////////////////////////////////////////////////////////
 
 PHTTPServiceThread::PHTTPServiceThread(PINDEX stackSize,
-                                       PHTTPServiceProcess & app,
-                                       PSocket & listeningSocket)
+                                       PHTTPServiceProcess & app)
   : PThread(stackSize, AutoDeleteThread, NormalPriority, "HTTP Service:%x"),
-    process(app),
-    listener(listeningSocket)
+    process(app)
 {
   process.httpThreadsMutex.Wait();
   process.httpThreads.Append(this);
   process.httpThreadsMutex.Signal();
 
   myStackSize = stackSize;
+  socket = NULL;
   Resume();
 }
 
@@ -601,41 +706,24 @@ PHTTPServiceThread::~PHTTPServiceThread()
   process.httpThreadsMutex.Wait();
   process.httpThreads.Remove(this);
   process.httpThreadsMutex.Signal();
+  delete socket;
+}
+
+
+void PHTTPServiceThread::Close()
+{
+  if (socket != NULL)
+    socket->Close();
 }
 
 
 void PHTTPServiceThread::Main()
 {
-  if (!listener.IsOpen())
-    return;
-
-  // get a socket when a client connects
-  if (!socket.Accept(listener)) {
-    if (socket.GetErrorCode() != PChannel::Interrupted)
-      PSYSTEMLOG(Error, "Accept failed for HTTP: " << socket.GetErrorText());
-    if (listener.IsOpen())
-      new PHTTPServiceThread(myStackSize, process, listener);
-    return;
+  PTCPSocket * socket = process.AcceptHTTP();
+  if (socket != NULL) {
+    new PHTTPServiceThread(myStackSize, process);
+    process.ProcessHTTP(*socket);
   }
-
-  new PHTTPServiceThread(myStackSize, process, listener);
-
-  PHTTPServer * server = process.CreateHTTPServer(socket);
-  if (server == NULL) {
-    PSYSTEMLOG(Error, "HTTP server creation/open failed.");
-    return;
-  }
-
-  // process requests
-  while (server->ProcessCommand())
-    ;
-
-  // always close after the response has been sent
-  delete server;
-
-  // if a restart was requested, then do it, but only if we are not shutting down
-  if (listener.IsOpen())
-    process.CompleteRestartSystem();
 }
 
 
@@ -688,7 +776,7 @@ BOOL PConfigPage::Post(PHTTPRequest & request,
   PSYSTEMLOG(Debug3, "Post to " << request.url << '\n' << data);
   BOOL retval = PHTTPConfig::Post(request, data, reply);
 
-  if (request.code == PHTTP::OK)
+  if (request.code == PHTTP::RequestOK)
     process.BeginRestartSystem();
 
   PServiceHTML::ProcessMacros(request, reply,
@@ -751,7 +839,7 @@ BOOL PConfigSectionsPage::Post(PHTTPRequest & request,
                                PHTML & reply)
 {
   BOOL retval = PHTTPConfigSectionList::Post(request, data, reply);
-  if (request.code == PHTTP::OK)
+  if (request.code == PHTTP::RequestOK)
     process.BeginRestartSystem();
   return retval;
 }
@@ -1032,7 +1120,7 @@ BOOL PRegisterPage::Post(PHTTPRequest & request,
     LoadText(request);
 
   BOOL retval = PHTTPConfig::Post(request, data, reply);
-  if (request.code != PHTTP::OK)
+  if (request.code != PHTTP::RequestOK)
     return FALSE;
 
   PSecureConfig sconf(process.GetProductKey(), process.GetSecuredKeys());
@@ -1137,7 +1225,7 @@ PString PServiceHTML::ExtractSignature(const PString & html,
     return tag(tag.Find("signature")+10, tag.FindLast('-')-2).Trim();
   }
 
-  return PString();
+  return PString::Empty();
 }
 
 
@@ -1289,7 +1377,7 @@ PObject::Comparison PServiceMacro::Compare(const PObject & obj) const
 
 PString PServiceMacro::Translate(PHTTPRequest &, const PString &, const PString &) const
 {
-  return PString();
+  return PString::Empty();
 };
 
 
@@ -1420,6 +1508,19 @@ PCREATE_SERVICE_MACRO(Time,P_EMPTY,args)
 }
 
 
+PCREATE_SERVICE_MACRO(StartTime,P_EMPTY,P_EMPTY)
+{
+  return PProcess::Current().GetStartTime().AsString(PTime::MediumDateTime);
+}
+
+
+PCREATE_SERVICE_MACRO(UpTime,P_EMPTY,P_EMPTY)
+{
+  PTimeInterval upTime = PTime() - PProcess::Current().GetStartTime();
+  return upTime.AsString(0, PTimeInterval::IncludeDays);
+}
+
+
 PCREATE_SERVICE_MACRO(LocalHost,request,P_EMPTY)
 {
   if (request.localAddr != 0)
@@ -1464,6 +1565,47 @@ PCREATE_SERVICE_MACRO(PeerIP,request,P_EMPTY)
     return "N/A";
 }
 
+PCREATE_SERVICE_MACRO(MonitorInfo,request,P_EMPTY)
+{
+  const PTime & compilationDate = PHTTPServiceProcess::Current().GetCompilationDate();
+
+  PString peerAddr = "N/A";
+  if (request.origin != 0)
+    peerAddr = request.origin.AsString();
+
+  PString localAddr = "127.0.0.1";
+  if (request.localAddr != 0)
+    localAddr = request.localAddr.AsString();
+
+  WORD localPort = 80;
+  if (request.localPort != 0)
+    localPort = request.localPort;
+
+  PString timeFormat = "yyyyMMdd hhmmss z";
+
+  PTime now;
+  PTimeInterval upTime = now - PProcess::Current().GetStartTime();
+
+  PStringStream monitorText; 
+  monitorText << "Program: "          << PHTTPServiceProcess::Current().GetProductName() << "\n"
+              << "Version: "          << PHTTPServiceProcess::Current().GetVersion(TRUE) << "\n"
+              << "Manufacturer: "     << PHTTPServiceProcess::Current().GetManufacturer() << "\n"
+              << "OS: "               << PHTTPServiceProcess::Current().GetOSClass() << " " << PHTTPServiceProcess::Current().GetOSName() << "\n"
+              << "OS Version: "       << PHTTPServiceProcess::Current().GetOSVersion() << "\n"
+              << "Hardware: "         << PHTTPServiceProcess::Current().GetOSHardware() << "\n"
+              << "Compilation date: " << compilationDate.AsString(timeFormat, PTime::GMT) << "\n"
+              << "Start Date: "       << PProcess::Current().GetStartTime().AsString(timeFormat, PTime::GMT) << "\n"
+              << "Current Date: "     << now.AsString(timeFormat, PTime::GMT) << "\n"
+              << "Up time: "          << upTime << "\n"
+              << "Peer Addr: "        << peerAddr << "\n"
+              << "Local Host: "       << PIPSocket::GetHostName() << "\n"
+              << "Local Addr: "       << localAddr << "\n"
+              << "Local Port: "       << localPort << "\n"
+       ;
+
+  return monitorText;
+}
+
 
 PCREATE_SERVICE_MACRO(RegInfo,P_EMPTY,P_EMPTY)
 {
@@ -1563,7 +1705,7 @@ PCREATE_SERVICE_MACRO(Query,request,args)
     if (!value)
       return value;
   }
-  return PString();
+  return PString::Empty();
 }
 
 
@@ -1582,7 +1724,7 @@ PCREATE_SERVICE_MACRO(Get,request,args)
       return config.GetString(variable, value);
     }
   }
-  return PString();
+  return PString::Empty();
 }
 
 
@@ -1634,12 +1776,84 @@ PCREATE_SERVICE_MACRO(SignedInclude,P_EMPTY,args)
 }
 
 
+#if PMEMORY_CHECK
+
+PCREATE_SERVICE_MACRO(HeapStatistics,P_EMPTY,P_EMPTY)
+{
+  BOOL oldIgnoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE);
+  PStringStream str(5000);
+  PMemoryHeap::SetIgnoreAllocations(oldIgnoreAllocations);
+  PMemoryHeap::DumpStatistics(str);
+  return str;
+}
+
+
+PCREATE_SERVICE_MACRO(HeapDump,request,P_EMPTY)
+{
+  static DWORD lastObjectNumber = 0;
+  DWORD objectNumber = lastObjectNumber;
+  PStringToString vars = request.url.GetQueryVars();
+  if (vars.Contains("object"))
+    objectNumber = vars["object"].AsUnsigned();
+
+  BOOL oldIgnoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE);
+  PStringStream str(32000);
+  str << "Dumping objects from #" << objectNumber << '\n';
+  PMemoryHeap::SetIgnoreAllocations(oldIgnoreAllocations);
+  PMemoryHeap::DumpObjectsSince(objectNumber, str);
+  lastObjectNumber = PMemoryHeap::GetAllocationRequest();
+  return str;
+}
+
+#endif
+
+
+PCREATE_SERVICE_MACRO_BLOCK(IfQuery,request,args,block)
+{
+  PStringToString vars = request.url.GetQueryVars();
+
+  PINDEX space = args.FindOneOf(" \t\r\n");
+  PString var = args.Left(space);
+  PString value = args.Mid(space).LeftTrim();
+
+  BOOL ok;
+  if (value.IsEmpty())
+    ok = vars.Contains(var);
+  else {
+    PString operation;
+    space = value.FindOneOf(" \t\r\n");
+    if (space != P_MAX_INDEX) {
+      operation = value.Left(space);
+      value = value.Mid(space).LeftTrim();
+    }
+
+    PString query = vars(var);
+    if (operation == "!=")
+      ok = query != value;
+    else if (operation == "<")
+      ok = query < value;
+    else if (operation == ">")
+      ok = query > value;
+    else if (operation == "<=")
+      ok = query <= value;
+    else if (operation == ">=")
+      ok = query >= value;
+    else if (operation == "*=")
+      ok = (query *= value);
+    else
+      ok = query == value;
+  }
+
+  return ok ? block : PString::Empty();
+}
+
+
 PCREATE_SERVICE_MACRO_BLOCK(IfInURL,request,args,block)
 {
   if (request.url.AsString().Find(args) != P_MAX_INDEX)
     return block;
 
-  return PString();
+  return PString::Empty();
 }
 
 
@@ -1648,7 +1862,7 @@ PCREATE_SERVICE_MACRO_BLOCK(IfNotInURL,request,args,block)
   if (request.url.AsString().Find(args) == P_MAX_INDEX)
     return block;
 
-  return PString();
+  return PString::Empty();
 }
 
 
@@ -1659,7 +1873,7 @@ static void SplitCmdAndArgs(const PString & text, PINDEX pos, PCaselessString &
   PINDEX endCmd = macro.FindOneOf(whitespace);
   if (endCmd == P_MAX_INDEX) {
     cmd = macro;
-    args = PString();
+    args = PString::Empty();
   }
   else {
     cmd = macro.Left(endCmd);
@@ -1765,7 +1979,7 @@ BOOL PServiceHTML::ProcessMacros(PHTTPRequest & request,
       if (!process.SubstituteEquivalSequence(request, cmd & args, substitution)) {
         PINDEX idx = ServiceMacros.GetValuesIndex(PServiceMacro(cmd, FALSE));
         if (idx != P_MAX_INDEX) {
-          substitution = ServiceMacros[idx].Translate(request, args, PString());
+          substitution = ServiceMacros[idx].Translate(request, args, PString::Empty());
           substitedMacro = TRUE;
         }
       }
