@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,12 +21,18 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * $Log: inetmail.cxx,v $
+ * Revision 1.24  2003/02/20 00:16:06  craigs
+ * Changed MIME_Version to MIME-Version
+ *
+ * Revision 1.23  2002/12/19 01:35:24  robertj
+ * Fixed problem with returning incorrect lastWriteLength on translated output.
+ *
+ * Revision 1.22  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
  * Revision 1.21  2002/01/07 05:26:47  robertj
  * Fixed getting scan list of messages, thanks xradish
  *
@@ -1114,7 +1120,7 @@ void PPOP3Server::HandleDeleteMessage(PINDEX, const PString &)
 //////////////////////////////////////////////////////////////////////////////
 // PRFC822Channel
 
-const char PRFC822Channel::MimeVersionTag[] = "MIME_version";
+const char PRFC822Channel::MimeVersionTag[] = "MIME-version";
 const char PRFC822Channel::FromTag[] = "From";
 const char PRFC822Channel::ToTag[] = "To";
 const char PRFC822Channel::CCTag[] = "cc";
@@ -1200,12 +1206,21 @@ BOOL PRFC822Channel::Write(const void * buf, PINDEX len)
     writePartHeaders = FALSE;
   }
 
+  BOOL ok;
   if (base64 == NULL)
-    return PIndirectChannel::Write(buf, len);
+    ok = PIndirectChannel::Write(buf, len);
+  else {
+    base64->ProcessEncoding(buf, len);
+    PString str = base64->GetEncodedString();
+    ok = PIndirectChannel::Write(str.GetPointer(), str.GetLength());
+  }
 
-  base64->ProcessEncoding(buf, len);
-  PString str = base64->GetEncodedString();
-  return PIndirectChannel::Write(str.GetPointer(), str.GetLength());
+  // Always return the lastWriteCount as the number of bytes expected to be
+  // written, not teh actual number which with base64 encoding etc may be
+  // significantly more.
+  if (ok)
+    lastWriteCount = len;
+  return ok;
 }
 
 
