@@ -5,7 +5,7 @@
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 1993-2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,16 +21,22 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
  * Portions bsed upon the file crypto/buffer/bss_sock.c 
  * Original copyright notice appears below
  *
- * $Id: pssl.cxx,v 1.33 2002/04/02 16:59:35 robertj Exp $
+ * $Id: pssl.cxx,v 1.36 2003/04/16 08:00:19 robertj Exp $
  * $Log: pssl.cxx,v $
+ * Revision 1.36  2003/04/16 08:00:19  robertj
+ * Windoes psuedo autoconf support
+ *
+ * Revision 1.35  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
+ * Revision 1.34  2002/06/07 02:55:23  robertj
+ * Fixed GNU warning
+ *
  * Revision 1.33  2002/04/02 16:59:35  robertj
  * Fixed GNU warning
  *
@@ -191,12 +197,19 @@ extern "C" {
 
 #include <openssl/ssl.h>
 #include <openssl/err.h>
-#include <openssl/crypto.h>
-#include <openssl/buffer.h>
 #include <openssl/rand.h>
 
 };
 
+
+#ifdef _MSC_VER
+#pragma comment(lib, P_SSL_LIB1)
+#pragma comment(lib, P_SSL_LIB2)
+#endif
+
+
+///////////////////////////////////////////////////////////////////////////////
+
 PARRAY(PSSLMutexArrayBase, PMutex);
 class PSSLMutexArray : public PSSLMutexArrayBase
 {
@@ -802,8 +815,6 @@ static void PSSLAssert(const char * msg)
 
 PSSLContext::PSSLContext(const void * sessionId, PINDEX idSize)
 {
-  PINDEX i;
-
   static PMutex InitialisationMutex;
   InitialisationMutex.Wait();
 
@@ -814,7 +825,7 @@ PSSLContext::PSSLContext(const void * sessionId, PINDEX idSize)
 
     // Seed the random number generator
     BYTE seed[128];
-    for (i = 0; i < sizeof(seed); i++)
+    for (size_t i = 0; i < sizeof(seed); i++)
       seed[i] = (BYTE)rand();
     RAND_seed(seed, sizeof(seed));
 
