@@ -1,11 +1,11 @@
 /*
  * pxml.cxx
  *
- * XML parser support library
+ * XML parser support
  *
  * Portable Windows Library
  *
- * Copyright (c) 1993-1998 Equivalence Pty. Ltd.
+ * Copyright (c) 2002 Equivalence Pty. Ltd.
  *
  * The contents of this file are subject to the Mozilla Public License
  * Version 1.0 (the "License"); you may not use this file except in
@@ -21,50 +21,103 @@
  *
  * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
  *
- * Portions are Copyright (C) 1993 Free Software Foundation, Inc.
- * All Rights Reserved.
- *
  * Contributor(s): ______________________________________.
  *
+ * $Log: pxml.cxx,v $
+ * Revision 1.35  2003/05/14 02:50:53  rjongbloed
+ * Simplified name space initialisation
+ *
+ * Revision 1.34  2003/05/06 07:55:25  craigs
+ * Fixed problem with initialising XML parser for namespaces
+ *
+ * Revision 1.33  2003/04/27 23:53:30  craigs
+ * Removed deprecated options
+ *
+ * Revision 1.32  2003/04/16 08:00:19  robertj
+ * Windoes psuedo autoconf support
+ *
+ * Revision 1.31  2003/04/08 12:47:07  craigs
+ * Fixed problem with handling of CDATA
+ *
+ * Revision 1.30  2003/04/08 05:08:41  craigs
+ * Fixed problems with additional spaces being included with metasequences
+ *
+ * Revision 1.29  2003/04/02 09:13:55  rogerh
+ * Add type casts because the variable 'expat' is now a void *
+ *
+ * Revision 1.28  2003/03/31 06:20:56  craigs
+ * Split the expat wrapper from the XML file handling to allow reuse of the parser
+ *
+ * Revision 1.27  2003/01/13 02:14:02  robertj
+ * Improved error logging for auto-loaded XML
+ *
+ * Revision 1.26  2002/12/16 06:38:59  robertj
+ * Added ability to specify certain elemets (by name) that are exempt from
+ *   the indent formatting. Useful for XML/RPC where leading white space is
+ *   not ignored by all servers.
+ *
+ * Revision 1.25  2002/12/10 04:41:16  robertj
+ * Added test for URL being empty, don't try and run auto load in background.
+ *
+ * Revision 1.24  2002/11/26 05:53:45  craigs
+ * Added ability to auto-reload from URL
+ *
+ * Revision 1.23  2002/11/21 08:08:52  craigs
+ * Changed to not overwrite XML data if load fails
+ *
+ * Revision 1.22  2002/11/19 07:37:25  craigs
+ * Added locking functions and LoadURL function
+ *
+ * Revision 1.21  2002/11/06 22:47:25  robertj
+ * Fixed header comment (copyright etc)
+ *
  */
 
 // This depends on the expat XML library by Jim Clark
 // See http://www.jclark.com/xml/expat.html for more information
 
+#include <ptlib.h>
+
 #ifdef __GNUC__
 #pragma implementation "pxml.h"
 #endif
 
-#include <ptlib.h>
+#include <ptclib/pxml.h>
 
 #if P_EXPAT
 
-#include <ptclib/pxml.h>
 #include <expat.h>
 
-#define	XMLSETTINGS_OPTIONS	(NewLineAfterElement | CloseExtended)
+
+#define CACHE_BUFFER_SIZE   1024
+#define	XMLSETTINGS_OPTIONS	(NewLineAfterElement)
+
+
+#ifdef _MSC_VER
+#pragma comment(lib, P_EXPAT_LIBRARY)
+#endif
 
 
 ////////////////////////////////////////////////////
-//
+
 static void PXML_StartElement(void * userData, const char * name, const char ** attrs)
 {
-  ((PXML *)userData)->StartElement(name, attrs);
+  ((PXMLParser *)userData)->StartElement(name, attrs);
 }
 
 static void PXML_EndElement(void * userData, const char * name)
 {
-  ((PXML *)userData)->EndElement(name);
+  ((PXMLParser *)userData)->EndElement(name);
 }
 
 static void PXML_CharacterDataHandler(void * userData, const char * data, int len)
 {
-  ((PXML *)userData)->AddCharacterData(data, len);
+  ((PXMLParser *)userData)->AddCharacterData(data, len);
 }
 
 static void PXML_XmlDeclHandler(void * userData, const char * version, const char * encoding, int standalone)
 {
-  ((PXML *)userData)->XmlDecl(version, encoding, standalone);
+  ((PXMLParser *)userData)->XmlDecl(version, encoding, standalone);
 }
 
 static void PXML_StartDocTypeDecl(void * userData,
@@ -73,64 +126,227 @@ static void PXML_StartDocTypeDecl(void * userData,
 		            const char * pubid,
 			             int hasInternalSubSet)
 {
-  ((PXML *)userData)->StartDocTypeDecl(docTypeName, sysid, pubid, hasInternalSubSet);
+  ((PXMLParser *)userData)->StartDocTypeDecl(docTypeName, sysid, pubid, hasInternalSubSet);
 }
 
 static void PXML_EndDocTypeDecl(void * userData)
 {
-  ((PXML *)userData)->EndDocTypeDecl();
+  ((PXMLParser *)userData)->EndDocTypeDecl();
 }
 
-////////////////////////////////////////////////////
-
-PXML::PXML(int _options)
+static void PXML_StartNamespaceDeclHandler(void *userData,
+                                 const XML_Char *prefix,
+                                 const XML_Char *uri)
 {
-  Construct();
-  if (_options >= 0)
-    options = _options;
+  ((PXMLParser *)userData)->StartNamespaceDeclHandler(prefix, uri);
 }
 
-PXML::PXML(const PString & data, int _options)
+static void PXML_EndNamespaceDeclHandler(void *userData, const XML_Char *prefix)
 {
-  Construct();
-  if (_options >= 0)
-    options = _options;
+  ((PXMLParser *)userData)->EndNamespaceDeclHandler(prefix);
+}
+
+PXMLParser::PXMLParser(int _options)
+  : options(_options)
+{
+  if (options < 0)
+    options = 0;
+
+  if ((options & WithNS) != 0)
+    expat = XML_ParserCreateNS(NULL, '|');
+  else
+    expat = XML_ParserCreate(NULL);
+
+  XML_SetUserData((XML_Parser)expat, this);
+
+  XML_SetElementHandler      ((XML_Parser)expat, PXML_StartElement, PXML_EndElement);
+  XML_SetCharacterDataHandler((XML_Parser)expat, PXML_CharacterDataHandler);
+  XML_SetXmlDeclHandler      ((XML_Parser)expat, PXML_XmlDeclHandler);
+  XML_SetDoctypeDeclHandler  ((XML_Parser)expat, PXML_StartDocTypeDecl, PXML_EndDocTypeDecl);
+  XML_SetNamespaceDeclHandler((XML_Parser)expat, PXML_StartNamespaceDeclHandler, PXML_EndNamespaceDeclHandler);
+
+  rootElement = NULL;
+  currentElement = NULL;
+  lastElement    = NULL;
+}
+
+PXMLParser::~PXMLParser()
+{
+  XML_ParserFree((XML_Parser)expat);
+}
+
+PXMLElement * PXMLParser::GetXMLTree() const
+{ 
+  return rootElement; 
+}
+
+PXMLElement * PXMLParser::SetXMLTree(PXMLElement * newRoot)
+{ 
+  PXMLElement * oldRoot = rootElement;
+  rootElement = newRoot; 
+  return oldRoot;
+}
+
+BOOL PXMLParser::Parse(const char * data, int dataLen, BOOL final)
+{
+  return XML_Parse((XML_Parser)expat, data, dataLen, final) != 0;  
+}
+
+void PXMLParser::GetErrorInfo(PString & errorString, PINDEX & errorCol, PINDEX & errorLine)
+{
+  XML_Error err = XML_GetErrorCode((XML_Parser)expat);
+  errorString = PString(XML_ErrorString(err));
+  errorCol    = XML_GetCurrentColumnNumber((XML_Parser)expat);
+  errorLine   = XML_GetCurrentLineNumber((XML_Parser)expat);
+}
+
+void PXMLParser::StartElement(const char * name, const char **attrs)
+{
+  PXMLElement * newElement = new PXMLElement(currentElement, name);
+  if (currentElement != NULL)
+    currentElement->AddSubObject(newElement, FALSE);
+
+  while (attrs[0] != NULL) {
+    newElement->SetAttribute(PString(attrs[0]), PString(attrs[1]));
+    attrs += 2;
+  }
+
+  currentElement = newElement;
+  lastElement    = NULL;
+
+  if (rootElement == NULL)
+    rootElement = currentElement;
+}
+
+void PXMLParser::EndElement(const char * /*name*/)
+{
+  currentElement = currentElement->GetParent();
+  lastElement    = NULL;
+}
+
+void PXMLParser::AddCharacterData(const char * data, int len)
+{
+  PString str(data, len);
+
+  if (lastElement != NULL) {
+    PAssert(!lastElement->IsElement(), "lastElement set by non-data element");
+    lastElement->SetString(lastElement->GetString() + str, FALSE);
+  } else {
+    PXMLData * newElement = new PXMLData(currentElement, str);
+    if (currentElement != NULL)
+      currentElement->AddSubObject(newElement, FALSE);
+    lastElement = newElement;
+  } 
+}
+
+
+void PXMLParser::XmlDecl(const char * _version, const char * _encoding, int _standAlone)
+{
+  version    = _version;
+  encoding   = _encoding;
+  standAlone = _standAlone;
+}
+
+void PXMLParser::StartDocTypeDecl(const char * /*docTypeName*/,
+		                              const char * /*sysid*/,
+				                          const char * /*pubid*/,
+				                          int /*hasInternalSubSet*/)
+{
+}
+
+void PXMLParser::EndDocTypeDecl()
+{
+}
+
+void PXMLParser::StartNamespaceDeclHandler(const XML_Char * /*prefix*/, 
+                                           const XML_Char * /*uri*/)
+{
+}
+
+void PXMLParser::EndNamespaceDeclHandler(const XML_Char * /*prefix*/)
+{
+}
+
+
+///////////////////////////////////////////////////////////////////////////////////////////////
+
+
+PXML::PXML(int options, const char * noIndentElements)
+  : PXMLBase(options) 
+{
+  Construct(options, noIndentElements);
+}
+
+PXML::PXML(const PString & data, int options, const char * noIndentElements)
+  : PXMLBase(options) 
+{
+  Construct(options, noIndentElements);
   Load(data);
 }
 
 PXML::~PXML()
 {
-  delete rootElement;
+  autoLoadTimer.Stop();
+  RemoveAll();
 }
 
 PXML::PXML(const PXML & xml)
+  : noIndentElements(xml.noIndentElements)
 {
-  Construct();
+  Construct(xml.options, NULL);
 
-  options      = xml.options;
   loadFromFile = xml.loadFromFile;
   loadFilename = xml.loadFilename;
   version      = xml.version;
   encoding     = xml.encoding;
   standAlone   = xml.standAlone;
 
+  PWaitAndSignal m(xml.rootMutex);
+
   PXMLElement * oldRootElement = xml.rootElement;
   if (oldRootElement != NULL)
     rootElement = (PXMLElement *)oldRootElement->Clone(NULL);
 }
 
-void PXML::Construct()
+void PXML::Construct(int _options, const char * _noIndentElements)
 {
   rootElement    = NULL;
-  currentElement = NULL;
-  lastElement    = NULL;
-  options        = 0;
+  options        = _options > 0 ? _options : 0;
   loadFromFile   = FALSE;
   standAlone     = -2;
+
+  if (_noIndentElements != NULL)
+    noIndentElements = PString(_noIndentElements).Tokenise(' ', FALSE);
+}
+
+PXMLElement * PXML::SetRootElement(const PString & documentType)
+{
+  PWaitAndSignal m(rootMutex);
+
+  if (rootElement != NULL)
+    delete rootElement;
+
+  rootElement = new PXMLElement(rootElement, documentType);
+
+  return rootElement;
+}
+
+PXMLElement * PXML::SetRootElement(PXMLElement * element)
+{
+  PWaitAndSignal m(rootMutex);
+
+  if (rootElement != NULL)
+    delete rootElement;
+
+  rootElement = element;
+
+  return rootElement;
 }
 
 BOOL PXML::IsDirty() const
 {
+  PWaitAndSignal m(rootMutex);
+
   if (rootElement == NULL)
     return FALSE;
 
@@ -139,6 +355,8 @@ BOOL PXML::IsDirty() const
 
 PCaselessString PXML::GetDocumentType() const
 { 
+  PWaitAndSignal m(rootMutex);
+
   if (rootElement == NULL)
     return PCaselessString();
   return rootElement->GetName();
@@ -146,6 +364,10 @@ PCaselessString PXML::GetDocumentType() const
 
 BOOL PXML::LoadFile(const PFilePath & fn, int _options)
 {
+  PTRACE(4, "XML\tLoading file " << fn);
+
+  PWaitAndSignal m(rootMutex);
+
   if (_options >= 0)
     options = _options;
 
@@ -153,42 +375,187 @@ BOOL PXML::LoadFile(const PFilePath & fn, int _options)
   loadFromFile = TRUE;
 
   PFile file;
-  if (!file.Open(fn, PFile::ReadOnly)) 
+  if (!file.Open(fn, PFile::ReadOnly)) {
+    errorString = "File open error" & file.GetErrorText();
     return FALSE;
+  }
 
   off_t len = file.GetLength();
   PString data;
-  if (!file.Read(data.GetPointer(len + 1), len))
+  if (!file.Read(data.GetPointer(len + 1), len)) {
+    errorString = "File read error" & file.GetErrorText();
     return FALSE;
+  }
 
   data[(PINDEX)len] = '\0';
 
   return Load(data);
 }
 
+BOOL PXML::LoadURL(const PURL & url)
+{
+  return LoadURL(url, PMaxTimeInterval, -1);
+}
+
+
+BOOL PXML::LoadURL(const PURL & url, const PTimeInterval & timeout, int _options)
+{
+  if (url.IsEmpty()) {
+    errorString = "Cannot load empty URL";
+    errorCol = errorLine = 0;
+    return FALSE;
+  }
+
+  PTRACE(4, "XML\tLoading URL " << url);
+
+  PString data;
+  if (url.GetScheme() == "file") 
+    return LoadFile(url.AsFilePath());
+
+  PHTTPClient client;
+  PINDEX contentLength;
+  PMIMEInfo outMIME, replyMIME;
+
+  // make sure we do not hang around for ever
+  client.SetReadTimeout(timeout);
+
+  // get the resource header information
+  if (!client.GetDocument(url, outMIME, replyMIME)) {
+    errorString = PString("Cannot load URL") & url.AsString();
+    errorCol = errorLine = 0;
+    return FALSE;
+  }
+
+  // get the length of the data
+  if (!replyMIME.Contains(PHTTPClient::ContentLengthTag))
+    contentLength = (PINDEX)replyMIME[PHTTPClient::ContentLengthTag].AsUnsigned();
+  else
+    contentLength = P_MAX_INDEX;
+
+  // download the resource into memory
+  PINDEX offs = 0;
+  for (;;) {
+    PINDEX len;
+    if (contentLength == P_MAX_INDEX)
+      len = CACHE_BUFFER_SIZE;
+    else if (offs == contentLength)
+      break;
+    else
+      len = PMIN(contentLength = offs, CACHE_BUFFER_SIZE);
+
+    if (!client.Read(offs + data.GetPointer(offs + len), len))
+      break;
+
+    len = client.GetLastReadCount();
+
+    offs += len;
+  }
+
+  return Load(data, _options);
+}
+
+BOOL PXML::StartAutoReloadURL(const PURL & url, 
+                              const PTimeInterval & timeout, 
+                              const PTimeInterval & refreshTime,
+                              int _options)
+{
+  if (url.IsEmpty()) {
+    autoLoadError = "Cannot auto-load empty URL";
+    return FALSE;
+  }
+
+  PWaitAndSignal m(autoLoadMutex);
+  autoLoadTimer.Stop();
+
+  SetOptions(_options);
+  autoloadURL      = url;
+  autoLoadWaitTime = timeout;
+  autoLoadError    = PString::Empty();
+  autoLoadTimer.SetNotifier(PCREATE_NOTIFIER(AutoReloadTimeout));
+
+  BOOL stat = AutoLoadURL();
+
+  autoLoadTimer = refreshTime;
+
+  return stat;
+}
+
+void PXML::AutoReloadTimeout(PTimer &, INT)
+{
+  PThread::Create(PCREATE_NOTIFIER(AutoReloadThread), PThread::AutoDeleteThread);
+}
+
+void PXML::AutoReloadThread(PThread &, INT)
+{
+  PWaitAndSignal m(autoLoadMutex);
+  OnAutoLoad(AutoLoadURL());
+  autoLoadTimer.Reset();
+}
+
+void PXML::OnAutoLoad(BOOL ok)
+{
+  PTRACE_IF(3, !ok, "XML\tFailed to load XML: " << GetErrorString());
+}
+
+BOOL PXML::AutoLoadURL()
+{
+  BOOL stat = LoadURL(autoloadURL, autoLoadWaitTime);
+  if (stat)
+    autoLoadError = PString::Empty();
+  else 
+    autoLoadError = GetErrorString() + psprintf(" at line %i, column %i", GetErrorLine(), GetErrorColumn());
+  return stat;
+}
+
+BOOL PXML::StopAutoReloadURL()
+{
+  PWaitAndSignal m(autoLoadMutex);
+  autoLoadTimer.Stop();
+  return TRUE;
+}
+
+
 BOOL PXML::Load(const PString & data, int _options)
 {
   if (_options >= 0)
     options = _options;
 
-  XML_Parser parser = XML_ParserCreate(NULL);
-  XML_SetUserData(parser, this);
+  BOOL stat = FALSE;
+  PXMLElement * loadingRootElement = NULL;
 
-  XML_SetElementHandler      (parser, PXML_StartElement, PXML_EndElement);
-  XML_SetCharacterDataHandler(parser, PXML_CharacterDataHandler);
-  XML_SetXmlDeclHandler      (parser, PXML_XmlDeclHandler);
-  XML_SetDoctypeDeclHandler  (parser, PXML_StartDocTypeDecl, PXML_EndDocTypeDecl);
+  {
+    PXMLParser parser(options);
+    int done = 1;
+    stat = parser.Parse(data, data.GetLength(), done) != 0;
+  
+    if (!stat)
+      parser.GetErrorInfo(errorString, errorCol, errorLine);
 
-  int done = 1;
-  if (XML_Parse(parser, (const char *)data, data.GetLength(), done) != 0)
-    return TRUE;
+    version    = parser.GetVersion();
+    encoding   = parser.GetEncoding();
+    standAlone = parser.GetStandAlone();
 
-  XML_Error err = XML_GetErrorCode(parser);
-  errorString = PString(XML_ErrorString(err));
-  errorCol    = XML_GetCurrentColumnNumber(parser);
-  errorLine   = XML_GetCurrentLineNumber(parser);
+    loadingRootElement = parser.GetXMLTree();
+  }
 
-  return FALSE;
+  if (stat) {
+    if (loadingRootElement == NULL) {
+      errorString = "XML\tFailed to create root node in XML!";
+      return FALSE;
+    }
+    else {
+      PWaitAndSignal m(rootMutex);
+      if (rootElement != NULL) {
+        delete rootElement;
+        rootElement = NULL;
+      }
+      rootElement = loadingRootElement;
+      PTRACE(4, "XML\tLoaded XML " << rootElement->GetName());
+    }
+    OnLoaded();
+  }
+
+  return stat;
 }
 
 BOOL PXML::Save(int _options)
@@ -204,6 +571,8 @@ BOOL PXML::Save(int _options)
 
 BOOL PXML::SaveFile(const PFilePath & fn, int _options)
 {
+  PWaitAndSignal m(rootMutex);
+
   PFile file;
   if (!file.Open(fn, PFile::WriteOnly)) 
     return FALSE;
@@ -217,6 +586,8 @@ BOOL PXML::SaveFile(const PFilePath & fn, int _options)
 
 BOOL PXML::Save(PString & data, int _options)
 {
+  PWaitAndSignal m(rootMutex);
+
   if (_options >= 0)
     options = _options;
 
@@ -226,6 +597,16 @@ BOOL PXML::Save(PString & data, int _options)
   return TRUE;
 }
 
+void PXML::RemoveAll()
+{
+  PWaitAndSignal m(rootMutex);
+
+  if (rootElement != NULL) {
+    delete rootElement;
+    rootElement = NULL;
+  }
+}
+
 PXMLElement * PXML::GetElement(const PCaselessString & name, PINDEX idx) const
 {
   if (rootElement == NULL)
@@ -244,78 +625,36 @@ PXMLElement * PXML::GetElement(PINDEX idx) const
   return (PXMLElement *)(rootElement->GetElement(idx));
 }
 
+BOOL PXML::RemoveElement(PINDEX idx)
+{
+  if (rootElement == NULL)
+    return FALSE;
+
+  if (idx >= rootElement->GetSize())
+    return FALSE;
+
+  rootElement->RemoveElement(idx);
+  return TRUE;
+}
+
+
 PINDEX PXML::GetNumElements() const
 {
-	if (rootElement == NULL) return 0;
-	else return rootElement->GetSize();
+	if (rootElement == NULL) 
+    return 0;
+	else 
+    return rootElement->GetSize();
 }
 
-void PXML::StartElement(const char * name, const char **attrs)
+BOOL PXML::IsNoIndentElement(const PString & elementName) const
 {
-  PXMLElement * newElement = new PXMLElement(currentElement, name);
-  if (currentElement != NULL)
-    currentElement->AddSubObject(newElement, FALSE);
-
-  while (attrs[0] != NULL) {
-    newElement->SetAttribute(PString(attrs[0]), PString(attrs[1]));
-    attrs += 2;
-  }
-
-  currentElement = newElement;
-  lastElement    = NULL;
-
-  if (rootElement == NULL) 
-    rootElement = currentElement;
+  return noIndentElements.GetValuesIndex(elementName) != P_MAX_INDEX;
 }
 
-void PXML::EndElement(const char * /*name*/)
-{
-  currentElement = currentElement->GetParent();
-  lastElement    = NULL;
-}
-
-void PXML::AddCharacterData(const char * data, int len)
-{
-  PString str(data, len);
-
-  if (lastElement != NULL) {
-    PAssert(lastElement->IsData(), "lastElement set by non-data element");
-    if ((options & NoIgnoreWhiteSpace) == 0)
-      str = str.Trim();
-    lastElement->SetString(lastElement->GetString() & str, FALSE);
-  } else {
-    if ((options & NoIgnoreWhiteSpace) == 0) {
-      str = str.Trim();
-      if (str.IsEmpty()) 
-        return;
-    }
-    PXMLData * newElement = new PXMLData(currentElement, str);
-    currentElement->AddSubObject(newElement, FALSE);
-    lastElement = newElement;
-  }
-}
-
-void PXML::XmlDecl(const char * _version, const char * _encoding, int _standAlone)
-{
-  version    = _version;
-  encoding   = _encoding;
-  standAlone = _standAlone;
-}
-
-void PXML::StartDocTypeDecl(const char * /*_docTypeName*/,
-	                    const char * /*sysid*/,
-			    const char * /*pubid*/,
-			             int /*hasInternalSubSet*/)
-{
-}
-
-void PXML::EndDocTypeDecl()
-{
-}
 
 void PXML::PrintOn(ostream & strm) const
 {
-  BOOL newLine = (options & PXML::NewLineAfterElement) != 0;
+  BOOL newLine = (options & (PXMLParser::Indent|PXMLParser::NewLineAfterElement)) != 0;
 
 //<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
 
@@ -350,29 +689,29 @@ void PXML::PrintOn(ostream & strm) const
     strm << "<!DOCTYPE " << rootElement->GetName() << '>';
     if (newLine)
       strm << endl;
-    rootElement->PrintOn(strm, (options & Indent) ? 0 : -1, options);
+    rootElement->Output(strm, *this, 2);
   }
 }
 
-PString CreateStartTag(const PString & text)
+PString PXML::CreateStartTag(const PString & text)
 {
   return '<' + text + '>';
 }
 
 
-PString CreateEndTag(const PString & text)
+PString PXML::CreateEndTag(const PString & text)
 {
   return "</" + text + '>';
 }
 
 
-PString CreateTagNoData(const PString & text)
+PString PXML::CreateTagNoData(const PString & text)
 {
   return '<' + text + "/>";
 }
 
 
-PString CreateTag(const PString & text, const PString & data)
+PString PXML::CreateTag(const PString & text, const PString & data)
 {
   return CreateStartTag(text) + data + CreateEndTag(text);
 }
@@ -387,6 +726,24 @@ void PXMLObject::SetDirty()
     parent->SetDirty();
 }
 
+PXMLObject * PXMLObject::GetNextObject()
+{
+	if (parent == NULL)
+		return NULL;
+
+  // find our index in our parent's list
+  PINDEX idx = parent->FindObject(this);
+  if (idx == P_MAX_INDEX)
+    return NULL;
+
+  // get the next object
+  ++idx;
+  if (idx >= parent->GetSize())
+    return NULL;
+
+  return (*parent).GetElement(idx);
+}
+
 ///////////////////////////////////////////////////////
 
 PXMLData::PXMLData(PXMLElement * _parent, const PString & _value)
@@ -401,12 +758,18 @@ PXMLData::PXMLData(PXMLElement * _parent, const char * data, int len)
   value = PString(data, len);
 }
 
-void PXMLData::PrintOn(ostream & strm, int indent, int options) const
+void PXMLData::Output(ostream & strm, const PXMLBase & xml, int indent) const
 {
-  if (indent > 0)
-    strm << setw(indent) << " ";
+  int options = xml.GetOptions();
+  if (xml.IsNoIndentElement(parent->GetName()))
+    options &= ~PXMLParser::Indent;
+
+  if (options & PXMLParser::Indent)
+    strm << setw(indent-1) << " ";
+
   strm << value;
-  if (indent >= 0 || ((options & PXML::NewLineAfterElement) != 0))
+
+  if ((options & (PXMLParser::Indent|PXMLParser::NewLineAfterElement)) != 0)
     strm << endl;
 }
 
@@ -439,6 +802,12 @@ PXMLElement::PXMLElement(PXMLElement * _parent, const PString & _name, const PSt
   AddSubObject(new PXMLData(this, data));
 }
 
+PINDEX PXMLElement::FindObject(PXMLObject * ptr) const
+{
+  return subObjects.GetObjectsIndex(ptr);
+}
+
+
 PXMLElement * PXMLElement::GetElement(const PCaselessString & name, PINDEX start) const
 {
   PINDEX idx;
@@ -464,6 +833,16 @@ PXMLObject * PXMLElement::GetElement(PINDEX idx) const
   return &subObjects[idx];
 }
 
+BOOL PXMLElement::RemoveElement(PINDEX idx)
+{
+  if (idx >= subObjects.GetSize())
+    return FALSE;
+
+  subObjects.RemoveAt(idx);
+  return TRUE;
+}
+
+
 PString PXMLElement::GetAttribute(const PCaselessString & key) const
 {
   return attributes(key);
@@ -499,12 +878,21 @@ BOOL PXMLElement::HasAttribute(const PCaselessString & key)
   return attributes.Contains(key);
 }
 
-void PXMLElement::PrintOn(ostream & strm, int indent, int options) const
+void PXMLElement::PrintOn(ostream & strm) const
 {
-  BOOL newLine = (options & PXML::NewLineAfterElement) != 0;
+  PXMLBase xml(-1);
+  Output(strm, xml, 0);
+}
+
+void PXMLElement::Output(ostream & strm, const PXMLBase & xml, int indent) const
+{
+  int options = xml.GetOptions();
+
+  BOOL newLine = (options & (PXMLParser::Indent|PXMLParser::NewLineAfterElement)) != 0;
+
+  if ((options & PXMLParser::Indent) != 0)
+    strm << setw(indent-1) << " ";
 
-  if (indent > 0)
-    strm << setw(indent) << " ";
   strm << '<' << name;
 
   PINDEX i;
@@ -515,24 +903,27 @@ void PXMLElement::PrintOn(ostream & strm, int indent, int options) const
     }
   }
 
-  if (((options & PXML::CloseExtended) != 0) &&
-      ((options & PXML::NoIgnoreWhiteSpace) == 0) &&
-      (subObjects.GetSize() == 0)) {
-    strm << " />";
+  // this ensures empty elements use the shortened form
+  if (subObjects.GetSize() == 0) {
+    strm << "/>";
     if (newLine)
       strm << endl;
-  } else {
+  }
+  else {
+    BOOL indenting = (options & PXMLParser::Indent) != 0 && !xml.IsNoIndentElement(name);
+
     strm << '>';
-    if (indent >= 0 || newLine)
+    if (indenting)
       strm << endl;
   
     for (i = 0; i < subObjects.GetSize(); i++) 
-      subObjects[i].PrintOn(strm, (indent < 0) ? -1 : (indent + 2), options);
+      subObjects[i].Output(strm, xml, indent + 2);
+
+    if (indenting)
+      strm << setw(indent-1) << " ";
 
-    if (indent > 0)
-      strm << setw(indent) << " ";
     strm << "</" << name << '>';
-    if (indent >= 0 || newLine)
+    if (newLine)
       strm << endl;
   }
 }
@@ -576,9 +967,12 @@ PString PXMLElement::GetData() const
   PString str;
   PINDEX idx;
   for (idx = 0; idx < subObjects.GetSize(); idx++) {
-    if (subObjects[idx].IsData()) {
+    if (!subObjects[idx].IsElement()) {
       PXMLData & dataElement = ((PXMLData &)subObjects[idx]);
-      str = str & dataElement.GetString();
+      PStringArray lines = dataElement.GetString().Lines();
+      PINDEX j;
+      for (j = 0; j < lines.GetSize(); j++)
+        str = str & lines[j];
     }
   }
   return str;
