@@ -27,6 +27,28 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: collect.cxx,v $
+ * Revision 1.59  2002/12/02 04:27:37  robertj
+ * Added extra bullet proofing for some pathological conditions.
+ *
+ * Revision 1.58  2002/11/12 08:57:18  robertj
+ * Changed scope of PAbstraSortedList::Element class so descendant classes
+ *   can get at it.
+ * Fixed problem with GetValuesIndex not returning first element of a certain
+ *   value if the object was a PCaselessString.
+ *
+ * Revision 1.57  2002/06/25 02:24:59  robertj
+ * Improved assertion system to allow C++ class name to be displayed if
+ *   desired, especially relevant to container classes.
+ *
+ * Revision 1.56  2002/06/12 09:40:58  robertj
+ * Fixed printing of a dictionary to utilise the stream fill character between
+ *   each dictiionary element, as per general container semantics.
+ *
+ * Revision 1.55  2002/04/26 05:40:21  robertj
+ * Removed assumption that GetAt() on a dictionary will automatically convert
+ *   the index to a POrdinalyKey. This breaks the PCollection semantics for
+ *   GetAt() which is to get based on the ordinal position not the hashed position.
+ *
  * Revision 1.54  2002/04/16 07:58:13  robertj
  * Fixed MakeUnique for lists and sorted lists.
  *
@@ -202,6 +224,8 @@
 
 #ifndef _WIN32_WCE
 #define new PNEW
+#undef  __CLASS__
+#define __CLASS__ GetClass()
 #endif
 
 
@@ -529,7 +553,10 @@ BOOL PAbstractList::Remove(const PObject * obj)
 
 PObject * PAbstractList::RemoveAt(PINDEX index)
 {
-  PAssert(SetCurrent(index), PInvalidArrayIndex);
+  if (!SetCurrent(index)) {
+    PAssertAlways(PInvalidArrayIndex);
+    return NULL;
+  }
 
   Element * elmt = info->lastElement;
 
@@ -658,7 +685,7 @@ PAbstractList::Element::Element(PObject * theData)
 
 ///////////////////////////////////////////////////////////////////////////////
 
-static PAbstractSortedList::Element nil = NULL;
+PAbstractSortedList::Element PAbstractSortedList::Element::nil = NULL;
 
 PAbstractSortedList::PAbstractSortedList()
 {
@@ -683,14 +710,14 @@ void PAbstractSortedList::CopyContents(const PAbstractSortedList & list)
 void PAbstractSortedList::CloneContents(const PAbstractSortedList * list)
 {
   Element * element = list->info->root;
-  while (element->left != &nil)
+  while (element->left != &Element::nil)
     element = element->left;
 
   info = new Info;
-  PAssertNULL(info);
+  PAssert(info != NULL, POutOfMemory);
   reference->size = 0;
 
-  while (element != &nil) {
+  while (element != &Element::nil) {
     Append(element->data->Clone());
     element = element->Successor();
   }
@@ -707,17 +734,17 @@ PObject::Comparison PAbstractSortedList::Compare(const PObject & obj) const
 {
   PAssert(obj.IsDescendant(PAbstractSortedList::Class()), PInvalidCast);
   Element * elmt1 = info->root;
-  while (elmt1->left != &nil)
+  while (elmt1->left != &Element::nil)
     elmt1 = elmt1->left;
 
   Element * elmt2 = ((const PAbstractSortedList &)obj).info->root;
-  while (elmt2->left != &nil)
+  while (elmt2->left != &Element::nil)
     elmt2 = elmt2->left;
 
-  while (elmt1 != &nil && elmt2 != &nil) {
-    if (elmt1 == &nil)
+  while (elmt1 != &Element::nil && elmt2 != &Element::nil) {
+    if (elmt1 == &Element::nil)
       return LessThan;
-    if (elmt2 == &nil)
+    if (elmt2 == &Element::nil)
       return GreaterThan;
     if (*elmt1->data < *elmt2->data)
       return LessThan;
@@ -734,14 +761,14 @@ PINDEX PAbstractSortedList::Append(PObject * obj)
 {
   Element * z = new Element(PAssertNULL(obj));
   Element * x = info->root;
-  Element * y = &nil;
-  while (x != &nil) {
+  Element * y = &Element::nil;
+  while (x != &Element::nil) {
     x->subTreeSize++;
     y = x;
     x = *z->data < *x->data ? x->left : x->right;
   }
   z->parent = y;
-  if (y == &nil)
+  if (y == &Element::nil)
     info->root = z;
   else if (*z->data < *y->data)
     y->left = z;
@@ -818,6 +845,9 @@ BOOL PAbstractSortedList::Remove(const PObject * obj)
 PObject * PAbstractSortedList::RemoveAt(PINDEX index)
 {
   Element * node = info->root->OrderSelect(index+1);
+  if (node == &Element::nil)
+    return NULL;
+
   PObject * data = node->data;
   RemoveElement(node);
   return reference->deleteObjects ? (PObject *)NULL : data;
@@ -826,10 +856,10 @@ PObject * PAbstractSortedList::RemoveAt(PINDEX index)
 
 void PAbstractSortedList::RemoveAll()
 {
-  if (info->root != &nil) {
+  if (info->root != &Element::nil) {
     info->root->DeleteSubTrees(reference->deleteObjects);
     delete info->root;
-    info->root = &nil;
+    info->root = &Element::nil;
     reference->size = 0;
   }
 }
@@ -888,14 +918,14 @@ PINDEX PAbstractSortedList::GetObjectsIndex(const PObject * obj) const
     PINDEX savePos = pos;
     Element * saveElmt = elmt;
     while (elmt->data != obj &&
-            (elmt = elmt->Predecessor()) != &nil &&
+            (elmt = elmt->Predecessor()) != &Element::nil &&
             *obj == *elmt->data)
       pos--;
     if (elmt->data != obj) {
       pos = savePos;
       elmt = saveElmt;
       while (elmt->data != obj &&
-              (elmt = elmt->Successor()) != &nil &&
+              (elmt = elmt->Successor()) != &Element::nil &&
               *obj == *elmt->data)
         pos++;
       if (elmt->data != obj)
@@ -919,7 +949,8 @@ PINDEX PAbstractSortedList::GetValuesIndex(const PObject & obj) const
   info->lastIndex = pos;
 
   Element * prev;
-  while ((prev = info->lastElement->Predecessor()) != &nil && obj == *prev->data) {
+  while ((prev = info->lastElement->Predecessor()) != &Element::nil &&
+                                  prev->data->Compare(obj) == EqualTo) {
     info->lastElement = prev;
     info->lastIndex--;
   }
@@ -932,21 +963,25 @@ void PAbstractSortedList::RemoveElement(Element * node)
 {
   PAssertNULL(node);
 
+  // Don't try an remove one of the special leaf nodes!
+  if (node == &Element::nil)
+    return;
+
   if (node->data != NULL && reference->deleteObjects)
     delete node->data;
 
-  Element * y = node->left == &nil || node->right == &nil ? node : node->Successor();
+  Element * y = node->left == &Element::nil || node->right == &Element::nil ? node : node->Successor();
 
   Element * t = y;
-  while (t != &nil) {
+  while (t != &Element::nil) {
     t->subTreeSize--;
     t = t->parent;
   }
 
-  Element * x = y->left != &nil ? y->left : y->right;
+  Element * x = y->left != &Element::nil ? y->left : y->right;
   x->parent = y->parent;
 
-  if (y->parent == &nil)
+  if (y->parent == &Element::nil)
     info->root = x;
   else if (y == y->parent->left)
     y->parent->left = x;
@@ -1026,10 +1061,10 @@ void PAbstractSortedList::LeftRotate(Element * node)
 {
   Element * pivot = PAssertNULL(node)->right;
   node->right = pivot->left;
-  if (pivot->left != &nil)
+  if (pivot->left != &Element::nil)
     pivot->left->parent = node;
   pivot->parent = node->parent;
-  if (node->parent == &nil)
+  if (node->parent == &Element::nil)
     info->root = pivot;
   else if (node == node->parent->left)
     node->parent->left = pivot;
@@ -1046,10 +1081,10 @@ void PAbstractSortedList::RightRotate(Element * node)
 {
   Element * pivot = PAssertNULL(node)->left;
   node->left = pivot->right;
-  if (pivot->right != &nil)
+  if (pivot->right != &Element::nil)
     pivot->right->parent = node;
   pivot->parent = node->parent;
-  if (node->parent == &nil)
+  if (node->parent == &Element::nil)
     info->root = pivot;
   else if (node == node->parent->right)
     node->parent->right = pivot;
@@ -1091,7 +1126,7 @@ PAbstractSortedList::Element * PAbstractSortedList::Element::Successor() const
 }
 
 
-PAbstractSortedList::Element*PAbstractSortedList::Element::Predecessor() const
+PAbstractSortedList::Element * PAbstractSortedList::Element::Predecessor() const
 {
   Element * pred;
   if (left != &nil) {
@@ -1126,7 +1161,7 @@ PAbstractSortedList::Element * PAbstractSortedList::Element::OrderSelect(PINDEX
       return right->OrderSelect(index - r);
   }
 
-  PAssertAlways("Order select failed!");
+  PAssertAlways2("PAbstractSortedList::Element", "Order select failed!");
   return &nil;
 }
 
@@ -1178,7 +1213,7 @@ void PAbstractSortedList::Element::DeleteSubTrees(BOOL deleteObject)
 
 PAbstractSortedList::Info::Info()
 {
-  root = &nil;
+  root = &Element::nil;
   lastElement = NULL;
   lastIndex = P_MAX_INDEX;
 }
@@ -1249,7 +1284,7 @@ PINDEX PHashTable::Table::AppendElement(PObject * key, PObject * data)
   PINDEX bucket = PAssertNULL(key)->HashFunction();
   Element * list = GetAt(bucket);
   Element * element = new Element;
-  PAssertNULL(element);
+  PAssert(element != NULL, POutOfMemory);
   element->key = key;
   element->data = data;
   if (list == NULL) {
@@ -1387,7 +1422,7 @@ PINDEX PHashTable::Table::GetElementsIndex(
 PHashTable::PHashTable()
   : hashTable(new PHashTable::Table)
 {
-  PAssertNULL(hashTable);
+  PAssert(hashTable != NULL, POutOfMemory);
   hashTable->lastElement = NULL;
 }
 
@@ -1413,7 +1448,7 @@ void PHashTable::CloneContents(const PHashTable * hash)
   PHashTable::Table * original = hash->hashTable;
 
   hashTable = new PHashTable::Table(original->GetSize());
-  PAssertNULL(hashTable);
+  PAssert(hashTable != NULL, POutOfMemory);
   hashTable->lastElement = NULL;
 
   for (PINDEX i = 0; i < sz; i++) {
@@ -1570,10 +1605,14 @@ PINDEX PAbstractDictionary::InsertAt(PINDEX index, PObject * obj)
 }
  
  
-BOOL PAbstractDictionary::Remove(const PObject *)
+BOOL PAbstractDictionary::Remove(const PObject * obj)
 {
-  PAssertAlways(PUnimplementedFunction);
-  return FALSE;
+  PINDEX idx = GetObjectsIndex(obj);
+  if (idx == P_MAX_INDEX)
+    return FALSE;
+
+  RemoveAt(idx);
+  return TRUE;
 }
 
 
@@ -1599,13 +1638,14 @@ PINDEX PAbstractDictionary::GetValuesIndex(const PObject & obj) const
 
 BOOL PAbstractDictionary::SetAt(PINDEX index, PObject * val)
 {
-  return AbstractSetAt(POrdinalKey(index), val);
+  return AbstractSetAt(AbstractGetKeyAt(index), val);
 }
 
 
 PObject * PAbstractDictionary::GetAt(PINDEX index) const
 {
-  return AbstractGetAt(POrdinalKey(index));
+  PAssert(hashTable->SetLastElementAt(index), PInvalidArrayIndex);
+  return hashTable->lastElement->data;
 }
  
  
@@ -1657,8 +1697,18 @@ PObject & PAbstractDictionary::GetRefAt(const PObject & key) const
 
 void PAbstractDictionary::PrintOn(ostream &strm) const
 {
-  for (PINDEX  i = 0; i < GetSize(); i++)
-    strm << AbstractGetKeyAt(i) << '=' << AbstractGetDataAt(i) << endl;
+  char separator = strm.fill();
+  if (separator == ' ')
+    separator = '\n';
+
+  for (PINDEX i = 0; i < GetSize(); i++) {
+    if (i > 0)
+      strm << separator;
+    strm << AbstractGetKeyAt(i) << '=' << AbstractGetDataAt(i);
+  }
+
+  if (separator == '\n')
+    strm << separator;
 }
 
 
