@@ -27,6 +27,88 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: contain.cxx,v $
+ * Revision 1.134  2003/07/28 18:44:01  dsandras
+ * Make use of the libc regex on Linux.
+ *
+ * Revision 1.133  2003/05/14 00:48:33  rjongbloed
+ * Added constructor to string lists/arrays etc that takes a single PString.
+ * Fixed bug in doing a MakeUnique on a container, it would lose the
+ *   DisallowDeleteObjects flag.
+ *
+ * Revision 1.132  2003/04/28 09:14:14  robertj
+ * Fixed bad sign extension problem in PBYTEArray output
+ *
+ * Revision 1.131  2003/04/15 07:08:37  robertj
+ * Changed read and write from streams for base array classes so operates in
+ *   the same way for both PIntArray and PArray<int> etc
+ *
+ * Revision 1.130  2003/03/31 01:24:23  robertj
+ * Added ReadFrom functions for standard container classes such as
+ *   PIntArray and PStringList etc
+ *
+ * Revision 1.129  2003/03/05 08:48:32  robertj
+ * Added PStringArray::ToCharAray() function at suggestion of Ravelli Rossano
+ *
+ * Revision 1.128  2003/02/02 23:29:34  robertj
+ * Fixed bug in RightTrim() (lost off last non blank char), tnanks Joerg Schoemer
+ *
+ * Revision 1.127  2002/12/16 08:10:35  robertj
+ * Fixed infinite loop when converting an illegal (incomplete) UTF-8 string
+ *   to UCS-2, thanks Chih-Wei Huang
+ *
+ * Revision 1.126  2002/11/26 01:08:33  robertj
+ * Fixed problem when using pre-initialised PStringStream greater than 255
+ *   bytes, would truncate and lose trailing null. Reported by Thien Nguyen
+ *
+ * Revision 1.125  2002/11/12 09:18:03  robertj
+ * Added PString::NumCompare() as functional equivalent of strncmp().
+ * Added PSortedStringList::GetNextStringsIndex() to do searches of binary
+ *   tree on partal strings.
+ *
+ * Revision 1.124  2002/11/01 05:10:01  robertj
+ * Fixed bug in UTF-8 to UCS-2 conversion, not compiler portable!
+ *
+ * Revision 1.123  2002/10/31 07:33:59  robertj
+ * Changed UTF-8 to UCS-2 conversion function to not include trailing null.
+ *
+ * Revision 1.122  2002/10/31 05:55:55  robertj
+ * Now comprehensively stated that a PString is ALWAYS an 8 bit string as
+ *   there are far too many inheerent assumptions every to make it 16 bit.
+ * Added UTF-8/UCS-2 conversion functions to PString.
+ *
+ * Revision 1.121  2002/10/10 04:43:44  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.120  2002/08/14 00:43:40  robertj
+ * Added ability to have fixed maximum length PStringStream's so does not do
+ *   unwanted malloc()'s while outputing data.
+ *
+ * Revision 1.119  2002/08/06 08:51:36  robertj
+ * Added missing va_end, thanks Klaus Kaempf
+ *
+ * Revision 1.118  2002/06/27 06:59:34  robertj
+ * Removed memory leak display of static that is not really a leak.
+ *
+ * Revision 1.117  2002/06/25 02:24:24  robertj
+ * Improved assertion system to allow C++ class name to be displayed if
+ *   desired, especially relevant to container classes.
+ *
+ * Revision 1.116  2002/06/24 06:18:36  robertj
+ * Fixed bug when getting extra space at start of outputing PBaseArray.
+ * Added ability to not include ASCII in PbaseArray output using ios::fixed.
+ *
+ * Revision 1.115  2002/06/19 04:04:30  robertj
+ * Fixed bug in setting/getting bits from PBitArray, could exceed array bounds.
+ *
+ * Revision 1.114  2002/06/17 09:16:18  robertj
+ * Fixed strange deadlock woth gcc 3.0.2, thanks Artis Kugevics
+ *
+ * Revision 1.113  2002/06/14 13:22:52  robertj
+ * Added PBitArray class.
+ *
+ * Revision 1.112  2002/06/05 12:29:15  craigs
+ * Changes for gcc 3.1
+ *
  * Revision 1.111  2002/04/09 02:30:18  robertj
  * Removed GCC3 variable as __GNUC__ can be used instead, thanks jason Spence
  *
@@ -432,14 +514,20 @@ extern "C" int vsprintf(char *, const char *, va_list);
 #endif
 
 #if !P_MACOS
+#ifdef  P_LINUX
+#include <regex.h>
+#else
 #include "regex/regex.h"
 #endif
+#endif
 
 #if !P_USE_INLINES
 #include "ptlib/contain.inl"
 #endif
 
 #define new PNEW
+#undef  __CLASS__
+#define __CLASS__ GetClass()
 
 
 ///////////////////////////////////////////////////////////////////////////////
@@ -447,20 +535,26 @@ extern "C" int vsprintf(char *, const char *, va_list);
 PContainer::PContainer(PINDEX initialSize)
 {
   reference = new Reference(initialSize);
-  PAssertNULL(reference);
+  PAssert(reference != NULL, POutOfMemory);
 }
 
 
 PContainer::PContainer(int, const PContainer * cont)
 {
+  PAssertNULL(cont);
+  PAssert2(cont->reference != NULL, cont->GetClass(), "Clone of deleted container");
+
   reference = new Reference(0);
-  *PAssertNULL(reference) = *cont->reference;
+  PAssert(reference != NULL, POutOfMemory);
+
+  *reference = *cont->reference;
 }
 
 
 PContainer::PContainer(const PContainer & cont)
-{                                                            
-  reference = PAssertNULL(cont.reference);
+{
+  PAssert2(cont.reference != NULL, cont.GetClass(), "Copy of deleted container");
+  reference = cont.reference;
   reference->count++;
 }
 
@@ -477,7 +571,8 @@ void PContainer::AssignContents(const PContainer & cont)
     delete reference;
   }
 
-  reference = PAssertNULL(cont.reference);
+  PAssert2(cont.reference != NULL, cont.GetClass(), "Assign of deleted container");
+  reference = cont.reference;
   reference->count++;
 }
 
@@ -511,7 +606,7 @@ BOOL PContainer::MakeUnique()
     return TRUE;
 
   reference->count--;
-  reference = new Reference(GetSize());
+  reference = new Reference(*reference);
   return FALSE;
 }
 
@@ -526,8 +621,10 @@ PAbstractArray::PAbstractArray(PINDEX elementSizeInBytes, PINDEX initialSize)
 
   if (GetSize() == 0)
     theArray = NULL;
-  else
+  else {
     theArray = (char *)calloc(GetSize(), elementSize);
+    PAssert(theArray != NULL, POutOfMemory);
+  }
 
   allocatedDynamically = TRUE;
 }
@@ -549,6 +646,7 @@ PAbstractArray::PAbstractArray(PINDEX elementSizeInBytes,
   else if (dynamicAllocation) {
     PINDEX sizebytes = elementSize*GetSize();
     theArray = (char *)malloc(sizebytes);
+    PAssert(theArray != NULL, POutOfMemory);
     memcpy(theArray, PAssertNULL(buffer), sizebytes);
   }
   else
@@ -588,6 +686,33 @@ void PAbstractArray::CloneContents(const PAbstractArray * array)
 }
 
 
+void PAbstractArray::PrintOn(ostream & strm) const
+{
+  char separator = strm.fill();
+  int width = strm.width();
+  for (PINDEX  i = 0; i < GetSize(); i++) {
+    if (i > 0 && separator != '\0')
+      strm << separator;
+    strm.width(width);
+    PrintElementOn(strm, i);
+  }
+  if (separator == '\n')
+    strm << '\n';
+}
+
+
+void PAbstractArray::ReadFrom(istream & strm)
+{
+  PINDEX i = 0;
+  while (strm.good()) {
+    ReadElementFrom(strm, i);
+    if (!strm.fail())
+      i++;
+  }
+  SetSize(i);
+}
+
+
 PObject::Comparison PAbstractArray::Compare(const PObject & obj) const
 {
   PAssert(obj.IsDescendant(PAbstractArray::Class()), PInvalidCast);
@@ -720,87 +845,13 @@ BOOL PAbstractArray::Concatenate(const PAbstractArray & array)
 }
 
 
-void PAbstractArray::PrintNumbersOn(ostream & strm, PINDEX size, BOOL is_signed) const
+void PAbstractArray::PrintElementOn(ostream & /*stream*/, PINDEX /*index*/) const
 {
-  PINDEX line_width = strm.width();
-  if (line_width == 0)
-    line_width = 16/size;
-
-  PINDEX indent = strm.precision();
-
-  PINDEX val_width;
-  switch (strm.flags()&ios::basefield) {
-    case ios::hex :
-      val_width = size*2;
-      is_signed = FALSE;
-      break;
-    case ios::oct :
-      val_width = ((size*8)+2)/3;
-      is_signed = FALSE;
-      break;
-    default :
-      switch (size) {
-        case 1 :
-          val_width = 3;
-          break;
-        case 2 :
-          val_width = 5;
-          break;
-        default :
-          val_width = 10;
-          break;
-      }
-      if (is_signed)
-        val_width++;
-  }
-
-  long mask = -1;
-  if (size < sizeof(mask))
-    mask = (1L << (size*8)) - 1;
-
-  PINDEX i = 0;
-  while (i < GetSize()) {
-    if (i > 0)
-      strm << '\n';
-    PINDEX j;
-    for (j = 0; j < indent; j++)
-      strm.put(' ');
-    for (j = 0; j < line_width; j++) {
-      if (j == line_width/2)
-        strm.put(' ');
-      if (i+j < GetSize()) {
-        strm << setw(val_width);
-        if (is_signed)
-          strm << GetNumberValueAt(i+j);
-        else
-          strm << (DWORD)(GetNumberValueAt(i+j)&mask);
-      }
-      else {
-        PINDEX k;
-        for (k = 0; k < val_width; k++)
-          strm.put(' ');
-      }
-      strm << ' ';
-    }
-    strm << "  ";
-    for (j = 0; j < line_width; j++) {
-      if (i+j < GetSize()) {
-        long val = GetNumberValueAt(i+j);
-        if (val >= 0 && val < 256 && isprint(val))
-          strm << (char)val;
-        else
-          strm << '.';
-      }
-    }
-    i += line_width;
-  }
 }
 
 
-long PAbstractArray::GetNumberValueAt(PINDEX) const
+void PAbstractArray::ReadElementFrom(istream & /*stream*/, PINDEX /*index*/)
 {
-  PAssertAlways(PUnimplementedFunction);
-  return 0;
 }
 
 
@@ -819,94 +870,167 @@ void PCharArray::PrintOn(ostream & strm) const
     strm.write(theArray, GetSize());
 
   while (width-- > 0)
-    strm.put(strm.fill());
+    strm << (char)strm.fill();
 
   if (!left)
     strm.write(theArray, GetSize());
 }
 
 
-void PShortArray::PrintOn(ostream & strm) const
+void PCharArray::ReadFrom(istream &strm)
 {
-  PrintNumbersOn(strm, sizeof(short), TRUE);
-}
+  PINDEX size = 0;
+  SetSize(size+100);
 
+  while (strm.good()) {
+    strm >> theArray[size++];
+    if (size >= GetSize())
+      SetSize(size+100);
+  }
 
-long PShortArray::GetNumberValueAt(PINDEX idx) const
-{
-  return ((short *)theArray)[idx];
-}
-
-
-void PIntArray::PrintOn(ostream & strm) const
-{
-  PrintNumbersOn(strm, sizeof(int), TRUE);
-}
-
-
-long PIntArray::GetNumberValueAt(PINDEX idx) const
-{
-  return ((int *)theArray)[idx];
-}
-
-
-void PLongArray::PrintOn(ostream & strm) const
-{
-  PrintNumbersOn(strm, sizeof(long), TRUE);
-}
-
-
-long PLongArray::GetNumberValueAt(PINDEX idx) const
-{
-  return ((long *)theArray)[idx];
+  SetSize(size);
 }
 
 
 void PBYTEArray::PrintOn(ostream & strm) const
 {
-  PrintNumbersOn(strm, sizeof(BYTE), FALSE);
+  PINDEX line_width = strm.width();
+  if (line_width == 0)
+    line_width = 16;
+  strm.width(0);
+
+  PINDEX indent = strm.precision();
+
+  PINDEX val_width = ((strm.flags()&ios::basefield) == ios::hex) ? 2 : 3;
+
+  PINDEX i = 0;
+  while (i < GetSize()) {
+    if (i > 0)
+      strm << '\n';
+    PINDEX j;
+    for (j = 0; j < indent; j++)
+      strm << ' ';
+    for (j = 0; j < line_width; j++) {
+      if (j == line_width/2)
+        strm << ' ';
+      if (i+j < GetSize())
+        strm << setw(val_width) << (theArray[i+j]&0xff);
+      else {
+        PINDEX k;
+        for (k = 0; k < val_width; k++)
+          strm << ' ';
+      }
+      strm << ' ';
+    }
+    if ((strm.flags()&ios::floatfield) != ios::fixed) {
+      strm << "  ";
+      for (j = 0; j < line_width; j++) {
+        if (i+j < GetSize()) {
+          unsigned val = theArray[i+j]&0xff;
+          if (isprint(val))
+            strm << (char)val;
+          else
+            strm << '.';
+        }
+      }
+    }
+    i += line_width;
+  }
 }
 
 
-long PBYTEArray::GetNumberValueAt(PINDEX idx) const
+void PBYTEArray::ReadFrom(istream &strm)
 {
-  return ((BYTE *)theArray)[idx];
+  PINDEX size = 0;
+  SetSize(size+100);
+
+  while (strm.good()) {
+    unsigned v;
+    strm >> v;
+    theArray[size] = (BYTE)v;
+    if (!strm.fail()) {
+      size++;
+      if (size >= GetSize())
+        SetSize(size+100);
+    }
+  }
+
+  SetSize(size);
 }
 
 
-void PWORDArray::PrintOn(ostream & strm) const
+///////////////////////////////////////////////////////////////////////////////
+
+PBitArray::PBitArray(PINDEX initialSize)
+  : PBYTEArray((initialSize+7)>>3)
 {
-  PrintNumbersOn(strm, sizeof(WORD), FALSE);
 }
 
 
-long PWORDArray::GetNumberValueAt(PINDEX idx) const
+PBitArray::PBitArray(const void * buffer,
+                     PINDEX length,
+                     BOOL dynamic)
+  : PBYTEArray((const BYTE *)buffer, (length+7)>>3, dynamic)
 {
-  return ((WORD *)theArray)[idx];
 }
 
 
-void PUnsignedArray::PrintOn(ostream & strm) const
+PObject * PBitArray::Clone() const
 {
-  PrintNumbersOn(strm, sizeof(unsigned), FALSE);
+  return new PBitArray(*this);
 }
 
 
-long PUnsignedArray::GetNumberValueAt(PINDEX idx) const
+PINDEX PBitArray::GetSize() const
 {
-  return ((unsigned *)theArray)[idx];
+  return PBYTEArray::GetSize()<<3;
 }
 
 
-void PDWORDArray::PrintOn(ostream & strm) const
+BOOL PBitArray::SetSize(PINDEX newSize)
 {
-  PrintNumbersOn(strm, sizeof(DWORD), FALSE);
+  return PBYTEArray::SetSize((newSize+7)>>3);
 }
 
 
-long PDWORDArray::GetNumberValueAt(PINDEX idx) const
+BOOL PBitArray::SetAt(PINDEX index, BOOL val)
 {
-  return ((DWORD *)theArray)[idx];
+  if (!SetMinSize(index+1))
+    return FALSE;
+
+  if (val)
+    theArray[index>>3] |= (1 << (index&7));
+  else
+    theArray[index>>3] &= ~(1 << (index&7));
+  return TRUE;
+}
+
+
+BOOL PBitArray::GetAt(PINDEX index) const
+{
+  PASSERTINDEX(index);
+  if (index >= GetSize())
+    return FALSE;
+
+  return (theArray[index>>3]&(1 << (index&7))) != 0;
+}
+
+
+void PBitArray::Attach(const void * buffer, PINDEX bufferSize)
+{
+  PBYTEArray::Attach((const BYTE *)buffer, (bufferSize+7)>>3);
+}
+
+
+BYTE * PBitArray::GetPointer(PINDEX minSize)
+{
+  return PBYTEArray::GetPointer((minSize+7)>>3);
+}
+
+
+BOOL PBitArray::Concatenate(const PBitArray & array)
+{
+  return PAbstractArray::Concatenate(array);
 }
 
 
@@ -914,31 +1038,22 @@ long PDWORDArray::GetNumberValueAt(PINDEX idx) const
 
 const PString & PString::Empty()
 {
+#if PMEMORY_CHECK
+  BOOL ignoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE);
+#endif
   static PString emptyString((const char *)NULL);
+#if PMEMORY_CHECK
+  PMemoryHeap::SetIgnoreAllocations(ignoreAllocations);
+#endif
   return emptyString;
 }
 
 
-#ifdef PHAS_UNICODE
-#define PSTRING_COPY(d, s, l) UnicodeCopy((WORD *)(d), (s), (l))
-#define PSTRING_MOVE(d, doff, s, soff, l) \
-            memmove(((WORD*)(d))+(doff), ((WORD*)(s))+(soff), (l)*sizeof(WORD))
-static void UnicodeCopy(WORD * theArray, char * src, size_t len)
-{
-  while (len-- > 0)
-    *theArray++ = *src++;
-}
-#else
-#define PSTRING_COPY(d, s, l) memcpy((d), (s), (l))
-#define PSTRING_MOVE(d, doff, s, soff, l) memmove((d)+(doff), (s)+(soff), (l))
-#endif
-
-
 PString::PString(const char * cstr)
-  : PSTRING_ANCESTOR_CLASS(cstr != NULL ? strlen(cstr)+1 : 1)
+  : PCharArray(cstr != NULL ? strlen(cstr)+1 : 1)
 {
   if (cstr != NULL)
-    PSTRING_COPY(theArray, cstr, GetSize());
+    memcpy(theArray, cstr, GetSize());
 }
 
 
@@ -947,40 +1062,32 @@ PString::PString(const WORD * ustr)
   if (ustr == NULL)
     SetSize(1);
   else {
-    const WORD * ptr = ustr;
     PINDEX len = 0;
-    while (*ptr++ != 0)
+    while (ustr[len] != 0)
       len++;
-    SetSize(len+1);
-#ifdef PHAS_UNICODE
-    memcpy(theArray, ustr, len*sizeof(WORD))
-#else
-    char * cstr = theArray;
-    while (len-- > 0)
-      *cstr++ = (char)*ustr++;
-#endif
+    InternalFromUCS2(ustr, len);
   }
 }
 
 
 PString::PString(const char * cstr, PINDEX len)
-  : PSTRING_ANCESTOR_CLASS(len+1)
+  : PCharArray(len+1)
 {
-  PSTRING_COPY(theArray, PAssertNULL(cstr), len);
+  if (len > 0)
+    memcpy(theArray, PAssertNULL(cstr), len);
 }
 
 
 PString::PString(const WORD * ustr, PINDEX len)
-  : PSTRING_ANCESTOR_CLASS(len+1)
+  : PCharArray(len+1)
 {
-  PAssertNULL(ustr);
-#ifdef PHAS_UNICODE
-  memcpy(theArray, ustr, len*sizeof(WORD))
-#else
-  char * cstr = theArray;
-  while (len-- > 0)
-    *cstr++ = (char)*ustr++;
-#endif
+  InternalFromUCS2(ustr, len);
+}
+
+
+PString::PString(const PWORDArray & ustr)
+{
+  InternalFromUCS2(ustr, ustr.GetSize());
 }
 
 
@@ -1042,7 +1149,7 @@ PString::PString(ConversionType type, const char * str, ...)
       if (*str != '\0') {
         PINDEX len = *str & 0xff;
         PAssert(SetSize(len+1), POutOfMemory);
-        PSTRING_COPY(theArray, str+1, len);
+        memcpy(theArray, str+1, len);
       }
       break;
 
@@ -1050,7 +1157,7 @@ PString::PString(ConversionType type, const char * str, ...)
       if (str[0] != '\0' && str[1] != '\0') {
         PINDEX len = str[0] | (str[1] << 8);
         PAssert(SetSize(len+1), POutOfMemory);
-        PSTRING_COPY(theArray, str+2, len);
+        memcpy(theArray, str+2, len);
       }
       break;
 
@@ -1064,6 +1171,7 @@ PString::PString(ConversionType type, const char * str, ...)
       va_list args;
       va_start(args, str);
       vsprintf(str, args);
+      va_end(args);
       break;
     }
 
@@ -1325,11 +1433,7 @@ PINDEX PString::HashFunction() const
 
 BOOL PString::IsEmpty() const
 {
-#ifdef PHAS_UNICODE
-  return *(WORD*)theArray == '\0';
-#else
   return *theArray == '\0';
-#endif
 }
 
 
@@ -1376,17 +1480,6 @@ BOOL PString::MakeUnique()
 }
 
 
-#ifdef PHAS_UNICODE
-PINDEX PString::GetLength() const
-{
-  for (len = 0; len < GetSize(); len++)
-    if (((WORD *)theArray)[len] == 0)
-      break;
-  return len;
-}
-#endif
-
-
 PString PString::operator+(const char * cstr) const
 {
   if (cstr == NULL)
@@ -1396,8 +1489,8 @@ PString PString::operator+(const char * cstr) const
   PINDEX alen = strlen(cstr)+1;
   PString str;
   str.SetSize(olen+alen);
-  PSTRING_MOVE(str.theArray, 0, theArray, 0, olen);
-  PSTRING_COPY(str.theArray+olen, cstr, alen);
+  memmove(str.theArray, theArray, olen);
+  memcpy(str.theArray+olen, cstr, alen);
   return str;
 }
 
@@ -1407,7 +1500,7 @@ PString PString::operator+(char c) const
   PINDEX olen = GetLength();
   PString str;
   str.SetSize(olen+2);
-  PSTRING_MOVE(str.theArray, 0, theArray, 0, olen);
+  memmove(str.theArray, theArray, olen);
   str.theArray[olen] = c;
   return str;
 }
@@ -1421,7 +1514,7 @@ PString & PString::operator+=(const char * cstr)
   PINDEX olen = GetLength();
   PINDEX alen = strlen(cstr)+1;
   SetSize(olen+alen);
-  PSTRING_COPY(theArray+olen, cstr, alen);
+  memcpy(theArray+olen, cstr, alen);
   return *this;
 }
 
@@ -1448,10 +1541,10 @@ PString PString::operator&(const char * cstr) const
   PString str;
   PINDEX space = olen > 0 && theArray[olen-1]!=' ' && *cstr!=' ' ? 1 : 0;
   str.SetSize(olen+alen+space);
-  PSTRING_MOVE(str.theArray, 0, theArray, 0, olen);
+  memmove(str.theArray, theArray, olen);
   if (space != 0)
     str.theArray[olen] = ' ';
-  PSTRING_COPY(str.theArray+olen+space, cstr, alen);
+  memcpy(str.theArray+olen+space, cstr, alen);
   return str;
 }
 
@@ -1462,7 +1555,7 @@ PString PString::operator&(char c) const
   PString str;
   PINDEX space = olen > 0 && theArray[olen-1] != ' ' && c != ' ' ? 1 : 0;
   str.SetSize(olen+2+space);
-  PSTRING_MOVE(str.theArray, 0, theArray, 0, olen);
+  memmove(str.theArray, theArray, olen);
   if (space != 0)
     str.theArray[olen] = ' ';
   str.theArray[olen+space] = c;
@@ -1483,7 +1576,7 @@ PString & PString::operator&=(const char * cstr)
   SetSize(olen+alen+space);
   if (space != 0)
     theArray[olen] = ' ';
-  PSTRING_COPY(theArray+olen+space, cstr, alen);
+  memcpy(theArray+olen+space, cstr, alen);
   return *this;
 }
 
@@ -1511,7 +1604,7 @@ void PString::Delete(PINDEX start, PINDEX len)
   if (len > slen - start)
     SetAt(start, '\0');
   else
-    PSTRING_MOVE(theArray, start, theArray, start+len, slen-start-len+1);
+    memmove(theArray+start, theArray+start+len, slen-start-len+1);
   MakeMinimumSize();
 }
 
@@ -1589,6 +1682,24 @@ BOOL PString::operator*=(const char * cstr) const
 }
 
 
+PObject::Comparison PString::NumCompare(const PString & str, PINDEX count, PINDEX offset) const
+{
+  PINDEX len = str.GetLength();
+  if (count > len)
+    count = len;
+  return InternalCompare(offset, count, str);
+}
+
+
+PObject::Comparison PString::NumCompare(const char * cstr, PINDEX count, PINDEX offset) const
+{
+  PINDEX len = ::strlen(cstr);
+  if (count > len)
+    count = len;
+  return InternalCompare(offset, count, cstr);
+}
+
+
 PObject::Comparison PString::InternalCompare(PINDEX offset, char c) const
 {
   char ch = theArray[offset];
@@ -1811,9 +1922,9 @@ void PString::Splice(const char * cstr, PINDEX pos, PINDEX len)
     if (clen > len)
       SetSize(newlen+1);
     if (pos+len < slen)
-      PSTRING_MOVE(theArray, pos+clen, theArray, pos+len, slen-pos-len+1);
+      memmove(theArray+pos+clen, theArray+pos+len, slen-pos-len+1);
     if (clen > 0)
-      PSTRING_COPY(theArray+pos, cstr, clen);
+      memcpy(theArray+pos, cstr, clen);
     theArray[newlen] = '\0';
   }
 }
@@ -1916,12 +2027,9 @@ PString PString::RightTrim() const
     rpos--;
   }
 
-#if defined(P_MACOSX) || defined(P_MACOS)	// make Apple gnu compiler happy
-  PString retval(theArray, rpos - theArray);
+  // make Apple & Tornado gnu compiler happy
+  PString retval(theArray, rpos - theArray + 1);
   return retval;
-#else
-  return PString(theArray, rpos - theArray);
-#endif
 }
 
 
@@ -1992,21 +2100,123 @@ double PString::AsReal() const
 }
 
 
+PWORDArray PString::AsUCS2() const
+{
+#ifdef P_HAS_G_CONVERT
+
+  gsize g_len = 0;
+  gchar * g_ucs2 = g_convert(theArray, GetSize()-1, "UCS-2", "UTF-8", 0, &g_len, 0);
+  if (g_ucs2 == NULL)
+    return PWORDArray();
+
+  PWORDArray ucs2((const WORD *)g_ucs2, (PINDEX)g_len);
+  g_free(g_ucs2)
+  return ucs2;
+
+#else
+
+  PWORDArray ucs2(GetSize()); // Always bigger than required
+
+  PINDEX count = 0;
+  PINDEX i = 0;
+  PINDEX length = GetSize()-1;
+  while (i < length) {
+    int c = theArray[i];
+    if ((c&0x80) == 0)
+      ucs2[count++] = (BYTE)theArray[i++];
+    else if ((c&0xe0) == 0xc0) {
+      if (i < length-1)
+        ucs2[count++] = (WORD)(((theArray[i  ]&0x1f)<<6)|
+                                (theArray[i+1]&0x3f));
+      i += 2;
+    }
+    else if ((c&0xf0) == 0xe0) {
+      if (i < length-2)
+        ucs2[count++] = (WORD)(((theArray[i  ]&0x0f)<<12)|
+                               ((theArray[i+1]&0x3f)<< 6)|
+                                (theArray[i+2]&0x3f));
+      i += 3;
+    }
+    else {
+      if ((c&0xf8) == 0xf0)
+        i += 4;
+      else if ((c&0xfc) == 0xf8)
+        i += 5;
+      else
+        i += 6;
+      if (i <= length)
+        ucs2[count++] = 0xffff;
+    }
+  }
+
+  ucs2.SetSize(count);
+  return ucs2;
+
+#endif
+}
+
+
+void PString::InternalFromUCS2(const WORD * ptr, PINDEX len)
+{
+  if (ptr == NULL || len == 0) {
+    *this = Empty();
+    return;
+  }
+
+#ifdef P_HAS_G_CONVERT
+
+  gsize g_len = 0;
+  gchar * g_utf8 = g_convert(ptr, len, "UTF-8", "UCS-2", 0, &g_len, 0);
+  if (g_utf8 == NULL) {
+    *this = Empty();
+    return;
+  }
+
+  SetSize(&g_len);
+  memcpy(theArray, g_char, g_len);
+  g_free(g_utf8);
+
+#else
+
+  PINDEX i;
+  PINDEX count = 1;
+  for (i = 0; i < len; i++) {
+    if (ptr[i] < 0x80)
+      count++;
+    else if (ptr[i] < 0x800)
+      count += 2;
+    else
+      count += 3;
+  }
+  SetSize(count);
+
+  count = 0;
+  for (i = 0; i < len; i++) {
+    unsigned v = *ptr++;
+    if (v < 0x80)
+      theArray[count++] = (char)v;
+    else if (v < 0x800) {
+      theArray[count++] = (char)(0xc0+(v>>6));
+      theArray[count++] = (char)(0x80+(v&0x3f));
+    }
+    else {
+      theArray[count++] = (char)(0xd0+(v>>12));
+      theArray[count++] = (char)(0x80+((v>>6)&0x3f));
+      theArray[count++] = (char)(0x80+(v&0x3f));
+    }
+  }
+
+#endif
+}
+
+
 PBYTEArray PString::ToPascal() const
 {
   PINDEX len = GetLength();
   PAssert(len < 256, "Cannot convert to PASCAL string");
   BYTE buf[256];
   buf[0] = (BYTE)len;
-#ifdef PHAS_UNICODE
-  WORD * ptr = (WORD *)theArray;
-  while (len > 0) {
-    buf[len] = (BYTE)(*ptr < 256 ? *ptr : 255);
-    len--;
-  }
-#else
   memcpy(&buf[1], theArray, len);
-#endif
   return PBYTEArray(buf, len+1);
 }
 
@@ -2049,12 +2259,20 @@ PString & PString::sprintf(const char * fmt, ...)
 
 PString & PString::vsprintf(const char * fmt, va_list arg)
 {
+#ifdef P_TORNADO
+  // The library provided with tornado 2.0 does not have the implementation
+  // for vsnprintf
+  // as workaround, just use a array size of 2000
+  PAssert(SetSize(2000), POutOfMemory);
+  ::vsprintf(theArray, fmt, arg);
+#else
   PINDEX len = theArray != NULL ? GetLength() : 0;
   PINDEX size = 0;
   do {
     size += 1000;
     PAssert(SetSize(size), POutOfMemory);
   } while (_vsnprintf(theArray+len, size-len, fmt, arg) == -1);
+#endif // P_TORNADO
 
   PAssert(MakeMinimumSize(), POutOfMemory);
   return *this;
@@ -2115,20 +2333,34 @@ PObject::Comparison PCaselessString::InternalCompare(
 
 ///////////////////////////////////////////////////////////////////////////////
 
+PStringStream::Buffer::Buffer(PStringStream & str, PINDEX size)
+  : string(str),
+    fixedBufferSize(size != 0)
+{
+  string.SetMinSize(size != 0 ? size : 256);
+  sync();
+}
+
+
 int PStringStream::Buffer::overflow(int c)
 {
   if (pptr() >= epptr()) {
+    if (fixedBufferSize)
+      return EOF;
+
     int gpos = gptr() - eback();
     int ppos = pptr() - pbase();
-    char * newptr = string->GetPointer(string->GetSize() + 32);
-    setp(newptr, newptr + string->GetSize() - 1);
+    char * newptr = string.GetPointer(string.GetSize() + 32);
+    setp(newptr, newptr + string.GetSize() - 1);
     pbump(ppos);
     setg(newptr, newptr + gpos, newptr + ppos);
   }
+
   if (c != EOF) {
     *pptr() = (char)c;
     pbump(1);
   }
+
   return 0;
 }
 
@@ -2141,10 +2373,10 @@ int PStringStream::Buffer::underflow()
 
 int PStringStream::Buffer::sync()
 {
-  char * base = string->GetPointer();
-  PINDEX len = string->GetLength();
+  char * base = string.GetPointer();
+  PINDEX len = string.GetLength();
   setg(base, base, base + len);
-  setp(base, base + string->GetSize() - 1);
+  setp(base, base + string.GetSize() - 1);
   pbump(len);
   return 0;
 }
@@ -2156,7 +2388,7 @@ streampos PStringStream::Buffer::seekoff(streamoff off,
                                  ios::seek_dir dir, int mode)
 #endif
 {
-  int len = string->GetLength();
+  int len = string.GetLength();
   int gpos = gptr() - eback();
   int ppos = pptr() - pbase();
   char * newgptr;
@@ -2196,7 +2428,7 @@ streampos PStringStream::Buffer::seekoff(streamoff off,
       break;
 
     default:
-      PAssertAlways(PInvalidParameter);
+      PAssertAlways2(string.GetClass(), PInvalidParameter);
       newgptr = gptr();
       newpptr = pptr();
   }
@@ -2211,26 +2443,39 @@ streampos PStringStream::Buffer::seekoff(streamoff off,
 }
 
 
+#ifdef _MSC_VER
+#pragma warning(disable:4355)
+#endif
+
 PStringStream::PStringStream()
-    : iostream(cout.rdbuf())
+  : iostream(new PStringStream::Buffer(*this, 0))
+{
+}
+
+
+PStringStream::PStringStream(PINDEX fixedBufferSize)
+  : iostream(new PStringStream::Buffer(*this, fixedBufferSize))
 {
-  init(new PStringStream::Buffer(this));
 }
 
 
 PStringStream::PStringStream(const PString & str)
-    : PString(str), iostream(cout.rdbuf())
+  : PString(str),
+    iostream(new PStringStream::Buffer(*this, 0))
 {
-  init(new PStringStream::Buffer(this));
 }
 
 
 PStringStream::PStringStream(const char * cstr)
-  : PString(cstr), iostream(cout.rdbuf())
+  : PString(cstr),
+    iostream(new PStringStream::Buffer(*this, 0))
 {
-  init(new PStringStream::Buffer(this));
 }
 
+#ifdef _MSC_VER
+#pragma warning(default:4355)
+#endif
+
 
 PStringStream::~PStringStream()
 {
@@ -2253,7 +2498,15 @@ PStringArray::PStringArray(PINDEX count, char const * const * strarr, BOOL casel
   if (count == 0)
     return;
 
-  PAssertNULL(strarr);
+  if (PAssertNULL(strarr) == NULL)
+    return;
+
+  if (count == P_MAX_INDEX) {
+    count = 0;
+    while (strarr[count] != NULL)
+      count++;
+  }
+
   SetSize(count);
   for (PINDEX i = 0; i < count; i++) {
     PString * newString;
@@ -2266,6 +2519,13 @@ PStringArray::PStringArray(PINDEX count, char const * const * strarr, BOOL casel
 }
 
 
+PStringArray::PStringArray(const PString & str)
+{
+  SetSize(1);
+  (*theArray)[0] = new PString(str);
+}
+
+
 PStringArray::PStringArray(const PStringList & list)
 {
   SetSize(list.GetSize());
@@ -2282,6 +2542,16 @@ PStringArray::PStringArray(const PSortedStringList & list)
 }
 
 
+void PStringArray::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    AppendString(str);
+  }
+}
+
+
 PString & PStringArray::operator[](PINDEX index)
 {
   PASSERTINDEX(index);
@@ -2292,6 +2562,40 @@ PString & PStringArray::operator[](PINDEX index)
 }
 
 
+char ** PStringArray::ToCharArray(PCharArray * storage) const
+{
+  PINDEX i;
+
+  PINDEX mySize = GetSize();
+  PINDEX storageSize = (mySize+1)*sizeof(char *);
+  for (i = 0; i < mySize; i++)
+    storageSize += (*this)[i].GetLength()+1;
+
+  char ** storagePtr;
+  if (storage != NULL)
+    storagePtr = (char **)storage->GetPointer(storageSize);
+  else
+    storagePtr = (char **)malloc(storageSize);
+
+  if (storagePtr == NULL)
+    return NULL;
+
+  char * strPtr = (char *)&storagePtr[GetSize()+1];
+
+  for (i = 0; i < mySize; i++) {
+    storagePtr[i] = strPtr;
+    const PString & str = (*this)[i];
+    PINDEX len = str.GetLength()+1;
+    memcpy(strPtr, (const char *)str, len);
+    strPtr += len;
+  }
+
+  storagePtr[i] = NULL;
+
+  return storagePtr;
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PStringList::PStringList(PINDEX count, char const * const * strarr, BOOL caseless)
@@ -2311,6 +2615,12 @@ PStringList::PStringList(PINDEX count, char const * const * strarr, BOOL caseles
 }
 
 
+PStringList::PStringList(const PString & str)
+{
+  AppendString(str);
+}
+
+
 PStringList::PStringList(const PStringArray & array)
 {
   for (PINDEX i = 0; i < array.GetSize(); i++)
@@ -2334,6 +2644,16 @@ PStringList & PStringList::operator += (const PStringList & v)
 }
 
 
+void PStringList::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    AppendString(str);
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PSortedStringList::PSortedStringList(PINDEX count,
@@ -2355,6 +2675,12 @@ PSortedStringList::PSortedStringList(PINDEX count,
 }
 
 
+PSortedStringList::PSortedStringList(const PString & str)
+{
+  AppendString(str);
+}
+
+
 PSortedStringList::PSortedStringList(const PStringArray & array)
 {
   for (PINDEX i = 0; i < array.GetSize(); i++)
@@ -2369,6 +2695,61 @@ PSortedStringList::PSortedStringList(const PStringList & list)
 }
 
 
+
+void PSortedStringList::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    AppendString(str);
+  }
+}
+
+
+PINDEX PSortedStringList::GetNextStringsIndex(const PString & str) const
+{
+  PINDEX len = str.GetLength();
+
+  info->lastIndex = InternalStringSelect(str, len, info->root, info->lastElement);
+
+  if (info->lastIndex != 0) {
+    Element * prev;
+    while ((prev = info->lastElement->Predecessor()) != &Element::nil &&
+                    ((PString *)prev->data)->NumCompare(str, len) >= EqualTo) {
+      info->lastElement = prev;
+      info->lastIndex--;
+    }
+  }
+
+  return info->lastIndex;
+}
+
+
+PINDEX PSortedStringList::InternalStringSelect(const char * str,
+                                               PINDEX len,
+                                               Element * thisElement,
+                                               Element * & lastElement)
+{
+  if (thisElement == &Element::nil)
+    return 0;
+
+  switch (((PString *)thisElement->data)->NumCompare(str, len)) {
+    case PObject::LessThan :
+    {
+      PINDEX index = InternalStringSelect(str, len, thisElement->right, lastElement);
+      return thisElement->left->subTreeSize + index + 1;
+    }
+
+    case PObject::GreaterThan :
+      return InternalStringSelect(str, len, thisElement->left, lastElement);
+
+    default :
+      lastElement = thisElement;
+      return thisElement->left->subTreeSize;
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PStringSet::PStringSet(PINDEX count, char const * const * strarr, BOOL caseless)
@@ -2386,6 +2767,22 @@ PStringSet::PStringSet(PINDEX count, char const * const * strarr, BOOL caseless)
 }
 
 
+PStringSet::PStringSet(const PString & str)
+{
+  Include(str);
+}
+
+
+void PStringSet::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    Include(str);
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 POrdinalToString::POrdinalToString(PINDEX count, const Initialiser * init)
@@ -2397,6 +2794,21 @@ POrdinalToString::POrdinalToString(PINDEX count, const Initialiser * init)
 }
 
 
+void POrdinalToString::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    POrdinalKey key;
+    char equal;
+    PString str;
+    strm >> key >> ws >> equal >> str;
+    if (equal != '=')
+      SetAt(key, PString::Empty());
+    else
+      SetAt(key, str.Mid(equal+1));
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PStringToOrdinal::PStringToOrdinal(PINDEX count,
@@ -2413,6 +2825,20 @@ PStringToOrdinal::PStringToOrdinal(PINDEX count,
 }
 
 
+void PStringToOrdinal::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    PINDEX equal = str.FindLast('=');
+    if (equal == P_MAX_INDEX)
+      SetAt(str, 0);
+    else
+      SetAt(str.Left(equal), str.Mid(equal+1).AsInteger());
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PStringToString::PStringToString(PINDEX count,
@@ -2436,6 +2862,20 @@ PStringToString::PStringToString(PINDEX count,
 }
 
 
+void PStringToString::ReadFrom(istream & strm)
+{
+  while (strm.good()) {
+    PString str;
+    strm >> str;
+    PINDEX equal = str.Find('=');
+    if (equal == P_MAX_INDEX)
+      SetAt(str, PString::Empty());
+    else
+      SetAt(str.Left(equal), str.Mid(equal+1));
+  }
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 
 PRegularExpression::PRegularExpression()
