@@ -27,6 +27,73 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: osutils.cxx,v $
+ * Revision 1.205  2003/07/24 22:01:42  dereksmithies
+ * Add fixes from Peter Nixon  for fixing install problems. Thanks.
+ *
+ * Revision 1.204  2003/03/27 07:27:44  robertj
+ * Added function to get a bunch of arguments as a string array.
+ *
+ * Revision 1.203  2003/01/24 10:21:06  robertj
+ * Fixed issues in RTEMS support, thanks Vladimir Nesic
+ *
+ * Revision 1.202  2002/12/18 05:31:31  robertj
+ * Moved PTimeInterval::GetInterval() to common code.
+ *
+ * Revision 1.201  2002/12/11 03:23:27  robertj
+ * Fixed deadlock in read/write mutex, at price of not having seemless upgrading
+ *   of read lock to write lock. There is now a window in which some other
+ *   thread may gain write lock from the thread that was trying to upgrade.
+ *
+ * Revision 1.200  2002/12/10 02:39:07  robertj
+ * Avoid odd trace output to stderr before trace file is set.
+ *
+ * Revision 1.199  2002/10/09 00:46:19  robertj
+ * Changed PThread::Create() so does not return PThread pointer if the thread
+ *   has been created auto-delete, the pointer is extremely dangerous to use
+ *   as it could be deleted at any time, so to remove temptation ...
+ *
+ * Revision 1.198  2002/10/08 03:35:15  robertj
+ * Fixed BSD warning
+ *
+ * Revision 1.197  2002/10/04 08:21:26  robertj
+ * Changed read/write mutex so can be called by same thread without deadlock.
+ *
+ * Revision 1.196  2002/07/30 02:55:09  craigs
+ * Added program start time to PProcess
+ *
+ * Revision 1.195  2002/06/27 06:38:58  robertj
+ * Changes to remove memory leak display for things that aren't memory leaks.
+ *
+ * Revision 1.194  2002/06/15 02:16:36  robertj
+ * Fixed bug (from rev 1.190) so can now use PTimer::Reset() after the timer
+ *   had previously expired (resetTiem was being zeroed), thanks Ted Szoczei
+ *
+ * Revision 1.193  2002/06/05 12:29:15  craigs
+ * Changes for gcc 3.1
+ *
+ * Revision 1.192  2002/06/04 00:25:31  robertj
+ * Fixed incorrectly initialised trace indent, thanks Artis Kugevics
+ *
+ * Revision 1.191  2002/05/31 04:10:44  robertj
+ * Fixed missing mutex in PTrace::SetStream, thanks Federico Pinna
+ *
+ * Revision 1.190  2002/05/28 13:05:26  robertj
+ * Fixed PTimer::SetInterval so it restarts timer as per operator=()
+ *
+ * Revision 1.189  2002/05/22 00:42:03  craigs
+ * Added GMTTime flag to tracing options
+ *
+ * Revision 1.188  2002/05/01 03:45:09  robertj
+ * Added initialisation of PreadWriteMutex and changed slightly to agree
+ *   with the text book definition of a semaphore for one of the mutexes.
+ *
+ * Revision 1.187  2002/04/30 06:21:38  robertj
+ * Fixed PReadWriteMutex class to implement text book algorithm!
+ *
+ * Revision 1.186  2002/04/30 03:39:21  robertj
+ * Changed PTimer::Stop() so does not return until timer is REALLY stopped, in
+ *   particular when a possibly executing OnTimeout() function has completed.
+ *
  * Revision 1.185  2002/04/24 01:19:07  robertj
  * Added milliseconds to PTRACE output timestamp
  *
@@ -681,33 +748,36 @@ static PTimeInterval ApplicationStartTick = PTimer::Tick();
 
 void PTrace::SetStream(ostream * s)
 {
-#if !defined(__NUCLEUS_PLUS__)
-  PTraceStream = s != NULL ? s : &cerr;
-#else
-
-#ifdef __NUCLEUS_PLUS__ 
-  PTraceStream = s;
+#ifndef __NUCLEUS_PLUS__
+  if (s == NULL)
+    s = &cerr;
 #endif
 
-#endif
+  if (PTraceMutex == NULL)
+    PTraceStream = s;
+  else {
+    PTraceMutex->Wait();
+    PTraceStream = s;
+    PTraceMutex->Signal();
+  }
 }
 
 
 void PTrace::Initialise(unsigned level, const char * filename, unsigned options)
 {
+  // If we have a tracing version, then open trace file and set modes
 #if PTRACING
   PProcess & process = PProcess::Current();
 #endif
 
-  // If we have a tracing version, then open trace file and set modes
-  PTraceOptions = options;
-  PTraceLevelThreshold = level;
+#if PMEMORY_CHECK
+  int ignoreAllocations = -1;
+#endif
 
   if (filename != NULL) {
 #if PMEMORY_CHECK
-    BOOL ignoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE);
+    ignoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE) ? 1 : 0;
 #endif
-
     PTextFile * traceOutput;
     if (options & AppendToFile) {
       traceOutput = new PTextFile(filename, PFile::ReadWrite);
@@ -715,9 +785,6 @@ void PTrace::Initialise(unsigned level, const char * filename, unsigned options)
     } else 
       traceOutput = new PTextFile(filename, PFile::WriteOnly);
 
-#if PMEMORY_CHECK
-    PMemoryHeap::SetIgnoreAllocations(ignoreAllocations);
-#endif
     if (traceOutput->IsOpen())
       PTrace::SetStream(traceOutput);
     else {
@@ -726,12 +793,20 @@ void PTrace::Initialise(unsigned level, const char * filename, unsigned options)
     }
   }
 
+  PTraceOptions = options;
+  PTraceLevelThreshold = level;
+
   PTRACE(1, process.GetName()
          << "\tVersion " << process.GetVersion(TRUE)
          << " by " << process.GetManufacturer()
          << " on " << process.GetOSClass() << ' ' << process.GetOSName()
          << " (" << process.GetOSVersion() << '-' << process.GetOSHardware()
          << ") at " << PTime().AsString("yyyy/M/d h:mm:ss.uuu"));
+
+#if PMEMORY_CHECK
+  if (ignoreAllocations >= 0)
+    PMemoryHeap::SetIgnoreAllocations(ignoreAllocations != 0);
+#endif
 }
 
 
@@ -792,7 +867,7 @@ ostream & PTrace::Begin(unsigned level, const char * fileName, int lineNum)
   else {
     if ((PTraceOptions&DateAndTime) != 0) {
       PTime now;
-      *PTraceStream << now.AsString("yyyy/MM/dd hh:mm:ss.uuu\t");
+      *PTraceStream << now.AsString("yyyy/MM/dd hh:mm:ss.uuu\t", (PTraceOptions&GMTTime) ? PTime::GMT : PTime::Local);
     }
 
     if ((PTraceOptions&Timestamp) != 0)
@@ -803,11 +878,7 @@ ostream & PTrace::Begin(unsigned level, const char * fileName, int lineNum)
       if (thread == NULL)
         *PTraceStream << "ThreadID=0x"
                       << setfill('0') << hex << setw(8)
-#ifdef _WIN32
-                      << GetCurrentThreadId()
-#elif defined(P_PTHREADS)
-                      << pthread_self()
-#endif
+                      << PThread::GetCurrentThreadId()
                       << setfill(' ') << dec;
       else {
         PString name = thread->GetThreadName();
@@ -821,7 +892,7 @@ ostream & PTrace::Begin(unsigned level, const char * fileName, int lineNum)
 
     if ((PTraceOptions&ThreadAddress) != 0)
       *PTraceStream << hex << setfill('0')
-                    << setw(7) << (unsigned)PThread::Current()
+                    << setw(7) << (void *)PThread::Current()
                     << dec << setfill(' ') << '\t';
   }
 
@@ -854,12 +925,17 @@ ostream & PTrace::End(ostream & s)
      stderr is used the unitbuf flag causes the out_waiting() not to work so we 
      must suffer with blank lines in that case.
    */
-  if ((s.flags()&ios::unitbuf) != 0 || s.rdbuf()->out_waiting() > 0) {
+#if __GNUC__ >= 3
+  s << endl;
+#else
+  ::streambuf & rb = *s.rdbuf();
+  if (((s.flags()&ios::unitbuf) != 0) || (rb.out_waiting() > 0)) {
     if ((PTraceOptions&SystemLogStream) != 0)
       s.flush();
     else
       s << endl;
   }
+#endif
 
   PTraceMutex->Signal();
 
@@ -911,6 +987,21 @@ void PDirectory::CloneContents(const PDirectory * d)
 }
 
 
+///////////////////////////////////////////////////////////////////////////////
+// PTimeInterval
+
+DWORD PTimeInterval::GetInterval() const
+{
+  if (milliseconds <= 0)
+    return 0;
+
+  if (milliseconds >= UINT_MAX)
+    return UINT_MAX;
+
+  return (DWORD)milliseconds;
+}
+
+
 ///////////////////////////////////////////////////////////////////////////////
 // PTimer
 
@@ -977,6 +1068,18 @@ PTimer::~PTimer()
 }
 
 
+void PTimer::SetInterval(PInt64 milliseconds,
+                         long seconds,
+                         long minutes,
+                         long hours,
+                         int days)
+{
+  timerList->processingMutex.Wait();
+  resetTime.SetInterval(milliseconds, seconds, minutes, hours, days);
+  StartRunning(oneshot);
+}
+
+
 void PTimer::RunContinuous(const PTimeInterval & time)
 {
   timerList->processingMutex.Wait();
@@ -1005,8 +1108,16 @@ void PTimer::Stop()
 {
   timerList->processingMutex.Wait();
   state = Stopped;
-  SetInterval(0);
+  milliseconds = 0;
+  BOOL isCurrentTimer = this == timerList->currentTimer;
   timerList->processingMutex.Signal();
+
+  // Make sure that the OnTimeout for this timer has completed before
+  // retruning from Stop() function,
+  if (isCurrentTimer) {
+    timerList->inTimeoutMutex.Wait();
+    timerList->inTimeoutMutex.Signal();
+  }
 }
 
 
@@ -1066,7 +1177,7 @@ void PTimer::Process(const PTimeInterval & delta, PTimeInterval & minTimeLeft)
       }
       else {
         if (oneshot) {
-          SetInterval(0);
+          milliseconds = 0;
           state = Stopped;
         }
         else {
@@ -1440,6 +1551,34 @@ PString PArgList::GetOptionStringByIndex(PINDEX idx, const char * dflt) const
 }
 
 
+PStringArray PArgList::GetParameters(PINDEX first, PINDEX last) const
+{
+  PStringArray array;
+
+  last += shift;
+  if (last < 0)
+    return array;
+
+  if (last >= parameterIndex.GetSize())
+    last = parameterIndex.GetSize()-1;
+
+  first += shift;
+  if (first < 0)
+    first = 0;
+
+  if (first > last)
+    return array;
+
+  array.SetSize(last-first+1);
+
+  PINDEX idx = 0;
+  while (first <= last)
+    array[idx++] = argumentArray[parameterIndex[first++]];
+
+  return array;
+}
+
+
 PString PArgList::GetParameter(PINDEX num) const
 {
   int idx = shift+(int)num;
@@ -1634,6 +1773,7 @@ PProcess::PProcess(const char * manuf, const char * name,
   PMemoryHeap::SetIgnoreAllocations(ignoreAllocations);
 #endif
 
+#ifndef P_RTEMS
   if (p_argv != 0 && p_argc > 0) {
     arguments.SetArgs(p_argc-1, p_argv+1);
 
@@ -1647,6 +1787,10 @@ PProcess::PProcess(const char * manuf, const char * name,
     if (productName.IsEmpty())
       productName = executableFile.GetTitle().ToLower();
   }
+#else
+  cout << "Enter program arguments:\n";
+  arguments.ReadFrom(cin);
+#endif
 
   InitialiseProcessThread();
 
@@ -1726,6 +1870,10 @@ void PProcess::SetThreadName(const PString & /*name*/)
 {
 }
 
+PTime PProcess::GetStartTime() const
+{ 
+  return programStartTime; 
+}
 
 PString PProcess::GetVersion(BOOL full) const
 {
@@ -1773,7 +1921,19 @@ PThread * PThread::Create(const PNotifier & notifier,
                           const PString & threadName,
                           PINDEX stackSize)
 {
-  return new PSimpleThread(notifier, parameter, deletion, priorityLevel, threadName, stackSize);
+  PThread * thread = new PSimpleThread(notifier,
+                                       parameter,
+                                       deletion,
+                                       priorityLevel,
+                                       threadName,
+                                       stackSize);
+  if (deletion != AutoDeleteThread)
+    return thread;
+
+  // Do not return a pointer to the thread if it is auto-delete as this
+  // pointer is extremely dangerous to use, it could be deleted at any moment
+  // from now on so using the pointer could crash the program.
+  return NULL;
 }
 
 
@@ -1809,9 +1969,7 @@ void PThread::InitialiseProcessThread()
   stackBase = NULL;
   link = this;
   ((PProcess*)this)->currentThread = this;
-#if PTRACING
   traceBlockIndentLevel = 0;
-#endif
 }
 
 
@@ -1825,9 +1983,7 @@ PThread::PThread(PINDEX stackSize,
   basePriority = priorityLevel;   // Threads user settable priority level
   dynamicPriority = 0;            // Run immediately
   suspendCount = 1;
-#if PTRACING
   traceBlockIndentLevel = 0;
-#endif
 
   AllocateStack(stackSize);
   PAssert(stackBase != NULL, "Insufficient near heap for thread");
@@ -2237,6 +2393,8 @@ PSyncPoint::PSyncPoint()
 
 #endif
 
+/////////////////////////////////////////////////////////////////////////////
+
 void PSyncPointAck::Signal()
 {
   PSyncPoint::Signal();
@@ -2257,6 +2415,8 @@ void PSyncPointAck::Acknowledge()
 }
 
 
+/////////////////////////////////////////////////////////////////////////////
+
 void PCondMutex::WaitCondition()
 {
   for (;;) {
@@ -2284,6 +2444,8 @@ void PCondMutex::OnWait()
 }
 
 
+/////////////////////////////////////////////////////////////////////////////
+
 PIntCondMutex::PIntCondMutex(int val, int targ, Operation op)
 {
   value = val;
@@ -2374,31 +2536,205 @@ PIntCondMutex & PIntCondMutex::operator-=(int dec)
 }
 
 
+/////////////////////////////////////////////////////////////////////////////
+
+PReadWriteMutex::PReadWriteMutex()
+  : readerSemaphore(1, 1),
+    writerSemaphore(1, 1)
+{
+  readerCount = 0;
+  writerCount = 0;
+}
+
+
+PReadWriteMutex::Nest * PReadWriteMutex::GetNest() const
+{
+  PWaitAndSignal mutex(nestingMutex);
+  return nestedThreads.GetAt(POrdinalKey((PINDEX)PThread::GetCurrentThreadId()));
+}
+
+
+void PReadWriteMutex::EndNest()
+{
+  nestingMutex.Wait();
+  nestedThreads.RemoveAt(POrdinalKey((PINDEX)PThread::GetCurrentThreadId()));
+  nestingMutex.Signal();
+}
+
+
+PReadWriteMutex::Nest & PReadWriteMutex::StartNest()
+{
+  POrdinalKey threadId = (PINDEX)PThread::GetCurrentThreadId();
+
+  nestingMutex.Wait();
+
+  Nest * nest = nestedThreads.GetAt(threadId);
+
+  if (nest == NULL) {
+    nest = new Nest;
+    nestedThreads.SetAt(threadId, nest);
+  }
+
+  nestingMutex.Signal();
+
+  return *nest;
+}
+
+
 void PReadWriteMutex::StartRead()
 {
+  // Get the nested thread info structure, create one it it doesn't exist
+  Nest & nest = StartNest();
+
+  // One more nested call to StartRead() by this thread, note this does not
+  // need to be mutexed as it is always in the context of a single thread.
+  nest.readerCount++;
+
+  // If this is the first call to StartRead() and there has not been a
+  // previous call to StartWrite() then actually do the text book read only
+  // lock, otherwise we leave it as just having incremented the reader count.
+  if (nest.readerCount == 1 && nest.writerCount == 0)
+    InternalStartRead();
+}
+
+
+void PReadWriteMutex::InternalStartRead()
+{
+  // Text book read only lock
+
   starvationPreventer.Wait();
+   readerSemaphore.Wait();
+    readerMutex.Wait();
+
+     readerCount++;
+     if (readerCount == 1)
+       writerSemaphore.Wait();
+
+    readerMutex.Signal();
+   readerSemaphore.Signal();
   starvationPreventer.Signal();
-  ++readers;
 }
 
 
 void PReadWriteMutex::EndRead()
 {
-  --readers;
+  // Get the nested thread info structure for the curent thread
+  Nest * nest = GetNest();
+
+  // If don't have an active read or write lock or there is a write lock but
+  // the StartRead() was never called, then assert and ignore call.
+  if (nest == NULL || nest->readerCount == 0) {
+    PAssertAlways("Unbalanced PReadWriteMutex::EndRead()");
+    return;
+  }
+
+  // One less nested lock by this thread, note this does not
+  // need to be mutexed as it is always in the context of a single thread.
+  nest->readerCount--;
+
+  // If this is a nested read or a write lock is present then we don't do the
+  // real unlock, the decrement is enough.
+  if (nest->readerCount > 0 || nest->writerCount > 0)
+    return;
+
+  // Do text book read lock
+  InternalEndRead();
+
+  // At this point all read and write locks are gone for this thread so we can
+  // reclaim the memory.
+  EndNest();
+}
+
+
+void PReadWriteMutex::InternalEndRead()
+{
+  // Text book read only unlock
+
+  readerMutex.Wait();
+
+  readerCount--;
+  if (readerCount == 0)
+    writerSemaphore.Signal();
+
+  readerMutex.Signal();
 }
 
 
 void PReadWriteMutex::StartWrite()
 {
-  starvationPreventer.Wait();
-  readers.WaitCondition();
+  // Get the nested thread info structure, create one it it doesn't exist
+  Nest & nest = StartNest();
+
+  // One more nested call to StartWrite() by this thread, note this does not
+  // need to be mutexed as it is always in the context of a single thread.
+  nest.writerCount++;
+
+  // If is a nested call to StartWrite() then simply return, the writer count
+  // increment is all we haev to do.
+  if (nest.writerCount > 1)
+    return;
+
+  // If have a read lock already in this thread then do the "real" unlock code
+  // but do not change the lock count, calls to EndRead() will now just
+  // decrement the count instead of doing the unlock (its already done!)
+  if (nest.readerCount > 0)
+    InternalEndRead();
+
+  // Note in this gap another thread could grab the write lock, thus
+
+  // Now do the text book write lock
+  writerMutex.Wait();
+
+  writerCount++;
+  if (writerCount == 1)
+    readerSemaphore.Wait();
+
+  writerMutex.Signal();
+
+  writerSemaphore.Wait();
 }
 
 
 void PReadWriteMutex::EndWrite()
 {
-  starvationPreventer.Signal();
-  readers.Signal();
+  // Get the nested thread info structure for the curent thread
+  Nest * nest = GetNest();
+
+  // If don't have an active read or write lock or there is a read lock but
+  // the StartWrite() was never called, then assert and ignore call.
+  if (nest == NULL || nest->writerCount == 0) {
+    PAssertAlways("Unbalanced PReadWriteMutex::EndWrite()");
+    return;
+  }
+
+  // One less nested lock by this thread, note this does not
+  // need to be mutexed as it is always in the context of a single thread.
+  nest->writerCount--;
+
+  // If this is a nested write lock then the decrement is enough and we
+  // don't do the actual write unlock.
+  if (nest->writerCount > 0)
+    return;
+
+  // Begin text book write unlock
+  writerSemaphore.Signal();
+
+  writerMutex.Wait();
+
+  writerCount--;
+  if (writerCount == 0)
+    readerSemaphore.Signal();
+
+  writerMutex.Signal();
+  // End of text book write unlock
+
+  // Now check to see if there was a read lock present for this thread, if so
+  // then reacquire the read lock (not changing the count) otherwise clean up the
+  // memory for the nested thread info structure
+  if (nest->readerCount > 0)
+    InternalStartRead();
+  else
+    EndNest();
 }
 
 
