@@ -27,6 +27,18 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: ptime.cxx,v $
+ * Revision 1.45  2003/08/14 02:07:11  dereksmithies
+ * Fix bug in AsString handling. Thanks Diego Tartara for pointing it out.
+ *
+ * Revision 1.44  2002/12/17 04:41:40  robertj
+ * Added ability to escape special characters.
+ *
+ * Revision 1.43  2002/12/10 05:22:16  robertj
+ * Fixed GNU warning
+ *
+ * Revision 1.42  2002/12/10 04:45:41  robertj
+ * Added support in PTime for ISO 8601 format.
+ *
  * Revision 1.41  2002/02/15 03:56:09  yurik
  * Warnings removed during compilation, patch courtesy of Jehan Bing, jehan@bravobrava.com
  *
@@ -405,16 +417,27 @@ PObject::Comparison PTime::Compare(const PObject & obj) const
 
 
 PString PTime::AsString(TimeFormat format, int zone) const
-{
-  if (format == RFC1123)
-    return AsString("wwwe, dd MMME yyyy hh:mm:ss z", zone);
+{ 
+  if (format >= NumTimeStrings)
+    return "Invalid format : " + AsString("yyyy-MM-dd T hh:mm:ss Z");
+
+  switch (format) {
+    case RFC1123 :
+      return AsString("wwwe, dd MMME yyyy hh:mm:ss z", zone);
+    case ShortISO8601 :
+      return AsString("yyyyMMddThhmmssZ");
+    case LongISO8601 :
+      return AsString("yyyy-MM-dd T hh:mm:ss Z");
+    default :
+      break;
+  }
 
   PString fmt, dsep;
 
   PString tsep = GetTimeSeparator();
   BOOL is12hour = GetTimeAMPM();
 
-  switch (format & 7) {
+  switch (format ) {
     case LongDateTime :
     case LongTime :
     case MediumDateTime :
@@ -442,7 +465,7 @@ PString PTime::AsString(TimeFormat format, int zone) const
       break;
   }
 
-  switch (format & 7) {
+  switch (format ) {
     case LongDateTime :
     case MediumDateTime :
     case ShortDateTime :
@@ -453,7 +476,7 @@ PString PTime::AsString(TimeFormat format, int zone) const
       break;
   }
 
-  switch (format & 7) {
+  switch (format ) {
     case LongDateTime :
     case LongDate :
       fmt += "wwww ";
@@ -607,15 +630,20 @@ PString PTime::AsString(const char * format, int zone) const
         break;
 
       case 'z' :
-        while (*++format == 'z')
-          ;
-        if (zone == 0)
-          str << "GMT";
+      case 'Z' :
+        if (zone == 0) {
+          if (*format == 'Z')
+            str << 'Z';
+          else
+            str << "GMT";
+        }
         else {
           str << (zone < 0 ? '-' : '+');
           zone = PABS(zone);
           str << setw(2) << (zone/60) << setw(2) << (zone%60);
         }
+        while (toupper(*++format) == 'z')
+          ;
         break;
 
       case 'u' :
@@ -637,6 +665,10 @@ PString PTime::AsString(const char * format, int zone) const
         }
         break;
 
+      case '\\' :
+        format++;
+        // Escaped character, put straight through to output string
+
       default :
         str << *format++;
     }
