@@ -26,6 +26,22 @@
  *		   Mark Cooke (mpc@star.sr.bham.ac.uk)
  *
  * $Log: vconvert.cxx,v $
+ * Revision 1.34  2003/06/14 02:57:36  rjongbloed
+ * REmoved redundent parameter, grey scale does not have rgb increment!
+ *
+ * Revision 1.33  2003/06/09 22:37:24  dereksmithies
+ * Fix from Clive Nicolson to make b/w colour conversions work (i.e. grey palette).
+ * many thanks!
+ *
+ * Revision 1.32  2003/04/03 09:28:37  robertj
+ * Added reversed RGB byte order versions (BGR24), thanks Damien Sandras
+ *
+ * Revision 1.31  2003/03/31 11:30:14  rogerh
+ * make 'cb' and 'cr' contain the values that their name implies.
+ *
+ * Revision 1.30  2002/09/01 23:00:05  dereks
+ * Fix noise in flipped RGB image. Thanks Alex Phtahov.
+ *
  * Revision 1.29  2002/02/26 02:23:21  dereks
  * Reduced verbosity in PTRACE output for when video is enabled.
  *
@@ -139,8 +155,14 @@
 
 static PColourConverterRegistration * RegisteredColourConvertersListHead = NULL;
 
+PSYNONYM_COLOUR_CONVERTER(Grey,   Grey);
+PSYNONYM_COLOUR_CONVERTER(GreyF,  GreyF);
 PSYNONYM_COLOUR_CONVERTER(RGB24,  RGB24);
+PSYNONYM_COLOUR_CONVERTER(RGB24F, RGB24F);
+PSYNONYM_COLOUR_CONVERTER(BGR24,  BGR24);
+PSYNONYM_COLOUR_CONVERTER(BGR24F, BGR24F);
 PSYNONYM_COLOUR_CONVERTER(RGB32,  RGB32);
+PSYNONYM_COLOUR_CONVERTER(RGB32F, RGB32F);
 PSYNONYM_COLOUR_CONVERTER(YUV411P,YUV411P);
 PSYNONYM_COLOUR_CONVERTER(YUV420P,YUV420P);
 PSYNONYM_COLOUR_CONVERTER(YUV420P,IYUV);
@@ -159,6 +181,22 @@ class PStandardColourConverter : public PColourConverter
       unsigned w, unsigned h
     ) : PColourConverter(srcFmt, dstFmt, w, h) { }
 
+    void GreytoYUV420PSameSize(
+      const BYTE * rgb,
+      BYTE * yuv,
+      BOOL flipVertical
+    ) const;
+    void GreytoYUV420PWithResize(
+      const BYTE * rgb,
+      BYTE * yuv,
+      BOOL flipVertical
+    ) const;
+    BOOL GreytoYUV420P(
+      const BYTE * rgb,
+      BYTE * yuv,
+      PINDEX * bytesReturned,
+      BOOL flipVertical
+    ) const;
     void RGBtoYUV420PSameSize(
       const BYTE * rgb,
       BYTE * yuv,
@@ -183,7 +221,8 @@ class PStandardColourConverter : public PColourConverter
       BYTE * rgb,
       PINDEX * bytesReturned,
       unsigned rgbIncrement,
-      BOOL flipVertical
+      BOOL flipVertical,
+      BOOL flipBR
     ) const;
     void ResizeYUV422(
       const BYTE * src,
@@ -278,7 +317,7 @@ BOOL PColourConverter::SetDstFrameSize(unsigned width, unsigned height, BOOL bSc
   dstFrameWidth  = width;
   dstFrameHeight = height;
   scaleNotCrop   = bScale;
-  
+
   dstFrameBytes = PVideoDevice::CalculateFrameBytes(dstFrameWidth, dstFrameHeight, dstColourFormat);
 
   PTRACE(6, "PColCnv\tSetDstFrameSize "
@@ -343,23 +382,152 @@ BOOL PSynonymColour::Convert(const BYTE *srcFrameBuffer,
 {
   if ((srcFrameWidth != dstFrameWidth) || (srcFrameHeight != dstFrameHeight))
     return FALSE;
-  
+
   if (srcFrameBuffer != dstFrameBuffer)
     memcpy(dstFrameBuffer, srcFrameBuffer, dstFrameBytes);
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
-  
+
   return TRUE;
 }
 
 ///////////////////////////////////////////////////////////////////////////////
 
-#define rgbtoyuv(b, g, r, y, u, v) \
-  y=(BYTE)(((int)30*r  +(int)59*g +(int)11*b)/100); \
-  u=(BYTE)(((int)-17*r  -(int)33*g +(int)50*b+12800)/100); \
-  v=(BYTE)(((int)50*r  -(int)42*g -(int)8*b+12800)/100); \
+#define BLACK_Y 0
+#define BLACK_U 128
+#define BLACK_V 128
 
+#define greytoy(r, y) y=r
+#define greytoyuv(r, y, u, v) greytoy(r,y); u=BLACK_U; v=BLACK_V
+
+void PStandardColourConverter::GreytoYUV420PSameSize(const BYTE * grey,
+						     BYTE * yuv,
+						     BOOL flip) const
+{
+  const unsigned planeSize = srcFrameWidth*srcFrameHeight;
+  const unsigned halfWidth = srcFrameWidth >> 1;
+
+  // get pointers to the data
+  BYTE * yplane  = yuv;
+  BYTE * uplane  = yuv + planeSize;
+  BYTE * vplane  = yuv + planeSize + (planeSize >> 2);
+  const BYTE * greyIndex = grey;
+
+  for (unsigned y = 0; y < srcFrameHeight; y++) {
+    BYTE * yline  = yplane + (y * srcFrameWidth);
+    BYTE * uline  = uplane + ((y >> 1) * halfWidth);
+    BYTE * vline  = vplane + ((y >> 1) * halfWidth);
+
+    if (flip)
+      greyIndex = grey + srcFrameWidth*(srcFrameHeight-1-y);
+
+    for (unsigned x = 0; x < srcFrameWidth; x+=2) {
+      greytoy(*greyIndex, *yline);
+      greyIndex++;
+      yline++;
+      greytoyuv(*greyIndex, *yline, *uline, *vline);
+      greyIndex++;
+      yline++;
+      uline++;
+      vline++;
+    }
+  }
+}
+
+
+// Simple crop/pad version.  Image aligned to top-left
+// and cropped / padded with black borders as required.
+void PStandardColourConverter::GreytoYUV420PWithResize(const BYTE * grey,
+						       BYTE * yuv,
+						       BOOL   flip) const
+{
+  int planeSize = dstFrameWidth*dstFrameHeight;
+  const int halfWidth = dstFrameWidth >> 1;
+  unsigned min_width, min_height;
+
+  min_width  = (dstFrameWidth  < srcFrameWidth)  ? dstFrameWidth  : srcFrameWidth;
+  min_height = (dstFrameHeight < srcFrameHeight) ? dstFrameHeight : srcFrameHeight;
+
+  // get pointers to the data
+  BYTE * yplane  = yuv;
+  BYTE * uplane  = yuv + planeSize;
+  BYTE * vplane  = yuv + planeSize + (planeSize >> 2);
+  const BYTE * greyIndex = grey;
+
+  for (unsigned y = 0; y < min_height; y++) 
+  {
+    BYTE * yline  = yplane + (y * dstFrameWidth);
+    BYTE * uline  = uplane + ((y >> 1) * halfWidth);
+    BYTE * vline  = vplane + ((y >> 1) * halfWidth);
+
+    if (flip)
+      greyIndex = grey + srcFrameWidth*(min_height-1-y);
+
+    for (unsigned x = 0; x < min_width; x+=2) 
+    {
+     greytoy(*greyIndex, *yline);
+     greyIndex++;
+     yline++;
+     greytoyuv(*greyIndex, *yline, *uline, *vline);
+     greyIndex++;
+     yline++;
+     uline++;
+     vline++;
+    }
+
+    // Crop if source width > dest width
+    if (srcFrameWidth > dstFrameWidth)
+      greyIndex += srcFrameWidth - dstFrameWidth;
+
+    // Pad if dest width < source width
+    if (dstFrameWidth > srcFrameWidth) {
+      memset(yline, BLACK_Y, dstFrameWidth - srcFrameWidth);
+      memset(uline, BLACK_U, (dstFrameWidth - srcFrameWidth)>>1);
+      memset(vline, BLACK_V, (dstFrameWidth - srcFrameWidth)>>1);
+    }
+  }
+
+  // Pad if dest height > source height
+  if (dstFrameHeight > srcFrameHeight) {
+    BYTE * yline  = yplane + (srcFrameHeight * dstFrameWidth);
+    BYTE * uline  = uplane + ((srcFrameHeight >> 1) * halfWidth);
+    BYTE * vline  = vplane + ((srcFrameHeight >> 1) * halfWidth);
+    unsigned fill = (dstFrameHeight - srcFrameHeight) * dstFrameWidth;
+
+    memset(yline, BLACK_Y, fill);
+    memset(uline, BLACK_U, fill >> 2);
+    memset(vline, BLACK_V, fill >> 2);
+  }
+}
+
+
+BOOL PStandardColourConverter::GreytoYUV420P(const BYTE * grey,
+					     BYTE * yuv,
+					     PINDEX * bytesReturned,
+					     BOOL flip) const
+{
+  if (grey == yuv)
+    return FALSE; // Cannot do in place conversion
+
+  if ((srcFrameWidth == dstFrameWidth) && (srcFrameHeight == dstFrameHeight)) 
+    GreytoYUV420PSameSize(grey, yuv, flip);
+  else
+    GreytoYUV420PWithResize(grey, yuv, flip);
+
+  if (bytesReturned != NULL)
+    *bytesReturned = dstFrameBytes;
+
+  return TRUE;
+}
+
+#define rgbtoy(b, g, r, y) \
+  y=(BYTE)(((int)30*r  +(int)59*g +(int)11*b)/100)
+
+#define rgbtoyuv(b, g, r, y, u, v) \
+  rgbtoy(b, g, r, y); \
+  u=(BYTE)(((int)-17*r  -(int)33*g +(int)50*b+12800)/100); \
+  v=(BYTE)(((int)50*r  -(int)42*g -(int)8*b+12800)/100)
 
 void PStandardColourConverter::RGBtoYUV420PSameSize(const BYTE * rgb,
                                                     BYTE * yuv,
@@ -368,7 +536,7 @@ void PStandardColourConverter::RGBtoYUV420PSameSize(const BYTE * rgb,
 {
   const unsigned planeSize = srcFrameWidth*srcFrameHeight;
   const unsigned halfWidth = srcFrameWidth >> 1;
-  
+
   // get pointers to the data
   BYTE * yplane  = yuv;
   BYTE * uplane  = yuv + planeSize;
@@ -384,7 +552,7 @@ void PStandardColourConverter::RGBtoYUV420PSameSize(const BYTE * rgb,
       rgbIndex = rgb + (srcFrameWidth*(srcFrameHeight-1-y)*rgbIncrement);
 
     for (unsigned x = 0; x < srcFrameWidth; x+=2) {
-      rgbtoyuv(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline, *uline, *vline);
+      rgbtoy(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline);
       rgbIndex += rgbIncrement;
       yline++;
       rgbtoyuv(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline, *uline, *vline);
@@ -397,10 +565,6 @@ void PStandardColourConverter::RGBtoYUV420PSameSize(const BYTE * rgb,
 }
 
 
-#define BLACK_Y 0
-#define BLACK_U 128
-#define BLACK_V 128
-
 // Simple crop/pad version.  Image aligned to top-left
 // and cropped / padded with black borders as required.
 void PStandardColourConverter::RGBtoYUV420PWithResize(const BYTE * rgb,
@@ -411,7 +575,7 @@ void PStandardColourConverter::RGBtoYUV420PWithResize(const BYTE * rgb,
   int planeSize = dstFrameWidth*dstFrameHeight;
   const int halfWidth = dstFrameWidth >> 1;
   unsigned min_width, min_height;
-  
+
   min_width  = (dstFrameWidth  < srcFrameWidth)  ? dstFrameWidth  : srcFrameWidth;
   min_height = (dstFrameHeight < srcFrameHeight) ? dstFrameHeight : srcFrameHeight;
 
@@ -432,7 +596,7 @@ void PStandardColourConverter::RGBtoYUV420PWithResize(const BYTE * rgb,
 
     for (unsigned x = 0; x < min_width; x+=2) 
     {
-     rgbtoyuv(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline, *uline, *vline);
+     rgbtoy(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline);
      rgbIndex += rgbIncrement;
      yline++;
      rgbtoyuv(rgbIndex[0], rgbIndex[1], rgbIndex[2],*yline, *uline, *vline);
@@ -476,12 +640,12 @@ BOOL PStandardColourConverter::RGBtoYUV420P(const BYTE * rgb,
 {
   if (rgb == yuv)
     return FALSE; // Cannot do in place conversion
-  
+
   if ((srcFrameWidth == dstFrameWidth) && (srcFrameHeight == dstFrameHeight)) 
     RGBtoYUV420PSameSize(rgb, yuv, rgbIncrement, flip);
   else
     RGBtoYUV420PWithResize(rgb, yuv, rgbIncrement, flip);
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
 
@@ -489,6 +653,12 @@ BOOL PStandardColourConverter::RGBtoYUV420P(const BYTE * rgb,
 }
 
 
+PSTANDARD_COLOUR_CONVERTER(Grey,YUV420P)
+{
+  return GreytoYUV420P(srcFrameBuffer, dstFrameBuffer, bytesReturned, doVFlip);
+}
+
+
 PSTANDARD_COLOUR_CONVERTER(RGB24,YUV420P)
 {
   return RGBtoYUV420P(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3,  doVFlip);
@@ -507,6 +677,12 @@ PSTANDARD_COLOUR_CONVERTER(RGB24F,YUV420P)
 }
 
 
+PSTANDARD_COLOUR_CONVERTER(GreyF,YUV420P)
+{
+  return GreytoYUV420P(srcFrameBuffer, dstFrameBuffer, bytesReturned, doVFlip);
+}
+
+
 PSTANDARD_COLOUR_CONVERTER(RGB32F,YUV420P)
 {
   return RGBtoYUV420P(srcFrameBuffer, dstFrameBuffer, bytesReturned, 4, doVFlip);
@@ -587,12 +763,12 @@ PSTANDARD_COLOUR_CONVERTER(YUV422,YUV422)
   
   if (srcFrameBuffer == dstFrameBuffer)
     return TRUE;
-  
+
   if ((srcFrameWidth == dstFrameWidth) && (srcFrameHeight == dstFrameHeight)) 
     memcpy(dstFrameBuffer,srcFrameBuffer,srcFrameWidth*srcFrameHeight*2);
   else
     ResizeYUV422(srcFrameBuffer, dstFrameBuffer);
-  
+
   return TRUE;
 }
 
@@ -655,7 +831,8 @@ BOOL PStandardColourConverter::YUV420PtoRGB(const BYTE * srcFrameBuffer,
                                             BYTE * dstFrameBuffer,
                                             PINDEX * bytesReturned,
                                             unsigned rgbIncrement,
-					    BOOL     flipVertical) const
+					    BOOL     flipVertical,
+                                            BOOL     flipBR) const
 {
   if (srcFrameBuffer == dstFrameBuffer)
     return FALSE;
@@ -665,15 +842,17 @@ BOOL PStandardColourConverter::YUV420PtoRGB(const BYTE * srcFrameBuffer,
   const BYTE    *yplane    = srcFrameBuffer;           		// 1 byte Y (luminance) for each pixel
   const BYTE    *uplane    = yplane+nbytes;              	// 1 byte U for a block of 4 pixels
   const BYTE    *vplane    = uplane+(nbytes >> 2);       	// 1 byte V for a block of 4 pixels
-  unsigned int   pixpos[4] = { 0, 1, srcFrameWidth, srcFrameWidth+1 };
+
+  unsigned int   pixpos[4] = {0, 1, srcFrameWidth, srcFrameWidth + 1};
+  unsigned int   originalPixpos[4] = {0, 1, srcFrameWidth, srcFrameWidth + 1};
+
   unsigned int   x, y, p;
 
   long     int   yvalue;
-  long     int   cr, cb, rd, gd, bd;
   long     int   l, r, g, b;
-            
+
   if(flipVertical) {
-    dstImageFrame = dstFrameBuffer + ((srcFrameHeight-2) * srcFrameWidth * rgbIncrement);
+    dstImageFrame = dstFrameBuffer + ((srcFrameHeight - 2) * srcFrameWidth * rgbIncrement);
     pixpos[0] = srcFrameWidth;
     pixpos[1] = srcFrameWidth +1;
     pixpos[2] = 0;
@@ -687,37 +866,46 @@ BOOL PStandardColourConverter::YUV420PtoRGB(const BYTE * srcFrameBuffer,
     for (x = 0; x < srcFrameWidth; x += 2)
     {
       // The RGB value without luminance
-     
-      cr = *uplane-128;
-      cb = *vplane-128;
-      rd = 104635*cb;			// 		  106986*cb
-      gd = -25690*cr-53294*cb;		// -26261*cr  +   -54496*cb 
-      bd = 132278*cr;			// 135221*cr
-     
+      long int  cr, cb, rd, gd, bd;
+
+      cb = *uplane-128;
+      cr = *vplane-128;
+      rd = 104635*cr;			// 		  106986*cr
+      gd = -25690*cb-53294*cr;		// -26261*cb  +   -54496*cr 
+      bd = 132278*cb;			// 135221*cb
+
       // Add luminance to each of the 4 pixels
-      
+
       for (p = 0; p < 4; p++)
       {
-        yvalue = *(yplane+pixpos[p])-16;
-        
+        yvalue = *(yplane + originalPixpos[p])-16;
+
         if (yvalue < 0) yvalue = 0;
-        
+
         l = 76310*yvalue;
-        
+
         r = l+rd;
         g = l+gd;
         b = l+bd;
-        
-        *(dstImageFrame + rgbIncrement*pixpos[p])   = LIMIT(b);
-        *(dstImageFrame + rgbIncrement*pixpos[p]+1) = LIMIT(g);
-        *(dstImageFrame + rgbIncrement*pixpos[p]+2) = LIMIT(r);
+
+        BYTE * rgpPtr = dstImageFrame + rgbIncrement*pixpos[p];
+        if (flipBR) {
+          *rgpPtr++ = LIMIT(r);
+          *rgpPtr++ = LIMIT(g);
+          *rgpPtr++ = LIMIT(b);
+        }
+        else {
+          *rgpPtr++= LIMIT(b);
+          *rgpPtr++= LIMIT(g);
+          *rgpPtr++ = LIMIT(r);
+        }
         if (rgbIncrement == 4)
-          *(dstImageFrame + 4*pixpos[p]+3) = 0;
+          *rgpPtr = 0;
       }
-      
+
       yplane += 2;
       dstImageFrame += rgbIncrement*2;
-      
+
       uplane++;
       vplane++;
     }
@@ -738,24 +926,32 @@ BOOL PStandardColourConverter::YUV420PtoRGB(const BYTE * srcFrameBuffer,
 
 PSTANDARD_COLOUR_CONVERTER(YUV420P,RGB24)
 {
-  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, doVFlip);
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, doVFlip, FALSE);
 }
 
+PSTANDARD_COLOUR_CONVERTER(YUV420P,BGR24)
+{
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, doVFlip, TRUE);
+}
 
 PSTANDARD_COLOUR_CONVERTER(YUV420P,RGB32)
 {
-  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 4, doVFlip);
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 4, doVFlip, FALSE);
 }
 
 PSTANDARD_COLOUR_CONVERTER(YUV420P,RGB24F)
 {
-  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, !doVFlip);
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, !doVFlip, FALSE);
 }
 
+PSTANDARD_COLOUR_CONVERTER(YUV420P,BGR24F)
+{
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 3, !doVFlip, TRUE);
+}
 
 PSTANDARD_COLOUR_CONVERTER(YUV420P,RGB32F)
 {
-  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 4, !doVFlip);
+  return YUV420PtoRGB(srcFrameBuffer, dstFrameBuffer, bytesReturned, 4, !doVFlip, FALSE);
 }
 
 
@@ -763,7 +959,7 @@ PSTANDARD_COLOUR_CONVERTER(RGB24,RGB32)
 {
   if ((dstFrameWidth != srcFrameWidth) || (dstFrameHeight != srcFrameHeight))
     return FALSE;
-  
+
   // Go from bottom to top so can do in place conversion
   const BYTE * src = srcFrameBuffer+srcFrameBytes-1;
   BYTE * dst = dstFrameBuffer+dstFrameBytes-1;
@@ -775,7 +971,7 @@ PSTANDARD_COLOUR_CONVERTER(RGB24,RGB32)
         *dst-- = *src--;
     }
   }
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
   return TRUE;
@@ -786,7 +982,7 @@ PSTANDARD_COLOUR_CONVERTER(RGB32,RGB24)
 {
   if ((dstFrameWidth != srcFrameWidth) || (dstFrameHeight != srcFrameHeight))
     return FALSE;
-  
+
   const BYTE * src = srcFrameBuffer;
   BYTE * dst = dstFrameBuffer;
 
@@ -797,7 +993,7 @@ PSTANDARD_COLOUR_CONVERTER(RGB32,RGB24)
       src++;
     }
   }
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
   return TRUE;
@@ -846,15 +1042,15 @@ PSTANDARD_COLOUR_CONVERTER(YUV420P,YUV411P)
 
   if ((dstFrameWidth != srcFrameWidth) || (dstFrameHeight != srcFrameHeight))
     return FALSE;
-  
+
   // Copy over the Y plane.
   memcpy(dstFrameBuffer, srcFrameBuffer, srcFrameWidth*srcFrameHeight);
-  
+
   unsigned linewidth = dstFrameWidth / 4;
-  
+
   // Source data is the start of the U plane
   const BYTE* src = srcFrameBuffer + srcFrameWidth * srcFrameHeight;
-  
+
   // Two output lines at a time
   BYTE *dst0 = dstFrameBuffer + dstFrameWidth * dstFrameHeight;
   BYTE *dst1 = dst0 + linewidth;
@@ -872,14 +1068,14 @@ PSTANDARD_COLOUR_CONVERTER(YUV420P,YUV411P)
     dst0 += linewidth;
     dst1 = dst0 + linewidth;
   }
-  
+
   // Source data is the start of the U plane
   src = srcFrameBuffer + srcFrameWidth * srcFrameHeight * 5 / 4;
-  
+
   // Two output lines at a time
   dst0 = dstFrameBuffer + dstFrameWidth * dstFrameHeight * 5 / 4;
   dst1 = dst0 + linewidth;
-  
+
   // V plane
   for (y = 0; y < dstFrameHeight; y += 2) {
     for (x = 0; x < dstFrameWidth; x += 4) {
@@ -891,7 +1087,7 @@ PSTANDARD_COLOUR_CONVERTER(YUV420P,YUV411P)
     dst0 += linewidth;
     dst1 = dst0 + linewidth;
   }
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
   
@@ -929,15 +1125,15 @@ PSTANDARD_COLOUR_CONVERTER(YUV411P,YUV420P)
 
   if ((dstFrameWidth != srcFrameWidth) || (dstFrameHeight != srcFrameHeight))
     return FALSE;
-  
+
   // Copy over the Y plane.
   memcpy(dstFrameBuffer, srcFrameBuffer, srcFrameWidth*srcFrameHeight);
-  
+
   unsigned linewidth = dstFrameWidth / 4;
-  
+
   // Source data is the start of the U plane
   const BYTE* src = srcFrameBuffer + srcFrameWidth * srcFrameHeight;
-  
+
   // Output line
   BYTE *dst0 = dstFrameBuffer + dstFrameWidth * dstFrameHeight;
 
@@ -946,26 +1142,26 @@ PSTANDARD_COLOUR_CONVERTER(YUV411P,YUV420P)
   // U plane
   for (y = 0; y < dstFrameHeight; y += 2) {
     for (x = 0; x < dstFrameWidth; x += 4) {
-      
+
       // Double up the horizontal samples
       *dst0++ = *src;
       *dst0++ = *src++;
     }
-    
+
     // Skip over the 2nd line we are decimating
     src += linewidth;
   }
-  
+
   // Source data is the start of the U plane
   src = srcFrameBuffer + srcFrameWidth * srcFrameHeight * 5 / 4;
-  
+
   // Output line
   dst0 = dstFrameBuffer + dstFrameWidth * dstFrameHeight * 5 / 4;
-  
+
   // V plane
   for (y = 0; y < dstFrameHeight; y += 2) {
     for (x = 0; x < dstFrameWidth; x += 4) {
-      
+
       // Double up the samples horizontal samples
       *dst0++ = *src;
       *dst0++ = *src++;
@@ -974,10 +1170,10 @@ PSTANDARD_COLOUR_CONVERTER(YUV411P,YUV420P)
     // Skip over the 2nd source line we already did.
     src += linewidth;
   }
-  
+
   if (bytesReturned != NULL)
     *bytesReturned = dstFrameBytes;
-  
+
   return TRUE;
 }
 
@@ -1117,5 +1313,5 @@ PSTANDARD_COLOUR_CONVERTER(UYV444,YUV420P)
   return TRUE;
 }
 
- 
+
 // End Of File ///////////////////////////////////////////////////////////////
