@@ -24,6 +24,44 @@
  * Contributor(s): Mark Cooke (mpc@star.sr.bham.ac.uk)
  *
  * $Log: videoio.cxx,v $
+ * Revision 1.39  2003/05/14 07:51:23  rjongbloed
+ * Changed SetColourFormatConverter so if converter already in place no
+ *   change is made.
+ * Fixed some trace logs.
+ *
+ * Revision 1.38  2003/04/03 23:21:34  robertj
+ * Added reversed RGB byte order versions (BGR24), thanks Damien Sandras
+ *
+ * Revision 1.37  2003/03/21 04:09:33  robertj
+ * Changed PPM video output device so you can specify the full format of the
+ *   output file uinng printf command for the frame number eg %u or %03i or
+ *   something. If there is no %u in the Opan() argument, a %u is added after
+ *   the filename.
+ * Fixed video output RGB SetFrameData so abide by correct semantics. The input
+ *   is aways to be what was set using SetColourFormat() or
+ *   SetColourFormatConverter().
+ *
+ * Revision 1.36  2003/03/20 23:42:01  dereks
+ * Make PPM video output device work correctly.
+ *
+ * Revision 1.35  2003/03/17 07:50:41  robertj
+ * Added OpenFull() function to open with all video parameters in one go.
+ * Made sure vflip variable is set in converter even if converter has not
+ *   been set yet, should not depend on the order of functions!
+ * Removed canCaptureVideo variable as this is really a virtual function to
+ *   distinguish PVideoOutputDevice from PVideoInputDevice, it is not dynamic.
+ * Made significant enhancements to PVideoOutputDevice class.
+ * Added PVideoOutputDevice descendants for NULL and PPM files.
+ *
+ * Revision 1.34  2002/09/01 23:25:26  dereks
+ * Documentation fix from Walter Whitlock. Many thanks.
+ *
+ * Revision 1.33  2002/09/01 22:38:21  dereks
+ * Remove previous clarification, cause it breaks openphone code.
+ *
+ * Revision 1.32  2002/08/30 02:31:43  dereks
+ * Make operation of the code more clear. Thanks Diego Tartara
+ *
  * Revision 1.31  2002/04/12 08:24:53  robertj
  * Added text string output for tracing video format.
  *
@@ -175,8 +213,7 @@ PVideoDevice::PVideoDevice()
   channelNumber = -1;	// -1 will find the first working channel number.
   frameWidth = CIFWidth;
   frameHeight = CIFHeight;
-
-  SetCanCaptureVideo(FALSE);    ///This device cannot grab video from a port.
+  doVFlip = FALSE;
 
   converter = NULL;
 
@@ -190,12 +227,104 @@ PVideoDevice::~PVideoDevice()
 }
 
 
+BOOL PVideoDevice::OpenFull(const OpenArgs & args, BOOL startImmediate)
+{
+  if (args.deviceName[0] == '#') {
+    PStringArray devices = GetDeviceNames();
+    PINDEX id = args.deviceName.Mid(1).AsUnsigned();
+    if (id == 0 || id > devices.GetSize())
+      return FALSE;
+
+    if (!Open(devices[id-1], FALSE))
+      return FALSE;
+  }
+  else {
+    if (!Open(args.deviceName, FALSE))
+      return FALSE;
+  }
+
+  if (!SetVideoFormat(args.videoFormat))
+    return FALSE;
+
+  if (!SetChannel(args.channelNumber))
+    return FALSE;
+
+  if (args.convertFormat) {
+    if (!SetColourFormatConverter(args.colourFormat))
+      return FALSE;
+  }
+  else {
+    if (!SetColourFormat(args.colourFormat))
+      return FALSE;
+  }
+
+  if (args.rate > 0) {
+    if (!SetFrameRate(args.rate))
+      return FALSE;
+  }
+
+  if (args.convertSize) {
+    if (!SetFrameSizeConverter(args.width, args.height, args.scaleSize))
+      return FALSE;
+  }
+  else {
+    if (!SetFrameSize(args.width, args.height))
+      return FALSE;
+  }
+
+  if (!SetVFlipState(args.flip))
+    return FALSE;
+
+  if (args.brightness >= 0) {
+    if (!SetBrightness(args.brightness))
+      return FALSE;
+  }
+
+  if (args.whiteness >= 0) {
+    if (!SetWhiteness(args.whiteness))
+      return FALSE;
+  }
+
+  if (args.contrast >= 0) {
+    if (!SetContrast(args.contrast))
+      return FALSE;
+  }
+
+  if (args.colour >= 0) {
+    if (!SetColour(args.colour))
+      return FALSE;
+  }
+
+  if (args.hue >= 0) {
+    if (!SetColour(args.hue))
+      return FALSE;
+  }
+
+  if (startImmediate)
+    return Start();
+
+  return TRUE;
+}
+
+
 BOOL PVideoDevice::Close()
 {
   return TRUE;  
 }
 
 
+BOOL PVideoDevice::Start()
+{
+  return TRUE;
+}
+
+
+BOOL PVideoDevice::Stop()
+{
+  return TRUE;
+}
+
+
 BOOL PVideoDevice::SetVideoFormat(VideoFormat videoFmt)
 {
   videoFormat = videoFmt;
@@ -218,9 +347,10 @@ int PVideoDevice::GetNumChannels()
 BOOL PVideoDevice::SetChannel(int channelNum)
 {
   if (channelNum < 0) { // Seek out the first available channel
-    for (int c = 0; c < GetNumChannels(); c++)
+    for (int c = 0; c < GetNumChannels(); c++) {
       if (SetChannel(c))
         return TRUE;
+    }
     return FALSE;
   }
 
@@ -246,6 +376,8 @@ static struct {
 } colourFormatBPPTab[] = {
   { "RGB24",   24 },
   { "RGB24F",  24 },
+  { "BGR24",   24 },
+  { "BGR24F",  24 },
   { "RGB32",   32 },
   { "RGB32F",  32 },
   { "YUV422",  16 },
@@ -271,7 +403,15 @@ static struct {
 
 BOOL PVideoDevice::SetColourFormatConverter(const PString & colourFmt)
 {
-  if (converter) {    
+  if (converter != NULL) {
+    if (CanCaptureVideo()) {
+      if (converter->GetDstColourFormat() == colourFmt)
+        return TRUE;
+    }
+    else {
+      if (converter->GetSrcColourFormat() == colourFmt)
+        return TRUE;
+    }
     delete converter;
     converter = NULL;
   }
@@ -279,12 +419,19 @@ BOOL PVideoDevice::SetColourFormatConverter(const PString & colourFmt)
   if (!preferredColourFormat.IsEmpty()) {
     PTRACE(4,"PVidDev\tSetColourFormatConverter, want " << colourFmt << " trying " << preferredColourFormat);
     if (SetColourFormat(preferredColourFormat)) {
-      PTRACE(4,"PVidDev\tSetColourFormatConverter set camera to native "<< preferredColourFormat);
-      converter = PColourConverter::Create(preferredColourFormat, colourFmt, frameWidth, frameHeight);
+      if (CanCaptureVideo()) {
+        PTRACE(4,"PVidDev\tSetColourFormatConverter set camera to native "<< preferredColourFormat);
+        converter = PColourConverter::Create(preferredColourFormat, colourFmt, frameWidth, frameHeight);
+      }
+      else {
+        PTRACE(4,"PVidDev\tSetColourFormatConverter set renderer to "<< preferredColourFormat);
+        converter = PColourConverter::Create(colourFmt, preferredColourFormat, frameWidth, frameHeight);
+      }
       if (converter != NULL) {
         // set converter properties that depend on this color format
         if (SetColourFormat(preferredColourFormat)) {
-          PTRACE(3, "PVidDev\tSetColourFormatConverter succeeded for converted " << colourFmt << " camera using " << preferredColourFormat);
+          PTRACE(3, "PVidDev\tSetColourFormatConverter succeeded for " << colourFmt << " and device using " << preferredColourFormat);
+          converter->SetVFlipState(doVFlip);
           return TRUE;
         }
       } 
@@ -312,12 +459,19 @@ BOOL PVideoDevice::SetColourFormatConverter(const PString & colourFmt)
     PString formatToTry = colourFormatBPPTab[knownFormatIdx].colourFormat;
     PTRACE(4,"PVidDev\tSetColourFormatConverter, want " << colourFmt << " trying " << formatToTry);
     if (SetColourFormat(formatToTry)) {
-      PTRACE(4,"PVidDev\tSetColourFormatConverter set camera to "<< formatToTry);
-      converter = PColourConverter::Create(formatToTry, colourFmt, frameWidth, frameHeight);
+      if (CanCaptureVideo()) {
+        PTRACE(4,"PVidDev\tSetColourFormatConverter set camera to "<< formatToTry);
+        converter = PColourConverter::Create(formatToTry, colourFmt, frameWidth, frameHeight);
+      }
+      else {
+        PTRACE(4,"PVidDev\tSetColourFormatConverter set renderer to "<< formatToTry);
+        converter = PColourConverter::Create(colourFmt, formatToTry, frameWidth, frameHeight);
+      }
       if (converter != NULL) {
         // set converter properties that depend on this color format
         if (SetColourFormat(formatToTry)) {
-          PTRACE(3, "PVidDev\tSetColourFormatConverter succeeded for converted " << colourFmt << " camera using " << formatToTry);
+          PTRACE(3, "PVidDev\tSetColourFormatConverter succeeded for " << colourFmt << " and device using " << formatToTry);
+          converter->SetVFlipState(doVFlip);
           return TRUE;
         }
       } 
@@ -423,7 +577,7 @@ BOOL PVideoDevice::SetFrameSizeConverter(unsigned width, unsigned height,
                  << prefResizeTable[prefResizeIdx].device_width << 'x'
                  << prefResizeTable[prefResizeIdx].device_height
                  << " --> " << width<< 'x' <<height);	       
-
+          converter->SetVFlipState(doVFlip);
 	  return TRUE;
 	}
 	PTRACE(2,"PVidDev\tSetFrameSizeConverter FAILED for "
@@ -443,16 +597,21 @@ BOOL PVideoDevice::SetFrameSizeConverter(unsigned width, unsigned height,
   unsigned minWidth, minHeight, maxWidth, maxHeight;
   GetFrameSizeLimits(minWidth, minHeight, maxWidth, maxHeight);
 
-  if (SetFrameSize(maxWidth, maxHeight)){
-    PTRACE(4,"PVidDev\tSuccess set hardware size to " << maxWidth << 'x' << maxHeight);
-    if (converter->SetDstFrameSize(width, height, bScaleNotCrop)){
-      PTRACE(3,"PVidDev\tSetFrameSizeConvert SUCCEEDED for " << width << 'x' << height);
-      return TRUE;
-    }
+  if (!SetFrameSize(maxWidth, maxHeight)){
+    PTRACE(2,"PVidDev\tHardware SetFrameSize FAILED for " << maxWidth << 'x' << maxHeight);
+    return FALSE;
   }
 
-  PTRACE(2,"PVidDev\tSetFrameSizeConverter FAILED for " << width << 'x' << height);
-  return FALSE;
+  PTRACE(4,"PVidDev\tSuccess set hardware size to " << maxWidth << 'x' << maxHeight);
+  if (!converter->SetDstFrameSize(width, height, bScaleNotCrop)){
+    PTRACE(2,"PVidDev\tSetFrameSizeConverter FAILED for " << maxWidth << 'x' << maxHeight);
+    return FALSE;
+  }
+
+  converter->SetVFlipState(doVFlip);
+
+  PTRACE(3,"PVidDev\tSetFrameSizeConvert SUCCEEDED for " << width << 'x' << height);
+  return TRUE;
 }
 
 
@@ -548,12 +707,6 @@ unsigned PVideoDevice::CalculateFrameBytes(unsigned width, unsigned height,
 }
  
 
-BOOL PVideoDevice::CanCaptureVideo()
-{
-  return deviceCanCaptureVideo;
-}
-
-
 int PVideoDevice::GetBrightness()
 {
   return frameBrightness;
@@ -619,8 +772,11 @@ BOOL PVideoDevice::SetHue(unsigned newHue)
 }
 
     
-BOOL PVideoDevice::GetParameters (int *whiteness, int *brightness, 
-                                      int *colour, int *contrast, int *hue)
+BOOL PVideoDevice::GetParameters (int *whiteness,
+                                  int *brightness, 
+                                  int *colour,
+                                  int *contrast,
+                                  int *hue)
 {
   if (!IsOpen())
     return FALSE;
@@ -646,47 +802,246 @@ BOOL PVideoDevice::SetVideoChannelFormat (int newNumber, VideoFormat newFormat)
 
 BOOL PVideoDevice::GetVFlipState()
 {
-  return FALSE;
+  if (converter != NULL)
+    doVFlip = converter->GetVFlipState();
+
+  return doVFlip;
 }
 
 
-BOOL PVideoDevice::SetVFlipState(BOOL /*newVFlipState*/)
+BOOL PVideoDevice::SetVFlipState(BOOL newVFlip)
 {
-  return FALSE;
+  doVFlip = newVFlip;
+
+  if (converter != NULL)
+    converter->SetVFlipState(newVFlip);
+
+  return TRUE;
 }
 
 
-BOOL PVideoDevice::ToggleVFlipState()
-{
-  return FALSE;
-}
-
 
 ///////////////////////////////////////////////////////////////////////////////
 // PVideoOutputDevice
 
 PVideoOutputDevice::PVideoOutputDevice()
 {
-  now = 0;
-  suppress = FALSE;
 }
 
 
-BOOL PVideoOutputDevice::Redraw(const void * /*frame*/)
+BOOL PVideoOutputDevice::CanCaptureVideo() const
 {
   return FALSE;
 }
 
 
-void PVideoOutputDevice::SetNow(int _now)
+///////////////////////////////////////////////////////////////////////////////
+// PVideoOutputDeviceNULL
+
+PVideoOutputDeviceNULL::PVideoOutputDeviceNULL()
 {
-  now = _now;
+  deviceName = "NULL";
+}
+
+
+BOOL PVideoOutputDeviceNULL::Open(const PString & /*deviceName*/,
+                                  BOOL /*startImmediate*/)
+{
+  return TRUE;
+}
+
+
+BOOL PVideoOutputDeviceNULL::IsOpen()
+{
+  return TRUE;
+}
+
+
+PStringList PVideoOutputDeviceNULL::GetDeviceNames() const
+{
+  PStringList list;
+  list += "NULL";
+  return list;
+}
+
+
+PINDEX PVideoOutputDeviceNULL::GetMaxFrameBytes()
+{
+  return CalculateFrameBytes(frameWidth, frameHeight, colourFormat);
+}
+
+
+BOOL PVideoOutputDeviceNULL::SetFrameData(unsigned /*x*/, unsigned /*y*/,
+                                          unsigned /*width*/, unsigned /*height*/,
+                                          const BYTE * /*data*/,
+                                          BOOL /*endFrame*/)
+{
+  return TRUE;
+}
+
+
+BOOL PVideoOutputDeviceNULL::EndFrame()
+{
+  return TRUE;
+}
+
+
+///////////////////////////////////////////////////////////////////////////////
+// PVideoOutputDeviceRGB
+
+PVideoOutputDeviceRGB::PVideoOutputDeviceRGB()
+{
+  PTRACE(6, "PPM\t Constructor of PVideoOutputDeviceRGB");
+
+  colourFormat = "RGB24";
+  bytesPerPixel = 3;
+  frameStore.SetSize(frameWidth * frameHeight * bytesPerPixel);
+}
+
+
+BOOL PVideoOutputDeviceRGB::SetColourFormat(const PString & colourFormat)
+{
+  if (colourFormat == "RGB32")
+    bytesPerPixel = 4;
+  else if (colourFormat == "RGB24")
+    bytesPerPixel = 3;
+  else
+    return FALSE;
+
+  return PVideoOutputDevice::SetColourFormat(colourFormat) &&
+         SetFrameSize(frameWidth, frameHeight);
+}
+
+
+BOOL PVideoOutputDeviceRGB::SetFrameSize(unsigned width, unsigned height)
+{
+  if (!PVideoOutputDevice::SetFrameSize(width, height))
+    return FALSE;
+
+  return frameStore.SetSize(frameWidth*frameHeight*bytesPerPixel);
+}
+
+
+PINDEX PVideoOutputDeviceRGB::GetMaxFrameBytes()
+{
+  return frameStore.GetSize();
+}
+
+
+BOOL PVideoOutputDeviceRGB::SetFrameData(unsigned x, unsigned y,
+                                         unsigned width, unsigned height,
+                                         const BYTE * data,
+                                         BOOL endFrame)
+{
+  if (x+width > frameWidth || y+height > frameHeight)
+    return FALSE;
+
+  if (x == 0 && width == frameWidth && y == 0 && height == frameHeight) {
+    if (converter != NULL)
+      converter->Convert(data, frameStore.GetPointer());
+    else
+      memcpy(frameStore.GetPointer(), data, height*width*bytesPerPixel);
+  }
+  else {
+    if (converter != NULL) {
+      PAssertAlways("Converted output of partial RGB frame not supported");
+      return FALSE;
+    }
+
+    if (x == 0 && width == frameWidth)
+      memcpy(frameStore.GetPointer() + y*width*bytesPerPixel, data, height*width*bytesPerPixel);
+    else {
+      for (unsigned dy = 0; dy < height; dy++)
+        memcpy(frameStore.GetPointer() + ((y+dy)*width + x)*bytesPerPixel,
+               data + dy*width*bytesPerPixel, width*bytesPerPixel);
+    }
+  }
+
+  if (endFrame)
+    return EndFrame();
+
+  return TRUE;
+}
+
+///////////////////////////////////////////////////////////////////////////////
+// PVideoOutputDevicePPM
+
+PVideoOutputDevicePPM::PVideoOutputDevicePPM()
+{
+  PTRACE(6, "PPM\t Constructor of PVideoOutputDevicePPM");
+  frameNumber = 0;
+}
+
+
+BOOL PVideoOutputDevicePPM::Open(const PString & name,
+                                 BOOL /*startImmediate*/)
+{
+  Close();
+
+  PFilePath path = name;
+  if (!PDirectory::Exists(path.GetDirectory()))
+    return FALSE;
+
+  if (path != psprintf(path, 12345))
+    deviceName = path;
+  else
+    deviceName = path.GetDirectory() + path.GetTitle() + "%u" + path.GetType();
+
+  return TRUE;
+}
+
+
+BOOL PVideoOutputDevicePPM::IsOpen()
+{
+  return !deviceName;
+}
+
+
+BOOL PVideoOutputDevicePPM::Close()
+{
+  deviceName = PString::Empty();
+  return TRUE;
+}
+
+
+PStringList PVideoOutputDevicePPM::GetDeviceNames() const
+{
+  PStringList list;
+  list += PDirectory();
+  return list;
+}
+
+
+BOOL PVideoOutputDevicePPM::EndFrame()
+{
+  PFile file;
+  if (!file.Open(psprintf(deviceName, frameNumber++), PFile::WriteOnly)) {
+    PTRACE(1, "PPMVid\tFailed to open PPM output file \""
+           << file.GetName() << "\": " << file.GetErrorText());
+    return FALSE;
+  }
+
+  file << "P6 " << frameWidth  << " " << frameHeight << " " << 255 << "\n";
+
+  if (!file.Write(frameStore, frameStore.GetSize())) {
+    PTRACE(1, "PPMVid\tFailed to write frame data to PPM output file " << file.GetName());
+    return FALSE;
+  }
+
+  PTRACE(6, "PPMVid\tFinished writing PPM file " << file.GetName());
+  return file.Close();
 }
 
 
 ///////////////////////////////////////////////////////////////////////////////
 // PVideoInputDevice
 
+BOOL PVideoInputDevice::CanCaptureVideo() const
+{
+  return TRUE;
+}
+
+
 PStringList PVideoInputDevice::GetDeviceNames() const
 {
   return GetInputDeviceNames();
@@ -704,29 +1059,4 @@ BOOL PVideoInputDevice::GetFrame(PBYTEArray & frame)
 }
 
 
-BOOL PVideoInputDevice::GetVFlipState() 
-{
-  if ( converter == NULL )
-    return FALSE;
-
-  return converter->GetVFlipState();
-} 
-
-BOOL PVideoInputDevice::SetVFlipState(BOOL newVFlip) 
-{
-  if ( converter == NULL )
-    // no converter (yet), any new converter will have doVFlip = FALSE
-    return !newVFlip;
-  converter->SetVFlipState(newVFlip);
-  return TRUE;
-} 
-
-BOOL PVideoInputDevice::ToggleVFlipState() 
-{
-  if ( converter == NULL )
-    return FALSE;
-  converter->ToggleVFlipState();
-  return TRUE;
-} 
-
 // End Of File ///////////////////////////////////////////////////////////////
