@@ -27,6 +27,83 @@
  * Contributor(s): Loopback feature: Philip Edelbrock <phil@netroedge.com>.
  *
  * $Log: oss.cxx,v $
+ * Revision 1.59  2003/05/24 10:57:05  rogerhardiman
+ * If a sound device cannot be opened in RW mode, try just 'R' or 'W' modes.
+ * Needed for USB web cams with Mics which are read only sound devices when using
+ * ALSA. Tested by Damien.
+ *
+ * Revision 1.58  2003/02/02 18:54:22  rogerh
+ * FreeBSD changes for support of dspN.M (eg dsp0.0) sound card entries.
+ * Problem reported by Lars Eggert <larse@isi.edu>
+ *
+ * Revision 1.57  2003/01/06 19:25:07  rogerh
+ * Add NetBSD video support.
+ * Add correct includes for OSS ioctls (note the proper way to do this now
+ * is via pwlib commands)
+ * From Andreas Wrede, taken in part from NetBSD's package system
+ *
+ * Revision 1.56  2003/01/06 19:10:22  rogerh
+ * NetBSD uses /dev/audio and not /dev/dsp
+ *
+ * Revision 1.55  2002/12/12 09:03:56  rogerh
+ * On two FreeBSD machines, Read() calls from the sound card were not blocking
+ * correctly and returned with less bytes than asked for. This made OpenH323
+ * close the sound channel.  Add a FreeBSD workaround so Read() loops until it
+ * has all the bytes requested.
+ *
+ * Revision 1.54  2002/12/03 23:03:54  rogerh
+ * oops - remove some test code which should not have been committed
+ *
+ * Revision 1.53  2002/12/03 19:11:58  rogerh
+ * Open sound device in non blocking mode incase it is already open.
+ *
+ * Revision 1.52  2002/11/28 12:15:24  rogerh
+ * Change SetVolume/GetVolume to use the mic and not the igain for the input
+ * volume.
+ * Our target audience are likely to be using mics and many broken
+ * sound drivers do not implement igain properly.
+ *
+ * Revision 1.51  2002/10/17 12:57:24  robertj
+ * Added ability to increase maximum file handles on a process.
+ *
+ * Revision 1.50  2002/10/15 10:42:45  rogerh
+ * Fix loopback mode, which was broken in a recent change.
+ *
+ * Revision 1.49  2002/09/29 16:19:28  rogerh
+ * if /dev/dsp does not exist, do not return it as the default audio device.
+ * Instead, return the first dsp device.
+ *
+ * Revision 1.48  2002/09/29 15:56:49  rogerh
+ * Revert back to checking for the /dev/soundcard directory to detect devfs.
+ * If seems that the .devfsd file is not removed when devfs is not being used.
+ *
+ * Revision 1.47  2002/09/29 09:26:16  rogerh
+ * Changes to sound card detection.
+ * For Damien Sandras, allow /dev/dsp to be added to the list of sound devices
+ * For FreeBSD, ignore /dev/dspN.M eg /dev/dsp0.2 which are virtual soundcards
+ *
+ * Revision 1.46  2002/08/30 07:58:27  craigs
+ * Added fix for when sound cards are already open, thanks to Damien Sandras
+ *
+ * Revision 1.45  2002/08/15 19:57:38  rogerh
+ * Linux defvs mode is detected with /dev/.devfsd
+ *
+ * Revision 1.44  2002/07/04 05:00:36  robertj
+ * Fixed order of calls for OSS driver setup.
+ *
+ * Revision 1.43  2002/06/24 20:01:53  rogerh
+ * Add support for linux devfs and /dev/sound. Based on code by Snark.
+ *
+ * Revision 1.42  2002/06/09 16:33:45  rogerh
+ * Use new location of soundcard.h on FreeBSD
+ *
+ * Revision 1.41  2002/06/05 12:29:15  craigs
+ * Changes for gcc 3.1
+ *
+ * Revision 1.40  2002/05/02 14:19:32  rogerh
+ * Handle Big Endian systems correctly.
+ * Patch submitted by andi@fischlustig.de
+ *
  * Revision 1.39  2002/02/11 07:21:46  rogerh
  * Fix some non portable code which which seeks out /dev/dsp devices and only
  * worked on Linux. (char device for /dev/dsp is 14 on Linux, 30 on FreeBSD)
@@ -160,15 +237,19 @@
 #pragma implementation "sound.h"
 
 #include <ptlib.h>
+#include <ptlib/socket.h>
 
 #ifdef P_LINUX
 #include <sys/soundcard.h>
-#include <sys/time.h>
 #endif
 
 #ifdef P_FREEBSD
+#if P_FREEBSD >= 500000
+#include <sys/soundcard.h>
+#else
 #include <machine/soundcard.h>
 #endif
+#endif
 
 #if defined(P_OPENBSD) || defined(P_NETBSD)
 #include <soundcard.h>
@@ -260,7 +341,7 @@ PDICTIONARY(SoundHandleDict, PString, SoundHandleEntry);
 
 static char buffer[LOOPBACK_BUFFER_SIZE];
 static int  startptr, endptr;
-static unsigned int bufferLen;
+static PINDEX bufferLen;
 static PMutex audioBufferMutex;
 PAudioDelay readDelay;
 PAudioDelay writeDelay;
@@ -391,8 +472,22 @@ PSoundChannel::~PSoundChannel()
   Close();
 }
 
+static BOOL IsNumericString(PString numbers) {
+  // return true if 'numbers' contains only digits (0 to 9)
+  // or if it contains digits followed by a '.'
 
-static void CollectSoundDevices(PDirectory devdir, POrdinalToString & dsp, POrdinalToString & mixer)
+  BOOL isNumber = FALSE;
+  for (PINDEX p = 0; p < numbers.GetLength(); p++) {
+    if (isdigit(numbers[p])) {
+      isNumber = TRUE;
+    } else {
+      return isNumber;
+    }
+  }
+  return isNumber;
+}
+
+static void CollectSoundDevices(PDirectory devdir, POrdinalToString & dsp, POrdinalToString & mixer, BOOL collect_with_names)
 {
   if (!devdir.Open())
     return;
@@ -401,45 +496,85 @@ static void CollectSoundDevices(PDirectory devdir, POrdinalToString & dsp, POrdi
     PString filename = devdir.GetEntryName();
     PString devname = devdir + filename;
     if (devdir.IsSubDir())
-      CollectSoundDevices(devname, dsp, mixer);
+      CollectSoundDevices(devname, dsp, mixer, collect_with_names);
     else {
-#if defined(P_LINUX)
-      // On Linux, look at the character device numbers
-      PFileInfo info;
-      if (devdir.GetInfo(info) &&info.type == PFileInfo::CharDevice) {
-        struct stat s;
-        if (lstat(devname, &s) == 0) {
-          // OSS compatible audio major device numbers (OSS, SAM9407, etc)
-          static const unsigned deviceNumbers[] = { 14, 145 };
-          for (PINDEX i = 0; i < PARRAYSIZE(deviceNumbers); i++) {
-            if ((s.st_rdev >> 8) == deviceNumbers[i]) {
-              PINDEX cardnum = (s.st_rdev >> 4) & 15;
-              if ((s.st_rdev & 15) == 3)   // Digital audio minor device number
-                dsp.SetAt(cardnum, devname);
-              else if ((s.st_rdev & 15) == 0)   // Digital audio minor device number
-                mixer.SetAt(cardnum, devname);
+      if (!collect_with_names) {
+        // On Linux, look at the character device numbers
+        PFileInfo info;
+        if (devdir.GetInfo(info) &&info.type == PFileInfo::CharDevice) {
+          struct stat s;
+          if (lstat(devname, &s) == 0) {
+            // OSS compatible audio major device numbers (OSS, SAM9407, etc)
+            static const unsigned deviceNumbers[] = { 14, 145 };
+            for (PINDEX i = 0; i < PARRAYSIZE(deviceNumbers); i++) {
+              if ((s.st_rdev >> 8) == deviceNumbers[i]) {
+                PINDEX cardnum = (s.st_rdev >> 4) & 15;
+                if ((s.st_rdev & 15) == 3)   // Digital audio minor device number
+                  dsp.SetAt(cardnum, devname);
+                else if ((s.st_rdev & 15) == 0)   // Digital audio minor device number
+                  mixer.SetAt(cardnum, devname);
+              }
             }
-          }
+	  }
+        }
+      } else {
+        // On Linux devfs systems, the major numbers can change dynamically.
+	// On FreeBSD and other OSs, the major numbes are different to Linux.
+	// So collect devices by looking for dsp(N) and mixer(N).
+        // (or /dev/audio(N) and mixer(N) on NetBSD
+	// Notes. FreeBSD supports audio stream mixing. A single sound card
+	// may have multiple /dev entries in the form /dev/dspN.M
+	// eg /dev/dsp0.0 /dev/dsp0.1 /dev/dsp0.2 and /dev/dsp0.3
+	// When adding these to the 'dsp' string array, only the first one
+	// found is used.
+
+#ifndef P_NETBSD
+        // Look for dsp
+        if (filename == "dsp") {
+          dsp.SetAt(0, devname);
+        }
+
+        // Look for dspN entries. Insert at position N + 1
+        // and look for dspN.M entries. Insert at position N + 1 (ignoring M)
+
+        if ((filename.GetLength() > 3) && (filename.Left(3) == "dsp")) {
+
+	  PString numbers = filename.Mid(3); // get everything after 'dsp'
+	  if (IsNumericString(numbers)) {
+            PINDEX cardnum = numbers.AsInteger(); //dspN.M is truncated to dspN.
+	    // If we have not yet inserted something for this cardnum, insert it
+	    if (dsp.GetAt(cardnum+1) == NULL) {
+              dsp.SetAt(cardnum+1, devname);
+	    }
+	  }
+        }
+#else
+        // Look for audio on NetBSD 
+        if (filename == "audio") {
+          dsp.SetAt(0, devname);
+        }
+        // Look for audioN. Insert at position cardnum + 1
+        if ((filename.GetLength() > 5) && (filename.Left(5) == "audio")) {
+	  PString numbers = filename.Mid(5); // get everything after 'audio'
+	  if (IsNumericString(numbers)) {
+            PINDEX cardnum = numbers.AsInteger();
+            dsp.SetAt(cardnum+1, devname);
+	  }
+        }
+#endif
+        // Look for mixer
+        if (filename == "mixer") {
+          mixer.SetAt(0, devname);
+        }
+        // Look for mixerN. Insert at position cardnum + 1
+        if ((filename.GetLength() > 5) && (filename.Left(5) == "mixer")) {
+	  PString numbers = filename.Mid(5); // get everything after 'mixer'
+	  if (IsNumericString(numbers)) {
+            PINDEX cardnum = numbers.AsInteger();
+            mixer.SetAt(cardnum+1, devname);
+	  }
         }
       }
-#else
-      // On FreeBSD, we could have looked for character device '30' for OSS
-      // but as we also suppose OSS on OpenBSD and NetBSD, we will use
-      // this OS independent code for all non Linux platforms.
-
-      // Look for dspN
-      if ( (filename.GetLength() >= 4) && (filename.Left(3) == "dsp")
-        && (filename.Mid(3,1) >= "0") && (filename.Mid(3,1) <= "9")) {
-        PINDEX cardnum = filename.Mid(3).AsInteger();
-        dsp.SetAt(cardnum, devname);
-      }
-      // Look for mixerN
-      if ( (filename.GetLength() >= 6) && (filename.Left(5) == "mixer")
-        && (filename.Mid(5,1) >= "0") && (filename.Mid(5,1) <= "9")) {
-        PINDEX cardnum = filename.Mid(5).AsInteger();
-        mixer.SetAt(cardnum, devname);
-      }
-#endif
     }
   } while (devdir.Next());
 }
@@ -447,10 +582,23 @@ static void CollectSoundDevices(PDirectory devdir, POrdinalToString & dsp, POrdi
 
 PStringArray PSoundChannel::GetDeviceNames(Directions /*dir*/)
 {
-  // First scan all of the devices and look for ones with major device
-  // numbers corresponding to OSS compatible drivers.
+  // First locate sound cards. On Linux with devfs and on the other platforms
+  // (eg FreeBSD), we search for filenames with dspN or mixerN.
+  // On linux without devfs we scan all of the devices and look for ones
+  // with major device numbers corresponding to OSS compatible drivers.
+
   POrdinalToString dsp, mixer;
-  CollectSoundDevices("/dev", dsp, mixer);
+
+#ifdef P_LINUX
+  PDirectory devdir = "/dev/sound";
+  if (devdir.Open()) {
+    CollectSoundDevices("/dev/sound", dsp, mixer, TRUE); // use names (devfs)
+  } else {
+    CollectSoundDevices("/dev", dsp, mixer, FALSE); // use major numbers
+  }
+#else
+  CollectSoundDevices("/dev", dsp, mixer, TRUE); // use names
+#endif
 
   // Now we go through the collected devices and see if any have a phyisical reality
   PStringList devices;
@@ -469,12 +617,13 @@ PStringArray PSoundChannel::GetDeviceNames(Directions /*dir*/)
         ::close(fd);
       }
       else {
+
         // mixer failed but this could still be a valid dsp...
         // warning this is just a hack to make this work on strange mixer and dsp configurations
         // on my machine the first sound card registers 1 mixer and 2 dsp, so when my webcam
         // registers itself as dsp2 this test would fail...
-        int fd = ::open(dsp[cardnum], O_RDONLY);
-        if (fd >= 0) {
+        int fd = ::open(dsp[cardnum], O_RDONLY | O_NONBLOCK);
+        if (fd >= 0 || errno == EBUSY) {
           devices.AppendString(dsp[cardnum]);
           ::close(fd);
         }
@@ -483,8 +632,9 @@ PStringArray PSoundChannel::GetDeviceNames(Directions /*dir*/)
     else {
       // No mixer available, try and open it directly, this could fail if
       // the device happens to be open already
-      int fd = ::open(dsp[cardnum], O_RDONLY);
-      if (fd >= 0) {
+      int fd = ::open(dsp[cardnum], O_RDONLY | O_NONBLOCK);
+
+      if (fd >= 0 || errno == EBUSY) {
         devices.AppendString(dsp[cardnum]);
         ::close(fd);
       }
@@ -497,11 +647,22 @@ PStringArray PSoundChannel::GetDeviceNames(Directions /*dir*/)
 }
 
 
-PString PSoundChannel::GetDefaultDevice(Directions /*dir*/)
+PString PSoundChannel::GetDefaultDevice(Directions dir)
 {
-  return "/dev/dsp";
-}
+  // Normally /dev/dsp points to the default sound device. If this is not
+  // present, probe /dev for sound devices and return the first detected device.
 
+  if (PFile::Exists("/dev/dsp")) {
+    return "/dev/dsp";
+  } else if (PFile::Exists("/dev/audio")) { // the NetBSD default name
+    return "/dev/audio";
+  } else {
+    // return the first dsp device detected
+    PStringArray devicenames;
+    devicenames = PSoundChannel::GetDeviceNames(dir);
+    return devicenames[0];
+  }
+}
 
 BOOL PSoundChannel::Open(const PString & _device,
                               Directions _dir,
@@ -542,9 +703,29 @@ BOOL PSoundChannel::Open(const PString & _device,
       startptr = endptr = 0;
       bufferLen = 0;
       os_handle = 0; // Use os_handle value 0 to indicate loopback, cannot ever be stdin!
+    } else {
+      // open the device in non-blocking mode to avoid hang if already open
+      os_handle = ::open((const char *)_device, O_RDWR | O_NONBLOCK);
+
+      if ((os_handle < 0) && (errno != EWOULDBLOCK)) {
+        // cannot open in read/write mode. Try in one direction only. (for ALSA)
+        if (_dir == Recorder) {
+          os_handle = ::open((const char *)_device, O_RDONLY | O_NONBLOCK);
+        } else {
+          os_handle = ::open((const char *)_device, O_WRONLY | O_NONBLOCK);
+        }
+
+        if ((os_handle < 0) && (errno != EWOULDBLOCK)) {
+          return ConvertOSError(os_handle);
+        }
+
+        // device opened, but in half duplex mode
+      }
+
+      // switch to blocking mode
+      DWORD cmd = 0;
+      ::ioctl(os_handle, FIONBIO, &cmd);
     }
-    else if (!ConvertOSError(os_handle = ::open((const char *)_device, O_RDWR))) 
-      return FALSE;
 
     // add the device to the dictionary
     SoundHandleEntry * entry = PNEW SoundHandleEntry;
@@ -621,21 +802,12 @@ BOOL PSoundChannel::Setup()
       arg = val = entry.fragmentValue;
       ::ioctl(os_handle, SNDCTL_DSP_SETFRAGMENT, &arg); 
 
-      audio_buf_info info;
-      ::ioctl(os_handle, SNDCTL_DSP_GETOSPACE, &info);
-      PTRACE(6, "OSS\tOutput: fragments = " << info.fragments
-                     << ", total frags = " << info.fragstotal
-                     << ", frag size   = " << info.fragsize
-                     << ", bytes       = " << info.bytes);
-
-      ::ioctl(os_handle, SNDCTL_DSP_GETISPACE, &info);
-      PTRACE(6, "OSS\tInput: fragments = " << info.fragments
-                     << ", total frags = " << info.fragstotal
-                     << ", frag size   = " << info.fragsize
-                     << ", bytes       = " << info.bytes);
-
       mBitsPerSample = entry.bitsPerSample;
+#if PBYTE_ORDER == PLITTLE_ENDIAN
       arg = val = (entry.bitsPerSample == 16) ? AFMT_S16_LE : AFMT_S8;
+#else
+      arg = val = (entry.bitsPerSample == 16) ? AFMT_S16_BE : AFMT_S8;
+#endif
       if (ConvertOSError(::ioctl(os_handle, SNDCTL_DSP_SETFMT, &arg)) || (arg != val)) {
 
         mNumChannels = entry.numChannels;
@@ -650,6 +822,21 @@ BOOL PSoundChannel::Setup()
           }
         }
       }
+
+#if PTRACING
+      audio_buf_info info;
+      ::ioctl(os_handle, SNDCTL_DSP_GETOSPACE, &info);
+      PTRACE(4, "OSS\tOutput: fragments = " << info.fragments
+                     << ", total frags = " << info.fragstotal
+                     << ", frag size   = " << info.fragsize
+                     << ", bytes       = " << info.bytes);
+
+      ::ioctl(os_handle, SNDCTL_DSP_GETISPACE, &info);
+      PTRACE(4, "OSS\tInput: fragments = " << info.fragments
+                     << ", total frags = " << info.fragstotal
+                     << ", frag size   = " << info.fragsize
+                     << ", bytes       = " << info.bytes);
+#endif
     }
   }
 
@@ -727,7 +914,7 @@ BOOL PSoundChannel::Write(const void * buf, PINDEX len)
     return TRUE; // the write failed, but we return OK
   }
 
-  unsigned int index = 0;
+  PINDEX index = 0;
   while (index < len) {
     buffer[endptr++] = ((char *)buf)[index++];
     if (endptr == LOOPBACK_BUFFER_SIZE)
@@ -747,6 +934,26 @@ BOOL PSoundChannel::Read(void * buf, PINDEX len)
 
   if (os_handle > 0) {
     PTRACE(6, "OSS\tRead start");
+
+#if defined(P_FREEBSD)
+    PINDEX total = 0;
+    while (total < len) {
+      PINDEX bytes = 0;
+      while (!ConvertOSError(bytes = ::read(os_handle, (void *)(((unsigned int)buf) + total), len-total))) {
+        if (GetErrorCode() != Interrupted) {
+          PTRACE(6, "OSS\tRead failed");
+          return FALSE;
+        }
+        PTRACE(6, "OSS\tRead interrupted");
+      }
+      total += bytes;
+      if (total != len)
+        PTRACE(6, "OSS\tRead completed short - " << total << " vs " << len << ". Reading more data");
+    }
+    lastReadCount = total;
+
+#else
+
     while (!ConvertOSError(lastReadCount = ::read(os_handle, (void *)buf, len))) {
       if (GetErrorCode() != Interrupted) {
         PTRACE(6, "OSS\tRead failed");
@@ -754,6 +961,8 @@ BOOL PSoundChannel::Read(void * buf, PINDEX len)
       }
       PTRACE(6, "OSS\tRead interrupted");
     }
+#endif
+
 
     if (lastReadCount != len)
       PTRACE(6, "OSS\tRead completed short - " << lastReadCount << " vs " << len);
@@ -785,9 +994,9 @@ BOOL PSoundChannel::Read(void * buf, PINDEX len)
 
   // There may not be enough audio data in the buffer, so find out
   // how much we can copy.
-  unsigned int copy = ((len < bufferLen) ? len : bufferLen);
+  PINDEX copy = ((len < bufferLen) ? len : bufferLen);
 
-  for (unsigned int index = 0; index < copy; index++) {
+  for (PINDEX index = 0; index < copy; index++) {
     ((char *)buf)[index]=buffer[startptr++];
     if (startptr == LOOPBACK_BUFFER_SIZE)
       startptr = 0;
@@ -1045,12 +1254,9 @@ BOOL PSoundChannel::StartRecording()
   if (os_handle == 0)
     return TRUE;
 
-  fd_set fds;
-  FD_ZERO(&fds);
-  FD_SET(os_handle, &fds);
-
-  struct timeval instant = {0, 0};
-  return ConvertOSError(::select(1, &fds, NULL, NULL, &instant));
+  P_fd_set fds = os_handle;
+  P_timeval instant;
+  return ConvertOSError(::select(1, fds, NULL, NULL, instant));
 }
 
 
@@ -1123,7 +1329,7 @@ BOOL PSoundChannel::SetVolume(unsigned newVal)
   if (direction  == Player) 
     rc = ::ioctl(os_handle, MIXER_WRITE(SOUND_MIXER_VOLUME), &deviceVol);
    else 
-    rc = ::ioctl(os_handle, MIXER_WRITE(SOUND_MIXER_IGAIN), &deviceVol);
+    rc = ::ioctl(os_handle, MIXER_WRITE(SOUND_MIXER_MIC), &deviceVol);
 
   if (rc < 0) {
     PTRACE(1, "PSoundChannel::SetVolume failed : " << ::strerror(errno));
@@ -1142,7 +1348,7 @@ BOOL  PSoundChannel::GetVolume(unsigned &devVol)
   if (direction == Player)
     rc = ::ioctl(os_handle, MIXER_READ(SOUND_MIXER_VOLUME), &vol);
   else
-    rc = ::ioctl(os_handle, MIXER_READ(SOUND_MIXER_IGAIN), &vol);
+    rc = ::ioctl(os_handle, MIXER_READ(SOUND_MIXER_MIC), &vol);
   
   if (rc < 0) {
     PTRACE(1,  "PSoundChannel::GetVolume failed : " << ::strerror(errno)) ;
