@@ -27,6 +27,57 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: svcproc.cxx,v $
+ * Revision 1.82  2003/01/14 04:43:21  robertj
+ * Improved output on error in getting service status.
+ *
+ * Revision 1.81  2002/12/12 00:54:29  robertj
+ * Fixed issue with setting max file handles causing a log file to be created
+ *   as root (real uid) rather than as the user determined by the -u arg.
+ * Utilised new PProcess functions for setting user and group.
+ *
+ * Revision 1.80  2002/12/02 08:27:43  robertj
+ * Fixed incorrectly set #if statement from RTEMS patch.
+ *
+ * Revision 1.79  2002/12/02 03:57:18  robertj
+ * More RTEMS support patches, thank you Vladimir Nesic.
+ *
+ * Revision 1.78  2002/11/02 00:32:21  robertj
+ * Further fixes to VxWorks (Tornado) port, thanks Andreas Sikkema.
+ *
+ * Revision 1.77  2002/10/30 03:14:15  robertj
+ * Added tab character between time and thread in system log (for db import).
+ *
+ * Revision 1.76  2002/10/23 04:26:27  robertj
+ * Fixed extra debug levels in system log dump.
+ *
+ * Revision 1.75  2002/10/17 13:44:27  robertj
+ * Port to RTEMS, thanks Vladimir Nesic.
+ *
+ * Revision 1.74  2002/10/17 12:57:24  robertj
+ * Added ability to increase maximum file handles on a process.
+ *
+ * Revision 1.73  2002/10/10 04:43:44  robertj
+ * VxWorks port, thanks Martijn Roest
+ *
+ * Revision 1.72  2002/10/03 01:27:51  robertj
+ * Added fail safe _exit() as it appears raise(SIGQUIT) does not always dump
+ *   core and exit the app causing endless SEGV's to appear the log file!
+ *
+ * Revision 1.71  2002/10/01 06:03:44  robertj
+ * Attempt to prevent recursive signals for SEGV etc
+ *
+ * Revision 1.70  2002/09/06 13:29:06  robertj
+ * Added missing help line for status command.
+ *
+ * Revision 1.69  2002/07/15 06:39:23  craigs
+ * Added function to allow raising of per-process file handle limit
+ *
+ * Revision 1.68  2002/06/13 08:50:11  rogerh
+ * GCC 3.1 uses slightly different #includes
+ *
+ * Revision 1.67  2002/06/05 12:29:16  craigs
+ * Changes for gcc 3.1
+ *
  * Revision 1.66  2002/04/17 03:31:48  robertj
  * Renamed system log file member variable to be common with Windows version.
  *
@@ -185,11 +236,28 @@
 #pragma implementation "svcproc.h"
 #include <ptlib/svcproc.h>
 
+#ifdef P_VXWORKS
+#include <logLib.h>
+#define LOG_EMERG			0
+#define LOG_ALERT			1
+#define LOG_CRIT			2
+#define LOG_ERR				3
+#define LOG_WARNING			4
+#define	LOG_NOTICE			5
+#define LOG_INFO			6
+#define LOG_DEBUG			7
+#else
 #include <syslog.h>
-#include <stdarg.h>
-#include <fstream.h>
 #include <pwd.h>
 #include <grp.h>
+#endif
+
+#include <stdarg.h>
+#if (__GNUC__ >= 3)
+#include <fstream>
+#else
+#include <fstream.h>
+#endif
 #include <signal.h>
 
 #include "uerror.h"
@@ -203,6 +271,7 @@
 
 #define	MAX_LOG_LINE_LEN	1024
 
+#ifndef P_VXWORKS
 static int PwlibLogToUnixLog[PSystemLog::NumLogLevels] = {
   LOG_CRIT,    // LogFatal,   
   LOG_ERR,     // LogError,   
@@ -210,8 +279,12 @@ static int PwlibLogToUnixLog[PSystemLog::NumLogLevels] = {
   LOG_INFO,    // LogInfo,    
   LOG_DEBUG,   // LogDebug
   LOG_DEBUG,
+  LOG_DEBUG,
+  LOG_DEBUG,
+  LOG_DEBUG,
   LOG_DEBUG
 };
+#endif // !P_VXWORKS
 
 static const char * const PLevelName[PSystemLog::NumLogLevels+1] = {
   "Message",
@@ -221,7 +294,10 @@ static const char * const PLevelName[PSystemLog::NumLogLevels+1] = {
   "Info",
   "Debug",
   "Debug2",
-  "Debug3"
+  "Debug3",
+  "Debug4",
+  "Debug5",
+  "Debug6",
 };
 
 #ifdef P_MAC_MPTHREADS
@@ -253,8 +329,14 @@ static pthread_mutex_t logMutex = PTHREAD_MUTEX_INITIALIZER;
 void PSystemLog::Output(Level level, const char * cmsg)
 {
   PString systemLogFileName = PServiceProcess::Current().systemLogFileName;
-  if (systemLogFileName.IsEmpty())
+  if (systemLogFileName.IsEmpty()) {
+#ifdef P_VXWORKS
+    printf("%s\n",cmsg);
+    logMsg((char *)(const char)cmsg,0,0,0,0,0,0);
+#else
     syslog(PwlibLogToUnixLog[level], "%s", cmsg);
+#endif
+  }
   else {
 #ifdef P_PTHREADS
     pthread_mutex_lock(&logMutex);
@@ -271,12 +353,14 @@ void PSystemLog::Output(Level level, const char * cmsg)
       out = new ofstream(systemLogFileName, ios::app);
 
     PTime now;
-    *out << now.AsString("yyyy/MM/dd hh:mm:ss.uuu ");
+    *out << now.AsString("yyyy/MM/dd hh:mm:ss.uuu\t");
 
     PThread * thread = PThread::Current();
     if (thread == NULL) {
 #ifdef P_MAC_MPTHREADS
       unsigned tid = (unsigned)MPCurrentTaskID();
+#elif defined(P_VXWORKS)
+      unsigned tid = ::taskIdSelf();
 #else
       unsigned tid = (unsigned) pthread_self();
 #endif
@@ -361,9 +445,11 @@ PServiceProcess::~PServiceProcess()
   if (!pidFileToRemove)
     PFile::Remove(pidFileToRemove);
 
+#ifndef P_VXWORKS
   // close the system log
   if (systemLogFileName.IsEmpty())
     closelog();
+#endif // !P_VXWORKS
 }
 
 
@@ -375,6 +461,7 @@ PServiceProcess & PServiceProcess::Current()
 }
 
 
+#ifndef P_VXWORKS
 static int KillProcess(int pid, int sig)
 {
   if (kill(pid, sig) != 0)
@@ -399,20 +486,18 @@ static int KillProcess(int pid, int sig)
 
   return 1;
 }
+#endif // !P_VXWORKS
 
 
 void PServiceProcess::_PXShowSystemWarning(PINDEX code, const PString & str)
 {
-  PSYSTEMLOG(Warning, "PWLib/Unix error #"
-                      << code
-                      << "-"
-                      << str
-                      << endl);
+  PSYSTEMLOG(Warning, "PWLib\t" << GetOSClass() << " error #" << code << '-' << str);
 }
 
 
 int PServiceProcess::InitialiseService()
 {
+#ifndef P_VXWORKS
 #if PMEMORY_CHECK
   PMemoryHeap::SetIgnoreAllocations(TRUE);
 #endif
@@ -435,6 +520,7 @@ int PServiceProcess::InitialiseService()
              "h-help."
              "x-execute."
              "p-pid-file:"
+	     "H-handlemax:"
              "i-ini-file:"
              "k-kill."
              "t-terminate."
@@ -485,10 +571,14 @@ int PServiceProcess::InitialiseService()
     }
 
     if (args.HasOption('s')) {
-      cout << "Process at " << pid << " is ";
-      if (kill(pid, 0) != 0)
-        cout << "not ";
-      cout << "running." << endl;
+      cout << "Process at " << pid << ' ';
+      if (kill(pid, 0) == 0)
+        cout << "is running.";
+      else if (errno == ESRCH)
+        cout << "does not exist.";
+      else
+        cout << " status could not be determined, error: " << strerror(errno);
+      cout << endl;
       return 0;
     }
 
@@ -520,6 +610,24 @@ int PServiceProcess::InitialiseService()
     return 1;
   }
 
+  // Set the gid we are running under
+  if (args.HasOption('g')) {
+    PString gidstr = args.GetOptionString('g');
+    if (!SetGroupName(gidstr)) {
+      cout << "Could not set GID to \"" << gidstr << "\" - " << strerror(errno) << endl;
+      return 1;
+    }
+  }
+
+  // Set the uid we are running under
+  if (args.HasOption('u')) {
+    PString uidstr = args.GetOptionString('u');
+    if (!SetUserName(uidstr)) {
+      cout << "Could not set UID to \"" << uidstr << "\" - " << strerror(errno) << endl;
+      return 1;
+    }
+  }
+
   BOOL helpAndExit = FALSE;
 
   // if displaying help, then do it
@@ -550,7 +658,7 @@ int PServiceProcess::InitialiseService()
     cout << "usage: [-c] -v|-d|-h|-x\n"
             "  -h --help           output this help message and exit\n"
             "  -v --version        display version information and exit\n"
-#ifndef BE_THREADS
+#if !defined(BE_THREADS) && !defined(P_RTEMS)
             "  -d --daemon         run as a daemon\n"
 #endif
             "  -u --uid uid        set user id to run as\n"
@@ -558,11 +666,13 @@ int PServiceProcess::InitialiseService()
             "  -p --pid-file       name or directory for pid file\n"
             "  -t --terminate      orderly terminate process in pid file\n"
             "  -k --kill           preemptively kill process in pid file\n"
+            "  -s --status         check to see if daemon is running\n"
             "  -c --console        output messages to stdout rather than syslog\n"
             "  -l --log-file file  output messages to file or directory instead of syslog\n"
             "  -x --execute        execute as a normal program\n"
             "  -i --ini-file       set the ini file to use, may be explicit file or\n"
             "                      a ':' separated set of directories to search.\n"
+	    "  -H --handlemax n    set maximum number of file handles (set before uid/gid)\n"
 #ifdef P_LINUX
             "  -C --core-size      set the maximum core file size\n"
 #endif
@@ -570,74 +680,6 @@ int PServiceProcess::InitialiseService()
     return 0;
   }
 
-  if (args.HasOption('i'))
-    SetConfigurationPath(args.GetOptionString('i'));
-
-  // Set the gid we are running under
-  if (args.HasOption('g')) {
-    PString gidstr = args.GetOptionString('g');
-    if (gidstr.IsEmpty())
-      gidstr = "nobody";
-    int gid;
-    if (strspn(gidstr, "0123456789") == gidstr.GetLength())
-      gid = gidstr.AsInteger();
-    else {
-      struct group * gr = getgrnam(gidstr);
-      if (gr == NULL) {
-        cout << "Could not find group \"" << gidstr << '"' << endl;
-        return 1;
-      }
-      gid = gr->gr_gid;
-    }
-    if (setgid(gid) != 0) {
-      cout << "Could not set GID to \"" << gidstr << "\" (" << gid << ")"
-              " - " << strerror(errno) << endl;
-      return 1;
-    }
-  }
-
-  // Set the uid we are running under
-  if (args.HasOption('u')) {
-    PString uidstr = args.GetOptionString('u');
-    if (uidstr.IsEmpty())
-      uidstr = "nobody";
-    int uid;
-    if (strspn(uidstr, "0123456789") == uidstr.GetLength())
-      uid = uidstr.AsInteger();
-    else {
-      struct passwd * pw = getpwnam(uidstr);
-      if (pw == NULL) {
-        cout << "Could not find user \"" << uidstr << '"' << endl;
-        return 1;
-      }
-      uid = pw->pw_uid;
-    }
-    if (setuid(uid) != 0) {
-      cout << "Could not set UID to \"" << uidstr << "\" (" << uid << ")"
-              " - " << strerror(errno) << endl;
-      return 1;
-    }
-  }
-
-  // set the core file size
-  if (args.HasOption('C')) {
-#ifdef P_LINUX
-    struct rlimit rlim;
-    if (getrlimit(RLIMIT_CORE, &rlim) != 0) 
-      cout << "Could not get current core file size : error = " << errno << endl;
-    else {
-      int v = args.GetOptionString('C').AsInteger();
-      rlim.rlim_cur = v;
-      if (setrlimit(RLIMIT_CORE, &rlim) != 0) 
-        cout << "Could not set current core file size to " << v << " : error = " << errno << endl;
-      else {
-        getrlimit(RLIMIT_CORE, &rlim);
-        cout << "Core file size set to " << rlim.rlim_cur << "/" << rlim.rlim_max << endl;
-      }
-    }
-#endif
-  }
-
   // open the system logger for this program
   if (systemLogFileName.IsEmpty())
     openlog((char *)(const char *)GetName(), LOG_PID, LOG_DAEMON);
@@ -651,8 +693,41 @@ int PServiceProcess::InitialiseService()
       return 1;
     }
   }
+  PSYSTEMLOG(StdError, "Starting service process \"" << GetName() << "\" v" << GetVersion(TRUE));
 
-#ifndef BE_THREADS
+  if (args.HasOption('i'))
+    SetConfigurationPath(args.GetOptionString('i'));
+
+  if (args.HasOption('H')) {
+    int uid = geteuid();
+    seteuid(getuid()); // Switch back to starting uid for next call
+    SetMaxHandles(args.GetOptionString('H').AsInteger());
+    seteuid(uid);
+  }
+
+  // set the core file size
+  if (args.HasOption('C')) {
+#ifdef P_LINUX
+    struct rlimit rlim;
+    if (getrlimit(RLIMIT_CORE, &rlim) != 0) 
+      cout << "Could not get current core file size : error = " << errno << endl;
+    else {
+      int uid = geteuid();
+      seteuid(getuid()); // Switch back to starting uid for next call
+      int v = args.GetOptionString('C').AsInteger();
+      rlim.rlim_cur = v;
+      if (setrlimit(RLIMIT_CORE, &rlim) != 0) 
+        cout << "Could not set current core file size to " << v << " : error = " << errno << endl;
+      else {
+        getrlimit(RLIMIT_CORE, &rlim);
+        cout << "Core file size set to " << rlim.rlim_cur << "/" << rlim.rlim_max << endl;
+      }
+      seteuid(uid);
+    }
+#endif
+  }
+
+#if !defined(BE_THREADS) && !defined(P_RTEMS)
   if (!args.HasOption('d'))
     return -1;
 
@@ -714,16 +789,14 @@ int PServiceProcess::InitialiseService()
   // Also if in background, don't want to get blocked on input from stdin
   ::close(STDIN_FILENO);
 
-#endif // BE_THREADS
-
+#endif // !BE_THREADS && !P_RTEMS
+#endif // !P_VXWORKS
   return -1;
 }
 
 int PServiceProcess::_main(void *)
 {
   if ((terminationValue = InitialiseService()) < 0) {
-    PSYSTEMLOG(Warning, "Starting service process \"" << GetName() << "\" v" << GetVersion(TRUE));
-
     // Make sure housekeeping thread is going so signals are handled.
     SignalTimerChange();
 
@@ -775,9 +848,11 @@ void PServiceProcess::Terminate()
   // Do the services stop code
   OnStop();
 
+#ifndef P_VXWORKS
   // close the system log
   if (systemLogFileName.IsEmpty())
     closelog();
+#endif // !P_VXWORKS
 
   // Now end the program
   exit(terminationValue);
@@ -814,14 +889,22 @@ void PServiceProcess::PXOnAsyncSignal(int sig)
       return;
   }
 
+  signal(SIGSEGV, SIG_DFL);
+  signal(SIGFPE, SIG_DFL);
+  signal(SIGBUS, SIG_DFL);
+
   static BOOL inHandler = FALSE;
-  if (inHandler)
+  if (inHandler) {
     raise(SIGQUIT); // Dump core
+    _exit(-1); // Fail safe if raise() didn't dump core and exit
+  }
 
   inHandler = TRUE;
 
 #ifdef P_MAC_MPTHREADS
   unsigned tid = (unsigned)MPCurrentTaskID();
+#elif defined(P_VXWORKS)
+  unsigned tid = ::taskIdSelf();
 #else
   unsigned tid = (unsigned) pthread_self();
 #endif
@@ -843,15 +926,19 @@ void PServiceProcess::PXOnAsyncSignal(int sig)
   strcat(msg, ", aborting.\n");
 
   if (systemLogFileName.IsEmpty()) {
-    syslog(LOG_CRIT, 
-#ifdef __BEOS__ // (Some?) BeOS versions of syslog.h have syslog() (wrongly) declared without const 
-    (char *) 
-#endif 
-    msg); 
+#ifdef P_VXWORKS
+  logMsg((char *)msg,0,0,0,0,0,0);
+#else
+    syslog(LOG_CRIT, msg); 
     closelog();
+#endif // !P_VXWORKS
   }
   else {
+#ifdef P_VXWORKS
+    int fd = open(systemLogFileName, O_WRONLY|O_APPEND, FWRITE|FAPPEND);
+#else
     int fd = open(systemLogFileName, O_WRONLY|O_APPEND);
+#endif // !P_VXWORKS
     if (fd >= 0) {
       write(fd, msg, strlen(msg));
       close(fd);
@@ -859,6 +946,7 @@ void PServiceProcess::PXOnAsyncSignal(int sig)
   }
 
   raise(SIGQUIT); // Dump core
+  _exit(-1); // Fail safe if raise() didn't dump core and exit
 }
 
 
@@ -886,6 +974,7 @@ void PServiceProcess::PXOnSignal(int sig)
       }
       break;
 
+#ifdef SIGWINCH
     case SIGWINCH :
       if (currentLogLevel > PSystemLog::Fatal) {
         currentLogLevel = (PSystemLog::Level)(currentLogLevel-1);
@@ -893,6 +982,7 @@ void PServiceProcess::PXOnSignal(int sig)
         s << "Log level decreased to " << PLevelName[currentLogLevel+1];
       }
       break;
+#endif
   }
 }
 
