@@ -27,6 +27,147 @@
  * Contributor(s): ______________________________________.
  *
  * $Log: tlibthrd.cxx,v $
+ * Revision 1.121  2003/05/16 17:40:55  shawn
+ * On Mac OS X, thread with the highest priority should use fixed priority
+ * scheduling policy.  This avoids starvation caused by desktop activity.
+ *
+ * Revision 1.120  2003/05/02 00:58:40  dereks
+ * Add test for linux at the end of PMutex::Signal. Thanks Robert!
+ *
+ * Revision 1.119  2003/05/02 00:39:11  dereks
+ * Changes to make threading work on Redhat 9
+ *
+ * Revision 1.118  2003/04/24 12:03:13  rogerh
+ * Calling pthread_mutex_unlock() on a mutex which is not locked can be
+ * considered an error. NetBSD now enforce this error so we need to quickly
+ * try locking the mutex before unlocking it in ~PThread and ~PSemaphore.
+ *
+ * Revision 1.117  2003/04/08 03:29:31  robertj
+ * Fixed IsSuspeneded() so returns TRUE if thread not started yet, this makes
+ *   it the same as the Win32 semantics.
+ *
+ * Revision 1.116  2003/03/10 15:37:00  rogerh
+ * fix IsTerminated() function.
+ *
+ * Revision 1.115  2003/03/07 00:07:15  robertj
+ * Fixed Mac OS X patch which broke every other platform.
+ *
+ * Revision 1.114  2003/03/06 08:58:48  rogerh
+ * P_MACOSX now carries the OSRELEASE. Use this to enable better threads
+ * support on Darwin 6.4. Submitted by Shawn.
+ *
+ * Revision 1.113  2003/02/20 23:32:00  robertj
+ * More RTEMS support patches, thanks Sebastian Meyer.
+ *
+ * Revision 1.112  2003/01/24 10:21:06  robertj
+ * Fixed issues in RTEMS support, thanks Vladimir Nesic
+ *
+ * Revision 1.111  2003/01/20 10:13:18  rogerh
+ * NetBSD thread changes
+ *
+ * Revision 1.110  2003/01/20 10:05:46  rogerh
+ * NetBSD thread changes
+ *
+ * Revision 1.109  2003/01/08 08:47:51  rogerh
+ * Add new Sleep() function for GNU PTH threads.
+ * Taken from NetBSD's package which uses PTH.
+ * Note: I am not sure this works correctly.
+ *
+ * Revision 1.108  2003/01/06 18:49:15  rogerh
+ * Back out pthead_kill to pthread_cancel change on NetBSD
+ *
+ * Revision 1.107  2002/12/11 05:39:26  robertj
+ * Added logging for file handle changes.
+ * Fixd bug where internal maxHandles not set when increased.
+ *
+ * Revision 1.106  2002/12/02 03:57:18  robertj
+ * More RTEMS support patches, thank you Vladimir Nesic.
+ *
+ * Revision 1.105  2002/11/22 10:14:07  robertj
+ * QNX port, thanks Xiaodan Tang
+ *
+ * Revision 1.104  2002/11/04 16:01:27  rogerh
+ * Using pthread_cancel and not pthread_kill with SIGKILL to terminate a thread
+ * On FreeBSD the thread does not have a handler for SIGKILL, it passes it up
+ * to the main process which gets killed! Assume the other BSDs are the same.
+ *
+ * Revision 1.103  2002/10/24 00:40:56  robertj
+ * Put back ability to terminate a thread from that threads context (removed
+ *   in revision 1.101) but requires that destructor not do so.
+ * Changed pipe close to allow for possible EINTR, and retry close.
+ *
+ * Revision 1.102  2002/10/24 00:25:13  robertj
+ * Changed high load thread problem fix from the termination function to start
+ *   function to finally, once and for all (I hope!) fix the race condition.
+ *
+ * Revision 1.101  2002/10/23 14:56:22  craigs
+ * Fixed problem with pipe leak under some circumstances
+ *
+ * Revision 1.100  2002/10/23 04:29:32  robertj
+ * Improved debugging for thread create/start/stop/destroy.
+ * Fixed race condition bug if auto-delete thread starts and completes before
+ *   pthread_create returns, PX_threadId is not set yet!
+ *
+ * Revision 1.99  2002/10/22 07:42:52  robertj
+ * Added extra debugging for file handle and thread leak detection.
+ *
+ * Revision 1.98  2002/10/18 03:05:39  robertj
+ * Fixed thread leak caused by fixing the thread crash a few revisions back,
+ *   caused by strange pthreads behaviour, at least under Linux.
+ *
+ * Revision 1.97  2002/10/17 13:44:27  robertj
+ * Port to RTEMS, thanks Vladimir Nesic.
+ *
+ * Revision 1.96  2002/10/17 12:57:24  robertj
+ * Added ability to increase maximum file handles on a process.
+ *
+ * Revision 1.95  2002/10/16 11:26:29  rogerh
+ * Add missing include. Noticed by Matthias on the GnomeMeeting IRC
+ *
+ * Revision 1.94  2002/10/10 03:09:48  robertj
+ * Fixed high load race condition when starting threads.
+ *
+ * Revision 1.93  2002/10/05 05:22:43  robertj
+ * Fixed adding GetThreadId() function.
+ *
+ * Revision 1.92  2002/10/01 06:27:48  robertj
+ * Added bullet proofing against possible EINTR error returns on all pthread
+ *   functions when under heavy load. Linux really should NOT do this, but ...
+ *
+ * Revision 1.91  2002/09/04 03:14:18  robertj
+ * Backed out changes submitted by Martin Froehlich as they do not appear to
+ *   actually do anything other than add a sychronisation point. The variables
+ *   the patches intended to protect were already protected.
+ * Fixed bug where if a PMutex was signalled by a thread that did not have it
+ *   locked, it would assert but continue to alter PMutex variables such that
+ *   a deadlock or crash is likely.
+ *
+ * Revision 1.90  2002/08/29 01:50:40  robertj
+ * Changed the pthread_create so does retries if get EINTR or EAGAIN errors
+ *   which indicate a (possibly) temporary resource limit.
+ * Enabled and adjusted tracing.
+ *
+ * Revision 1.89  2002/08/22 13:05:57  craigs
+ * Fixed problems with mutex implementation thanks to Martin Froehlich
+ *
+ * Revision 1.88  2002/07/15 06:56:59  craigs
+ * Fixed missing brace
+ *
+ * Revision 1.87  2002/07/15 06:39:23  craigs
+ * Added function to allow raising of per-process file handle limit
+ *
+ * Revision 1.86  2002/06/27 06:38:58  robertj
+ * Changes to remove memory leak display for things that aren't memory leaks.
+ *
+ * Revision 1.85  2002/06/27 02:04:01  robertj
+ * Fixed NetBSD compatibility issue, thanks Motoyuki OHMORI.
+ *
+ * Revision 1.84  2002/06/04 00:25:31  robertj
+ * Fixed incorrectly initialised trace indent, thanks Artis Kugevics
+ *
+ * Revision 1.83  2002/05/21 09:13:00  robertj
+ * Fixed problem when using posix recursive mutexes, thanks Artis Kugevics
+ *
  * Revision 1.82  2002/04/24 01:11:37  robertj
  * Fixed problem with PTRACE_BLOCK indent level being correct across threads.
  *
@@ -292,11 +433,62 @@
  *
  */
 
+#include <ptlib/socket.h>
+#include <sched.h>	// for sched_yield
 #include <pthread.h>
 #include <sys/resource.h>
 
-
+#ifdef P_RTEMS
+#define	SUSPEND_SIG SIGALRM
+#include <sched.h>
+#else
 #define	SUSPEND_SIG SIGVTALRM
+#endif
+
+#ifdef P_MACOSX
+#include <mach/mach.h>
+#include <mach/thread_policy.h>
+#include <sys/param.h>
+#include <sys/sysctl.h>
+// going to need the main thread for adjusting relative priority
+static pthread_t baseThread;
+#endif
+
+int PX_NewHandle(const char *, int);
+
+
+#define PAssertPTHREAD(func, args) \
+  { \
+    unsigned threadOpRetry = 0; \
+    while (PAssertThreadOp(func args, threadOpRetry, #func, __FILE__, __LINE__)); \
+  }
+
+static BOOL PAssertThreadOp(int retval,
+                            unsigned & retry,
+                            const char * funcname,
+                            const char * file,
+                            unsigned line)
+{
+  if (retval == 0) {
+    PTRACE_IF(2, retry > 0, "PWLib\t" << funcname << " required " << retry << " retries!");
+    return FALSE;
+  }
+
+  if (errno == EINTR || errno == EAGAIN) {
+    if (++retry < 1000) {
+#if defined(P_RTEMS)
+      sched_yield();
+#else
+      usleep(10000); // Basically just swap out thread to try and clear blockage
+#endif
+      return TRUE;   // Return value to try again
+    }
+    // Give up and assert
+  }
+
+  PAssertFunc(file, line, NULL, psprintf("Function %s failed", funcname));
+  return FALSE;
+}
 
 
 PDECLARE_CLASS(PHouseKeepingThread, PThread)
@@ -329,12 +521,9 @@ void PHouseKeepingThread::Main()
 
     int fd = process.timerChangePipe[0];
 
-    fd_set read_fds;
-    FD_ZERO(&read_fds);
-    FD_SET(fd, &read_fds);
-
-    struct timeval tval;
-    if (::select(fd+1, &read_fds, NULL, NULL, delay.AsTimeVal(tval)) == 1) {
+    P_fd_set read_fds = fd;
+    P_timeval tval = delay;
+    if (::select(fd+1, read_fds, NULL, NULL, tval) == 1) {
       BYTE ch;
       ::read(fd, &ch, 1);
     }
@@ -348,11 +537,11 @@ void PProcess::SignalTimerChange()
 {
   if (housekeepingThread == NULL) {
 #if PMEMORY_CHECK
-  PMemoryHeap::SetIgnoreAllocations(TRUE);
+    BOOL oldIgnoreAllocations = PMemoryHeap::SetIgnoreAllocations(TRUE);
 #endif
     housekeepingThread = new PHouseKeepingThread;
 #if PMEMORY_CHECK
-  PMemoryHeap::SetIgnoreAllocations(FALSE);
+    PMemoryHeap::SetIgnoreAllocations(oldIgnoreAllocations);
 #endif
   }
 
@@ -363,21 +552,56 @@ void PProcess::SignalTimerChange()
 
 void PProcess::Construct()
 {
-  // set the file descriptor limit to something sensible
+#ifndef P_RTEMS
+  // get the file descriptor limit
   struct rlimit rl;
   PAssertOS(getrlimit(RLIMIT_NOFILE, &rl) == 0);
-  rl.rlim_cur = rl.rlim_max;
-  PAssertOS(setrlimit(RLIMIT_NOFILE, &rl) == 0);
+  maxHandles = rl.rlim_cur;
+  PTRACE(4, "PWLib\tMaximum per-process file handles is " << maxHandles);
 
   ::pipe(timerChangePipe);
+#else
+  maxHandles = 500; // arbitrary value
+  socketpair(AF_INET,SOCK_STREAM,0,timerChangePipe);
+#endif
 
   // initialise the housekeeping thread
   housekeepingThread = NULL;
 
+#ifdef P_MACOSX
+    // records the main thread for priority adjusting
+    baseThread = pthread_self();
+#endif
+
   CommonConstruct();
 }
 
 
+BOOL PProcess::SetMaxHandles(int newMax)
+{
+#ifndef P_RTEMS
+  // get the current process limit
+  struct rlimit rl;
+  PAssertOS(getrlimit(RLIMIT_NOFILE, &rl) == 0);
+
+  // set the new current limit
+  rl.rlim_cur = newMax;
+  if (setrlimit(RLIMIT_NOFILE, &rl) == 0) {
+    PAssertOS(getrlimit(RLIMIT_NOFILE, &rl) == 0);
+    maxHandles = rl.rlim_cur;
+    if (maxHandles == newMax) {
+      PTRACE(2, "PWLib\tNew maximum per-process file handles set to " << maxHandles);
+      return TRUE;
+    }
+  }
+#endif // !P_RTEMS
+
+  PTRACE(1, "PWLib\tCannot set per-process file handle limit to "
+         << newMax << " (is " << maxHandles << ") - check permissions");
+  return FALSE;
+}
+
+
 PProcess::~PProcess()
 {
   // Don't wait for housekeeper to stop if Terminate() is called from it.
@@ -415,14 +639,16 @@ void PThread::InitialiseProcessThread()
 
   PX_suspendMutex = MutexInitialiser;
 
+#ifdef P_RTEMS
+  PAssertOS(socketpair(AF_INET,SOCK_STREAM,0,unblockPipe) == 0);
+#else
   PAssertOS(::pipe(unblockPipe) == 0);
+#endif
 
   ((PProcess *)this)->activeThreads.DisallowDeleteObjects();
   ((PProcess *)this)->activeThreads.SetAt((unsigned)PX_threadId, this);
 
-#if PTRACING
   traceBlockIndentLevel = 0;
-#endif
 }
 
 
@@ -447,36 +673,42 @@ PThread::PThread(PINDEX stackSize,
 
   PX_suspendMutex = MutexInitialiser;
 
+#ifdef P_RTEMS
+  PAssertOS(socketpair(AF_INET,SOCK_STREAM,0,unblockPipe) == 0);
+#else
   PAssertOS(::pipe(unblockPipe) == 0);
+#endif
+  PX_NewHandle("Thread unblock pipe", PMAX(unblockPipe[0], unblockPipe[1]));
 
   // new thread is actually started the first time Resume() is called.
   PX_firstTimeStart = TRUE;
 
-#if PTRACING
   traceBlockIndentLevel = 0;
-#endif
+
+  PTRACE(5, "PWLib\tCreated thread " << this << ' ' << threadName);
 }
 
 
 PThread::~PThread()
 {
-  if (!IsTerminated()) 
+  if (PX_threadId != 0 && PX_threadId != pthread_self())
     Terminate();
 
-  ::close(unblockPipe[0]);
-  ::close(unblockPipe[1]);
+  PAssertPTHREAD(::close, (unblockPipe[0]));
+  PAssertPTHREAD(::close, (unblockPipe[1]));
 
 #ifndef P_HAS_SEMAPHORES
   pthread_mutex_destroy(&PX_WaitSemMutex);
 #endif
 
+#ifdef P_NETBSD
+  // If the mutex was not locked, the unlock will fail */
+  pthread_mutex_trylock(&PX_suspendMutex);
+#endif
+  pthread_mutex_unlock(&PX_suspendMutex);
   pthread_mutex_destroy(&PX_suspendMutex);
-}
 
-
-pthread_t PThread::PX_GetThreadId() const
-{
-  return PX_threadId;
+  PTRACE(5, "PWLib\tDestroyed thread " << this << ' ' << threadName);
 }
 
 
@@ -489,7 +721,7 @@ void PThread::Restart()
   pthread_attr_init(&threadAttr);
   pthread_attr_setdetachstate(&threadAttr, PTHREAD_CREATE_DETACHED);
 
-#ifdef P_LINUX
+#if defined(P_LINUX)
   /*
     Set realtime scheduling if our effective user id is root (only then is this
     allowed) AND our priority is Highest.
@@ -502,10 +734,24 @@ void PThread::Restart()
     neccessary.
   */
   if ((geteuid() == 0) && (PX_priority == HighestPriority))
-    PAssertOS(pthread_attr_setschedpolicy(&threadAttr, SCHED_FIFO) == 0);
+    PAssertPTHREAD(pthread_attr_setschedpolicy, (&threadAttr, SCHED_FIFO));
+#elif defined(P_RTEMS)
+  pthread_attr_setstacksize(&threadAttr, 2*PTHREAD_MINIMUM_STACK_SIZE);
+  pthread_attr_setinheritsched(&threadAttr, PTHREAD_EXPLICIT_SCHED);
+  pthread_attr_setschedpolicy(&threadAttr, SCHED_OTHER);
+  struct sched_param sched_param;
+  sched_param.sched_priority = 125; /* set medium priority */
+  pthread_attr_setschedparam(&threadAttr, &sched_param);
 #endif
 
-  PAssertOS(pthread_create(&PX_threadId, &threadAttr, PX_ThreadStart, this) == 0);
+  PAssertPTHREAD(pthread_create, (&PX_threadId, &threadAttr, PX_ThreadStart, this));
+
+#ifdef P_MACOSX
+  if (PX_priority == HighestPriority) {
+    PTRACE(1, "set thread to have the highest priority (MACOSX)");
+    SetPriority(HighestPriority);
+  }
+#endif
 }
 
 
@@ -519,14 +765,16 @@ void PX_SuspendSignalHandler(int)
   while (notResumed) {
     BYTE ch;
     notResumed = ::read(thread->unblockPipe[0], &ch, 1) < 0 && errno == EINTR;
+#if !( defined(P_NETBSD) && defined(P_NO_CANCEL) )
     pthread_testcancel();
+#endif
   }
 }
 
 
 void PThread::Suspend(BOOL susp)
 {
-  PAssertOS(pthread_mutex_lock(&PX_suspendMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&PX_suspendMutex));
 
   // Check for start up condition, first time Resume() is called
   if (PX_firstTimeStart) {
@@ -541,11 +789,11 @@ void PThread::Suspend(BOOL susp)
       }
     }
 
-    PAssertOS(pthread_mutex_unlock(&PX_suspendMutex) == 0);
+    PAssertPTHREAD(pthread_mutex_unlock, (&PX_suspendMutex));
     return;
   }
 
-#ifdef P_MACOSX
+#if defined(P_MACOSX) && (P_MACOSX <= 55)
   // Suspend - warn the user with an Assertion
   PAssertAlways("Cannot suspend threads on Mac OS X due to lack of pthread_kill()");
 #else
@@ -560,7 +808,7 @@ void PThread::Suspend(BOOL susp)
           pthread_kill(PX_threadId, SUSPEND_SIG);
         }
         else {
-          PAssertOS(pthread_mutex_unlock(&PX_suspendMutex) == 0);
+          PAssertPTHREAD(pthread_mutex_unlock, (&PX_suspendMutex));
           PX_SuspendSignalHandler(SUSPEND_SIG);
           return;  // Mutex already unlocked
         }
@@ -575,7 +823,7 @@ void PThread::Suspend(BOOL susp)
     }
   }
 
-  PAssertOS(pthread_mutex_unlock(&PX_suspendMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&PX_suspendMutex));
 #endif // P_MACOSX
 }
 
@@ -588,12 +836,15 @@ void PThread::Resume()
 
 BOOL PThread::IsSuspended() const
 {
+  if (PX_firstTimeStart)
+    return TRUE;
+
   if (IsTerminated())
     return FALSE;
 
-  PAssertOS(pthread_mutex_lock((pthread_mutex_t *)&PX_suspendMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, ((pthread_mutex_t *)&PX_suspendMutex));
   BOOL suspended = PX_suspendCount != 0;
-  PAssertOS(pthread_mutex_unlock((pthread_mutex_t *)&PX_suspendMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, ((pthread_mutex_t *)&PX_suspendMutex));
   return suspended;
 }
 
@@ -604,12 +855,67 @@ void PThread::SetAutoDelete(AutoDeleteFlag deletion)
   autoDelete = deletion == AutoDeleteThread;
 }
 
+#ifdef P_MACOSX
+// obtain thread priority of the main thread
+static unsigned long
+GetThreadBasePriority ()
+{
+    thread_basic_info_data_t threadInfo;
+    policy_info_data_t       thePolicyInfo;
+    unsigned int             count;
+
+    if (baseThread == 0) {
+      return 0;
+    }
+
+    // get basic info
+    count = THREAD_BASIC_INFO_COUNT;
+    thread_info (pthread_mach_thread_np (baseThread), THREAD_BASIC_INFO,
+		 (integer_t*)&threadInfo, &count);
+
+    switch (threadInfo.policy) {
+    case POLICY_TIMESHARE:
+        count = POLICY_TIMESHARE_INFO_COUNT;
+	thread_info(pthread_mach_thread_np (baseThread),
+		    THREAD_SCHED_TIMESHARE_INFO,
+		    (integer_t*)&(thePolicyInfo.ts), &count);
+	return thePolicyInfo.ts.base_priority;
+	break;
+
+    case POLICY_FIFO:
+        count = POLICY_FIFO_INFO_COUNT;
+	thread_info(pthread_mach_thread_np (baseThread),
+		    THREAD_SCHED_FIFO_INFO,
+		    (integer_t*)&(thePolicyInfo.fifo), &count);
+	if (thePolicyInfo.fifo.depressed) {
+	    return thePolicyInfo.fifo.depress_priority;
+	} else {
+	    return thePolicyInfo.fifo.base_priority;
+	}
+	break;
+
+    case POLICY_RR:
+      count = POLICY_RR_INFO_COUNT;
+      thread_info(pthread_mach_thread_np (baseThread),
+		  THREAD_SCHED_RR_INFO,
+		  (integer_t*)&(thePolicyInfo.rr), &count);
+      if (thePolicyInfo.rr.depressed) {
+	  return thePolicyInfo.rr.depress_priority;
+      } else {
+	  return thePolicyInfo.rr.base_priority;
+      }
+      break;
+    }
+
+    return 0;
+}
+#endif
 
 void PThread::SetPriority(Priority priorityLevel)
 {
   PX_priority = priorityLevel;
 
-#ifdef P_LINUX
+#if defined(P_LINUX)
   if (IsTerminated())
     return;
 
@@ -618,13 +924,56 @@ void PThread::SetPriority(Priority priorityLevel)
   if ((priorityLevel == HighestPriority) && (geteuid() == 0) ) {
     sched_param.sched_priority = sched_get_priority_min( SCHED_FIFO );
     
-    PAssertOS(pthread_setschedparam(PX_threadId, SCHED_FIFO, &sched_param) == 0);
+    PAssertPTHREAD(pthread_setschedparam, (PX_threadId, SCHED_FIFO, &sched_param));
   }
   else if (priorityLevel != HighestPriority) {
     /* priority 0 is the only permitted value for the SCHED_OTHER scheduler */ 
     sched_param.sched_priority = 0;
     
-    PAssertOS(pthread_setschedparam(PX_threadId, SCHED_OTHER, &sched_param) == 0);
+    PAssertPTHREAD(pthread_setschedparam, (PX_threadId, SCHED_OTHER, &sched_param));
+  }
+#endif
+
+#if defined(P_MACOSX)
+  if (IsTerminated())
+    return;
+
+  if (priorityLevel == HighestPriority) {
+    /* get fixed priority */
+    {
+      int result;
+
+      thread_extended_policy_data_t   theFixedPolicy;
+      thread_precedence_policy_data_t thePrecedencePolicy;
+      long                            relativePriority;
+
+      theFixedPolicy.timeshare = false; // set to true for a non-fixed thread
+      result = thread_policy_set (pthread_mach_thread_np(PX_threadId),
+				  THREAD_EXTENDED_POLICY,
+				  (thread_policy_t)&theFixedPolicy,
+				  THREAD_EXTENDED_POLICY_COUNT);
+      if (result != KERN_SUCCESS) {
+	PTRACE(1, "thread_policy - Couldn't set thread as fixed priority.");
+      }
+
+      // set priority
+
+      // precedency policy's "importance" value is relative to
+      // spawning thread's priority
+      
+      relativePriority = 62 - GetThreadBasePriority();
+      PTRACE(1,  "relativePriority is " << relativePriority <<
+	     " base priority is " << GetThreadBasePriority());
+      
+      thePrecedencePolicy.importance = relativePriority;
+      result = thread_policy_set (pthread_mach_thread_np(PX_threadId),
+				  THREAD_PRECEDENCE_POLICY,
+				  (thread_policy_t)&thePrecedencePolicy, 
+				  THREAD_PRECEDENCE_POLICY_COUNT);
+      if (result != KERN_SUCCESS) {
+	PTRACE(1, "thread_policy - Couldn't set thread priority.");
+      }
+    }
   }
 #endif
 }
@@ -632,12 +981,11 @@ void PThread::SetPriority(Priority priorityLevel)
 
 PThread::Priority PThread::GetPriority() const
 {
-#ifdef LINUX
+#if defined(LINUX)
   int schedulingPolicy;
   struct sched_param schedParams;
   
-  PAssertOS( pthread_getschedparam( PX_threadId, &schedulingPolicy, 
-                                    &schedParams ) );
+  PAssertPTHREAD(pthread_getschedparam, (PX_threadId, &schedulingPolicy, &schedParams));
   
   switch( schedulingPolicy )
   {
@@ -650,8 +998,7 @@ PThread::Priority PThread::GetPriority() const
       
     default:
       /* Unknown scheduler. We don't know what priority this thread has. */
-      PTRACE( 1, "PWLib\tPThread::GetPriority: unknown scheduling policy #" 
-              << schedulingPolicy );
+      PTRACE(1, "PWLib\tPThread::GetPriority: unknown scheduling policy #" << schedulingPolicy);
   }
 #endif
 
@@ -662,29 +1009,54 @@ PThread::Priority PThread::GetPriority() const
 #ifndef P_HAS_SEMAPHORES
 void PThread::PXSetWaitingSemaphore(PSemaphore * sem)
 {
-  PAssertOS(pthread_mutex_lock(&PX_WaitSemMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&PX_WaitSemMutex));
   PX_waitingSemaphore = sem;
-  PAssertOS(pthread_mutex_unlock(&PX_WaitSemMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&PX_WaitSemMutex));
 }
 #endif
 
 
+#ifdef P_GNU_PTH
+// GNU PTH threads version (used by NetBSD)
+// Taken from NetBSD pkg patches
 void PThread::Sleep(const PTimeInterval & timeout)
 {
   PTime lastTime;
-  PTime targetTime = lastTime + timeout;
-  do {
-    PTimeInterval delay = targetTime - lastTime;
-    struct timeval tval;
-    if (select(0, NULL, NULL, NULL, delay.AsTimeVal(tval)) < 0 && errno != EINTR)
+  PTime targetTime = PTime() + timeout;
+
+  sched_yield();
+  lastTime = PTime();
+
+  while (lastTime < targetTime) {
+    P_timeval tval = targetTime - lastTime;
+    if (select(0, NULL, NULL, NULL, tval) < 0 && errno != EINTR)
       break;
 
     pthread_testcancel();
 
     lastTime = PTime();
-  } while (lastTime < targetTime);
+  }
 }
 
+#else
+// Normal Posix threads version
+void PThread::Sleep(const PTimeInterval & timeout)
+{
+  PTime lastTime;
+  PTime targetTime = lastTime + timeout;
+  do {
+    P_timeval tval = targetTime - lastTime;
+    if (select(0, NULL, NULL, NULL, tval) < 0 && errno != EINTR)
+      break;
+
+#if !( defined(P_NETBSD) && defined(P_NO_CANCEL) )
+    pthread_testcancel();
+#endif
+
+    lastTime = PTime();
+  } while (lastTime < targetTime);
+}
+#endif
 
 void PThread::Yield()
 {
@@ -707,54 +1079,56 @@ void PThread::Terminate()
   if (PX_origStackSize <= 0)
     return;
 
+  // don't use PThread::Current, as the thread may already not be in the
+  // active threads list
+  if (PX_threadId == pthread_self()) {
+    pthread_exit(0);
+    return;
+  }
+
   if (IsTerminated())
     return;
 
-  PTRACE(1, "PWLib\tForcing termination of thread " << (void *)this);
+  PTRACE(2, "PWLib\tForcing termination of thread " << (void *)this);
 
-  if (Current() == this)
-    pthread_exit(NULL);
-  else {
-    PXAbortBlock();
-    WaitForTermination(20);
+  PXAbortBlock();
+  WaitForTermination(20);
 
 #ifndef P_HAS_SEMAPHORES
-    PAssertOS(pthread_mutex_lock(&PX_WaitSemMutex) == 0);
-    if (PX_waitingSemaphore != NULL) {
-      PAssertOS(pthread_mutex_lock(&PX_waitingSemaphore->mutex) == 0);
-      PX_waitingSemaphore->queuedLocks--;
-      PAssertOS(pthread_mutex_unlock(&PX_waitingSemaphore->mutex) == 0);
-      PX_waitingSemaphore = NULL;
-    }
-    PAssertOS(pthread_mutex_unlock(&PX_WaitSemMutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&PX_WaitSemMutex));
+  if (PX_waitingSemaphore != NULL) {
+    PAssertPTHREAD(pthread_mutex_lock, (&PX_waitingSemaphore->mutex));
+    PX_waitingSemaphore->queuedLocks--;
+    PAssertPTHREAD(pthread_mutex_unlock, (&PX_waitingSemaphore->mutex));
+    PX_waitingSemaphore = NULL;
+  }
+  PAssertPTHREAD(pthread_mutex_unlock, (&PX_WaitSemMutex));
 #endif
 
-#if defined(P_FREEBSD) || defined(P_OPENBSD) || defined(P_NETBSD) || defined (P_AIX)
-    pthread_kill(PX_threadId, SIGKILL);
+#if ( defined(P_NETBSD) && defined(P_NO_CANCEL) )
+  pthread_kill(PX_threadId,SIGKILL);
 #else
+  if (PX_threadId) {
     pthread_cancel(PX_threadId);
-#endif
   }
+#endif
 }
 
 
 BOOL PThread::IsTerminated() const
 {
-  if (PX_threadId == 0) {
-    //PTRACE(1, "PWLib\tIsTerminated(" << (void *)this << ") = 0");
+  if (PX_threadId == 0)
     return TRUE;
-  }
 
-#ifndef P_MACOSX
-  // MacOS X does not support pthread_kill so we cannot use it
+#if defined(P_MACOSX) && (P_MACOSX <= 55)
+  // MacOS X (darwin 5.5) does not support pthread_kill so we cannot use it
   // to test the validity of the thread
-  if (pthread_kill(PX_threadId, 0) != 0)  {
-    //PTRACE(1, "PWLib\tIsTerminated(" << (void *)this << ") terminated");
+#else
+  if (pthread_kill(PX_threadId, 0) != 0)
     return TRUE;
-  }
 #endif
 
-  //PTRACE(1, "PWLib\tIsTerminated(" << (void *)this << ") not dead yet");
+  PTRACE(7, "PWLib\tIsTerminated(" << (void *)this << ") not dead yet");
   return FALSE;
 }
 
@@ -774,7 +1148,7 @@ BOOL PThread::WaitForTermination(const PTimeInterval & maxWait) const
 {
   PAssert(Current() != this, "Waiting for self termination!");
   
-  //PTRACE(1, "PWLib\tWaitForTermination(" << maxWait << ')');
+  PTRACE(6, "PWLib\tWaitForTermination(" << maxWait << ')');
 
   PTimer timeout = maxWait;
   while (!IsTerminated()) {
@@ -795,20 +1169,36 @@ void * PThread::PX_ThreadStart(void * arg)
   pthread_detach(threadId);
 
   PThread * thread = (PThread *)arg;
+
+  // Added this to guarantee that the thread creation (PThread::Restart)
+  // has completed before we start the thread. Then the PX_threadId has
+  // been set.
+  pthread_mutex_lock(&thread->PX_suspendMutex);
+
   thread->SetThreadName(thread->GetThreadName());
 
+  pthread_mutex_unlock(&thread->PX_suspendMutex);
+
   PProcess & process = PProcess::Current();
 
+  PINDEX newHighWaterMark = 0;
+  static PINDEX highWaterMark = 0;
+
   // add thread to thread list
   process.threadMutex.Wait();
   process.activeThreads.SetAt((unsigned)threadId, thread);
+  if (process.activeThreads.GetSize() > highWaterMark)
+    newHighWaterMark = highWaterMark = process.activeThreads.GetSize();
   process.threadMutex.Signal();
 
+  PTRACE_IF(4, newHighWaterMark > 0, "PWLib\tThread high water mark set: " << newHighWaterMark);
+
   // make sure the cleanup routine is called when the thread exits
   pthread_cleanup_push(PThread::PX_ThreadEnd, arg);
 
+  PTRACE(5, "PWLib\tStarted thread " << thread << ' ' << thread->threadName);
+
   // now call the the thread main routine
-  //PTRACE(1, "PWLib\tAbout to call Main");
   thread->Main();
 
   // execute the cleanup routine
@@ -821,21 +1211,28 @@ void * PThread::PX_ThreadStart(void * arg)
 void PThread::PX_ThreadEnd(void * arg)
 {
   PThread * thread = (PThread *)arg;
-  PProcess & process = PProcess::Current();
-  
-  pthread_t id = thread->PX_GetThreadId();
-  if (id != 0) {
-
-    // remove this thread from the active thread list
-    process.threadMutex.Wait();
-    process.activeThreads.SetAt((unsigned)id, NULL);
-    process.threadMutex.Signal();
+  pthread_t id = thread->GetThreadId();
+  if (id == 0) {
+    // Don't know why, but pthreads under Linux at least can call this function
+    // multiple times! Probably a bug, but we have to allow for it.
+    PTRACE(2, "PWLib\tAttempted to multiply end thread " << thread << " ThreadID=" << (void *)id);
+    return;
   }
 
+  PTRACE(5, "PWLib\tEnded thread " << thread << ' ' << thread->threadName);
+
+  // remove this thread from the active thread list
+  PProcess & process = PProcess::Current();
+  process.threadMutex.Wait();
+  process.activeThreads.SetAt((unsigned)id, NULL);
+  process.threadMutex.Signal();
+
   // delete the thread if required, note this is done this way to avoid
   // a race condition, the thread ID cannot be zeroed before the if!
   if (thread->autoDelete) {
     thread->PX_threadId = 0;  // Prevent terminating terminated thread
+
+    // Now should be safe to delete the thread!
     delete thread;
   }
   else
@@ -845,37 +1242,37 @@ void PThread::PX_ThreadEnd(void * arg)
 
 int PThread::PXBlockOnIO(int handle, int type, const PTimeInterval & timeout)
 {
-  //PTRACE(1,"PWLib\tPThread::PXBlockOnIO(" << handle << ',' << type << ')');
+  PTRACE(7, "PWLib\tPThread::PXBlockOnIO(" << handle << ',' << type << ')');
 
-  if ((handle < 0) || (handle >= FD_SETSIZE)) {
+  if ((handle < 0) || (handle >= PProcess::Current().GetMaxHandles())) {
+    PTRACE(2, "PWLib\tAttempt to use illegal handle in PThread::PXBlockOnIO, handle=" << handle);
     errno = EBADF;
     return -1;
   }
 
   // make sure we flush the buffer before doing a write
-  fd_set tmp_rfd, tmp_wfd, tmp_efd;
-  fd_set * read_fds      = &tmp_rfd;
-  fd_set * write_fds     = &tmp_wfd;
-  fd_set * exception_fds = &tmp_efd;
+  P_fd_set read_fds;
+  P_fd_set write_fds;
+  P_fd_set exception_fds;
 
   int retval;
   do {
-
-    FD_ZERO(read_fds);
-    FD_ZERO(write_fds);
-    FD_ZERO(exception_fds);
-
     switch (type) {
       case PChannel::PXReadBlock:
       case PChannel::PXAcceptBlock:
-        FD_SET(handle, read_fds);
+        read_fds = handle;
+        write_fds.Zero();
+        exception_fds.Zero();
         break;
       case PChannel::PXWriteBlock:
-        FD_SET(handle, write_fds);
+        read_fds.Zero();
+        write_fds = handle;
+        exception_fds.Zero();
         break;
       case PChannel::PXConnectBlock:
-        FD_SET(handle, write_fds);
-        FD_SET(handle, exception_fds);
+        read_fds.Zero();
+        write_fds = handle;
+        exception_fds = handle;
         break;
       default:
         PAssertAlways(PLogicError);
@@ -883,20 +1280,19 @@ int PThread::PXBlockOnIO(int handle, int type, const PTimeInterval & timeout)
     }
 
     // include the termination pipe into all blocking I/O functions
-    FD_SET(unblockPipe[0], read_fds);
+    read_fds += unblockPipe[0];
 
-    struct timeval tval;
+    P_timeval tval = timeout;
     retval = ::select(PMAX(handle, unblockPipe[0])+1,
-                      read_fds, write_fds, exception_fds,
-                      timeout.AsTimeVal(tval));
+                      read_fds, write_fds, exception_fds, tval);
   } while (retval < 0 && errno == EINTR);
 
-  if ((retval == 1) && FD_ISSET(unblockPipe[0], read_fds)) {
+  if ((retval == 1) && read_fds.IsPresent(unblockPipe[0])) {
     BYTE ch;
     ::read(unblockPipe[0], &ch, 1);
     errno = EINTR;
     retval =  -1;
-    PTRACE(4, "PWLib\tUnblocked I/O");
+    PTRACE(6, "PWLib\tUnblocked I/O");
   }
 
   return retval;
@@ -934,7 +1330,7 @@ PSemaphore::PSemaphore(unsigned initial, unsigned maxCount)
   maxCountVar = maxCount;
 
 #ifdef P_HAS_SEMAPHORES
-  PAssertOS(sem_init(&semId, 0, initial) == 0);
+  PAssertPTHREAD(sem_init, (&semId, 0, initial));
 #else
   PAssert(maxCount > 0, "Invalid semaphore maximum.");
   if (initial > maxCount)
@@ -956,7 +1352,7 @@ PSemaphore::PSemaphore(const PSemaphore & sem)
   maxCountVar = sem.GetMaxCount();
 
 #ifdef P_HAS_SEMAPHORES
-  PAssertOS(sem_init(&semId, 0, initialVar) == 0);
+  PAssertPTHREAD(sem_init, (&semId, 0, initialVar));
 #else
   PAssert(maxCountVar > 0, "Invalid semaphore maximum.");
   if (initialVar > maxCountVar)
@@ -971,12 +1367,18 @@ PSemaphore::PSemaphore(const PSemaphore & sem)
 PSemaphore::~PSemaphore()
 {
   pthread_cond_destroy(&condVar);
+
+#ifdef P_NETBSD
+  // If the mutex was not locked, the unlock will fail */
+  pthread_mutex_trylock(&mutex);
+#endif
+
   pthread_mutex_unlock(&mutex);
   pthread_mutex_destroy(&mutex);
 
   if (pxClass == PXSemaphore) {
 #ifdef P_HAS_SEMAPHORES
-    PAssertOS(sem_destroy(&semId) == 0);
+    PAssertPTHREAD(sem_destroy, (&semId));
 #else
     PAssert(queuedLocks == 0, "Semaphore destroyed with queued locks");
 #endif
@@ -987,9 +1389,9 @@ PSemaphore::~PSemaphore()
 void PSemaphore::Wait()
 {
 #ifdef P_HAS_SEMAPHORES
-  PAssertOS(sem_wait(&semId) == 0);
+  PAssertPTHREAD(sem_wait, (&semId));
 #else
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
 
   queuedLocks++;
   PThread::Current()->PXSetWaitingSemaphore(this);
@@ -1004,7 +1406,7 @@ void PSemaphore::Wait()
 
   currentCount--;
 
-  PAssertOS(pthread_mutex_unlock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
 #endif
 }
 
@@ -1040,7 +1442,7 @@ BOOL PSemaphore::Wait(const PTimeInterval & waitTime)
   absTime.tv_sec  = finishTime.GetTimeInSeconds();
   absTime.tv_nsec = finishTime.GetMicrosecond() * 1000;
 
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
 
   PThread * thread = PThread::Current();
   thread->PXSetWaitingSemaphore(this);
@@ -1063,7 +1465,7 @@ BOOL PSemaphore::Wait(const PTimeInterval & waitTime)
   if (ok)
     currentCount--;
 
-  PAssertOS(pthread_mutex_unlock((pthread_mutex_t *)&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, ((pthread_mutex_t *)&mutex));
 
   return ok;
 #endif
@@ -1073,17 +1475,17 @@ BOOL PSemaphore::Wait(const PTimeInterval & waitTime)
 void PSemaphore::Signal()
 {
 #ifdef P_HAS_SEMAPHORES
-  PAssertOS(sem_post(&semId) == 0);
+  PAssertPTHREAD(sem_post, (&semId));
 #else
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
 
   if (currentCount < maximumCount)
     currentCount++;
 
   if (queuedLocks > 0) 
-    pthread_cond_signal(&condVar);
+    PAssertPTHREAD(pthread_cond_signal, (&condVar));
 
-  PAssertOS(pthread_mutex_unlock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
 #endif
 }
 
@@ -1092,16 +1494,19 @@ BOOL PSemaphore::WillBlock() const
 {
 #ifdef P_HAS_SEMAPHORES
   if (sem_trywait((sem_t *)&semId) != 0) {
-    PAssertOS(errno == EAGAIN);
+    PAssertOS(errno == EAGAIN || errno == EINTR);
     return TRUE;
   }
-  PAssertOS(sem_post((sem_t *)&semId) == 0);
+  PAssertPTHREAD(sem_post, ((sem_t *)&semId));
   return FALSE;
 #else
   return currentCount == 0;
 #endif
 }
 
+#if defined(P_QNX) && defined(P_HAS_RECURSIVE_MUTEX)
+#define PTHREAD_MUTEX_RECURSIVE_NP PTHREAD_MUTEX_RECURSIVE
+#endif
 
 PMutex::PMutex()
   : PSemaphore(PXMutex)
@@ -1139,18 +1544,21 @@ void PMutex::Wait()
   // if the mutex is already acquired by this thread,
   // then just increment the lock count
   if (pthread_equal(ownerThreadId, currentThreadId)) {
+    // Note this does not need a lock as it can only be touched by the thread
+    // which already has the mutex locked.
     lockCount++;
     return;
   }
 #endif
 
   // acquire the lock for real
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
 
 #ifndef P_HAS_RECURSIVE_MUTEX
-  // set flags
   PAssert((ownerThreadId == (pthread_t)-1) && (lockCount == 0),
           "PMutex acquired whilst locked by another thread");
+  // Note this is protected by the mutex itself only the thread with
+  // the lock can alter it.
   ownerThreadId = currentThreadId;
 #endif
 }
@@ -1170,6 +1578,8 @@ BOOL PMutex::Wait(const PTimeInterval & waitTime)
 
   // if we already have the mutex, return immediately
   if (pthread_equal(ownerThreadId, currentThreadId)) {
+    // Note this does not need a lock as it can only be touched by the thread
+    // which already has the mutex locked.
     lockCount++;
     return TRUE;
   }
@@ -1184,8 +1594,11 @@ BOOL PMutex::Wait(const PTimeInterval & waitTime)
 #ifndef P_HAS_RECURSIVE_MUTEX
       PAssert((ownerThreadId == (pthread_t)-1) && (lockCount == 0),
               "PMutex acquired whilst locked by another thread");
+      // Note this is protected by the mutex itself only the thread with
+      // the lock can alter it.
       ownerThreadId = currentThreadId;
 #endif
+
       return TRUE;
     }
 
@@ -1199,10 +1612,14 @@ BOOL PMutex::Wait(const PTimeInterval & waitTime)
 void PMutex::Signal()
 {
 #ifndef P_HAS_RECURSIVE_MUTEX
-  PAssert(pthread_equal(ownerThreadId, pthread_self()),
-          "PMutex signal failed - no matching wait or signal by wrong thread");
+  if (!pthread_equal(ownerThreadId, pthread_self())) {
+    PAssertAlways("PMutex signal failed - no matching wait or signal by wrong thread");
+    return;
+  }
 
   // if lock was recursively acquired, then decrement the counter
+  // Note this does not need a separate lock as it can only be touched by the thread
+  // which already has the mutex locked.
   if (lockCount > 0) {
     lockCount--;
     return;
@@ -1212,22 +1629,27 @@ void PMutex::Signal()
   ownerThreadId = (pthread_t)-1;
 #endif
 
-  // and unlock for sure
-  PAssert(pthread_mutex_unlock(&mutex) == 0,
-          "PMutex signal failed - no matching wait or signal by wrong thread");
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
+#ifdef P_LINUX
+  PThread::Yield();
+#endif
 }
 
 
 BOOL PMutex::WillBlock() const
 {
+#ifndef P_HAS_RECURSIVE_MUTEX
   pthread_t currentThreadId = pthread_self();
   if (currentThreadId == ownerThreadId)
     return FALSE;
+#endif
 
   pthread_mutex_t * mp = (pthread_mutex_t*)&mutex;
   if (pthread_mutex_trylock(mp) != 0)
     return TRUE;
-  return pthread_mutex_unlock(mp) != 0;
+
+  PAssertPTHREAD(pthread_mutex_unlock, (mp));
+  return FALSE;
 }
 
 
@@ -1245,17 +1667,17 @@ PSyncPoint::PSyncPoint(const PSyncPoint &)
 
 void PSyncPoint::Wait()
 {
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
   while (signalCount == 0)
     pthread_cond_wait(&condVar, &mutex);
   signalCount--;
-  PAssertOS(pthread_mutex_unlock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
 }
 
 
 BOOL PSyncPoint::Wait(const PTimeInterval & waitTime)
 {
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
 
   PTime finishTime;
   finishTime += waitTime;
@@ -1275,7 +1697,7 @@ BOOL PSyncPoint::Wait(const PTimeInterval & waitTime)
   if (err == 0)
     signalCount--;
 
-  PAssertOS(pthread_mutex_unlock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
 
   return err == 0;
 }
@@ -1283,10 +1705,10 @@ BOOL PSyncPoint::Wait(const PTimeInterval & waitTime)
 
 void PSyncPoint::Signal()
 {
-  PAssertOS(pthread_mutex_lock(&mutex) == 0);
+  PAssertPTHREAD(pthread_mutex_lock, (&mutex));
   signalCount++;
-  PAssertOS(pthread_cond_signal(&condVar) == 0);
-  PAssertOS(pthread_mutex_unlock(&mutex) == 0);
+  PAssertPTHREAD(pthread_cond_signal, (&condVar));
+  PAssertPTHREAD(pthread_mutex_unlock, (&mutex));
 }
 
 
