@@ -25,6 +25,29 @@
  *                 Mark Cooke (mpc@star.sr.bham.ac.uk)
  *
  * $Log: video4linux.cxx,v $
+ * Revision 1.40  2003/07/28 09:31:12  dsandras
+ *
+ * Fixed palette problem with recent versions of the bttv driver.
+ *
+ * Revision 1.39  2003/03/20 23:38:06  dereks
+ * Fixes in handling of flags, and test for device opening
+ *
+ * Revision 1.38  2003/03/17 07:52:52  robertj
+ * Removed canCaptureVideo variable as this is really a virtual function to
+ *   distinguish PVideoOutputDevice from PVideoInputDevice, it is not dynamic.
+ *
+ * Revision 1.37  2003/03/06 02:43:43  dereks
+ * Make error messages slightly more descriptive.
+ *
+ * Revision 1.36  2002/09/09 22:16:54  dereks
+ * update hints for spca50x camera.
+ *
+ * Revision 1.35  2002/09/08 22:01:56  dereks
+ * Add support for Intel PC Pro Camera, spca50x driver, Thanks Dennis Gilmore!
+ *
+ * Revision 1.34  2002/06/05 12:29:16  craigs
+ * Changes for gcc 3.1
+ *
  * Revision 1.33  2002/04/21 22:02:35  dereks
  * Tidy up test for existance of video devices. Thanks Guilhem Tardy.
  *
@@ -198,7 +221,7 @@ static struct {
     HINT_CSPICT_ALWAYS_WORKS |
     HINT_CGPICT_DOESNT_SET_PALETTE |
     HINT_HAS_PREF_PALETTE,
-    VIDEO_PALETTE_YUV420P },
+    VIDEO_PALETTE_YUV422 },
 
   /** Sony Vaio Motion Eye camera
       Linux kernel 2.4.7 has meye.c driver module.
@@ -232,6 +255,17 @@ static struct {
    VIDEO_PALETTE_YUV422
   },
 
+ /** Intel PC Pro Camera
+ 
+ */
+  { "SPCA50X USB Camera",
+    "Intel PC Pro Camera uses the spca50x driver",
+    HINT_ONLY_WORKS_PREF_PALETTE        |
+    HINT_HAS_PREF_PALETTE,
+    VIDEO_PALETTE_RGB24
+  },
+
+
   /** Default device with no special settings
    */
   { "",
@@ -295,7 +329,12 @@ BOOL PVideoInputDevice::Open(const PString & devName, BOOL startImmediate)
     return FALSE;
   }
   
-  SetCanCaptureVideo(videoCapability.type & VID_TYPE_CAPTURE);
+  if ((videoCapability.type & VID_TYPE_CAPTURE) == 0) {
+    PTRACE(1,"PVideoInputDevice:: device capablilities reports cannot capture");
+    ::close (videoFd);
+    videoFd = -1;
+    return FALSE;
+  }
 
   hint_index = PARRAYSIZE(driver_hints) - 1;
   PString driver_name(videoCapability.name);  
@@ -393,8 +432,8 @@ PStringList PVideoInputDevice::GetInputDeviceNames()
 
 	if ((entry.Left(5) == "video") || (entry.Left(7) == "capture")) {
 	  PString thisDevice = "/dev/video" + entry.Right(1);
-	  int videoFd;  
-	  if ((videoFd = open(thisDevice, O_RDONLY))) {
+	  int videoFd = ::open((const char *)thisDevice, O_RDONLY);
+	  if (videoFd > 0) {
 	    struct video_capability  videoCaps;
 	    if (ioctl(videoFd, VIDIOCGCAP, &videoCaps) >= 0 &&
 		(videoCaps.type & VID_TYPE_CAPTURE) != 0)
@@ -482,7 +521,7 @@ BOOL PVideoInputDevice::SetChannel(int newChannel)
   struct video_channel channel;
   channel.channel = channelNumber;
   if (::ioctl(videoFd, VIDIOCGCHAN, &channel) < 0) {
-    PTRACE(1,"VideoInputDevice:: Get Channel info failed : "<< ::strerror(errno));    
+    PTRACE(1,"VideoInputDevice:: Get info on channel " << channelNumber << " failed : "<< ::strerror(errno));    
     return FALSE;
   }
   
@@ -491,7 +530,7 @@ BOOL PVideoInputDevice::SetChannel(int newChannel)
 
   // set the information
   if (::ioctl(videoFd, VIDIOCSCHAN, &channel) < 0) {
-    PTRACE(1,"VideoInputDevice:: Setchannel info failed : "<< ::strerror(errno));    
+    PTRACE(1,"VideoInputDevice:: Set info on channel " << channelNumber << " failed : "<< ::strerror(errno));    
     return FALSE;
   }
   
@@ -827,7 +866,7 @@ BOOL PVideoInputDevice::NormalReadProcess(BYTE *resultBuffer, PINDEX *bytesRetur
       }      
     }
 
-    if ((unsigned) ret != frameBytes) {
+    if ((PINDEX)ret != frameBytes) {
       PTRACE(1,"PVideoInputDevice::NormalReadProcess() returned a short read");
       // Not a completely fatal. Maybe it should return FALSE instead of a partial
       // image though?
